﻿//-----------------------------------------------------------------------
// <copyright file="App.xaml.cs" company="Century Systems">
// Copyright (c) 2015 Century Systems Co., Ltd. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System;
using System.Windows;
using CenturySystems.WPFSample.Model;
using CenturySystems.WPFSample.View;
using CenturySystems.WPFSample.ViewModel;

namespace CenturySystems.WPFSample
{
    /// <summary>
    /// App.xaml の相互作用ロジック
    /// </summary>
    public partial class App : Application, IDisposable
    {
        /// <summary>
        /// モデル
        /// </summary>
        private ModelMain model;

        /// <summary>
        /// ビューモデル
        /// </summary>
        private ViewModelMain viewModel;

        /// <summary>
        /// Dispose()が呼ばれたか
        /// </summary>
        private bool disposed = false;

        /// <summary>
        /// オブジェクトを破棄する
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// オブジェクトを破棄する
        /// </summary>
        /// <param name="disposing">マネージドオブジェクトを破棄するか</param>
        protected virtual void Dispose(bool disposing)
        {
            if (this.disposed)
            {
                return;
            }

            if (disposing)
            {
                this.model.Dispose();
            }

            this.disposed = true;
        }
 
        /// <summary>
        /// 起動時の処理
        /// </summary>
        /// <param name="sender">sender</param>
        /// <param name="e">StartupEventArgs</param>
        private void ApplicationStartup(object sender, StartupEventArgs e)
        {
            this.model = new ModelMain();
            this.viewModel = new ViewModelMain(this.model);

            // メインウィンドウ表示
            var window = new MainWindow(this.viewModel);
            window.Show();
        }

        /// <summary>
        /// 終了時の処理
        /// </summary>
        /// <param name="sender">sender</param>
        /// <param name="e">ExitEventArgs</param>
        private void ApplicationExit(object sender, ExitEventArgs e)
        {
            this.model.Finally();
        }
    }
}
