﻿//-----------------------------------------------------------------------
// <copyright file="ModelBase.cs" company="Century Systems">
// Copyright (c) 2015 Century Systems Co., Ltd. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System;
using System.ComponentModel;
using System.Windows;

namespace CenturySystems.WPFSample.Common
{
    /// <summary>
    /// Modelの基本クラス。INotifyPropertyChangedの実装を提供します。 
    /// Model同士でのプロパティ変更イベントの受け渡しにも対応します。
    /// </summary>
    public class ModelBase : INotifyPropertyChanged, IWeakEventListener
    {
        /// <summary>
        /// プロパティの変更があったときに発行されます。 
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// イベントハンドラ(受け側)
        /// </summary>
        public event PropertyChangedEventHandler ModelPropertyChanged;

        /// <summary>
        /// モデルの状態が変化したイベントを受け取る
        /// </summary>
        /// <param name="managerType">マネージャータイプ</param>
        /// <param name="sender">イベントが発生したオブジェクト</param>
        /// <param name="e">引数</param>
        /// <returns>イベントを処理したならtrue</returns>
        public bool ReceiveWeakEvent(Type managerType, object sender, EventArgs e)
        {
            if (managerType == typeof(PropertyChangedEventManager) && e is PropertyChangedEventArgs)
            {
                this.ModelPropertyChanged(sender, (PropertyChangedEventArgs)e);
            }
            else
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// PropertyChangedイベントを発行します。 
        /// </summary>
        /// <param name="propertyName">プロパティ名</param>
        protected virtual void RaisePropertyChanged(string propertyName)
        {
            var h = this.PropertyChanged;
            if (h != null)
            {
                h(this, new PropertyChangedEventArgs(propertyName));
            }
        }
    }
}
