﻿//-----------------------------------------------------------------------
// <copyright file="ViewModelBase.cs" company="Century Systems">
// Copyright (c) 2015 Century Systems Co., Ltd. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System;
using System.ComponentModel;
using System.Windows;

namespace CenturySystems.WPFSample.Common
{
    /// <summary>
    /// ViewModelの基本クラス。
    /// </summary>
    public class ViewModelBase : INotifyPropertyChanged, IWeakEventListener
    {
        /// <summary>
        /// モデルのプロパティの変更があったときに発行されます。
        /// </summary>
        public event PropertyChangedEventHandler ModelPropertyChanged;

        /// <summary>
        /// 自分のプロパティの変更があったときに発行されます。 
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// モデルのプロパティ変更イベントを受け取ります。
        /// </summary>
        /// <param name="managerType">マネージャタイプ</param>
        /// <param name="sender">イベント発行元</param>
        /// <param name="e">引数</param>
        /// <returns>イベントを処理したらtrue</returns>
        public bool ReceiveWeakEvent(Type managerType, object sender, EventArgs e)
        {
            if (managerType == typeof(PropertyChangedEventManager) && e is PropertyChangedEventArgs)
            {
                this.ModelPropertyChanged(sender, (PropertyChangedEventArgs)e);
            }
            else
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// 自分のPropertyChangedイベントを発行します。 
        /// </summary>
        /// <param name="propertyName">プロパティ名</param>
        protected virtual void RaisePropertyChanged(string propertyName)
        {
            var h = this.PropertyChanged;
            if (h != null)
            {
                h(this, new PropertyChangedEventArgs(propertyName));
            }
        }
    }
}
