﻿//-----------------------------------------------------------------------
// <copyright file="SetULS.cs" company="Century Systems">
// Copyright (c) 2015 Century Systems Co., Ltd. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System;
using System.ComponentModel;
using CenturySystems.FLNet;
using CenturySystems.WPFSample.Common;

namespace CenturySystems.WPFSample.Model
{
    /// <summary>
    /// その他のモデル
    /// </summary>
    public class SetULS : ModelBase
    {
        /// <summary>
        /// モデルメイン
        /// </summary>
        private ModelMain modelMain;

        /// <summary>
        /// インスタンスを初期化します。
        /// </summary>
        /// <param name="modelMain">モデルメイン</param>
        public SetULS(ModelMain modelMain)
        {
            this.ModelMain = modelMain;
        }

        /// <summary>
        /// モデルメイン
        /// </summary>
        public ModelMain ModelMain
        {
            get
            {
                return this.modelMain;
            }

            set
            {
                this.modelMain = value;
                if (this.modelMain != null)
                {
                    // イベントを受ける設定
                    this.ModelPropertyChanged += this.OnListenerPropertyChanged;
                    PropertyChangedEventManager.AddListener(this.modelMain, this, string.Empty);
                }
            }
        }

        /// <summary>
        /// FLPCIのインスタンスのメソッドが呼べる状態ならtrue
        /// </summary>
        public bool CanCallFLPCIMethod
        {
            get
            {
                return this.modelMain.CanCallFLPCIMethod;
            }
        }

        /// <summary>
        /// 上位層の状態(ULS)を設定します。
        /// </summary>
        /// <param name="uplStatus">上位層の状態</param>
        public void SetUPLStatus(UPLStatus uplStatus)
        {
            lock (this.modelMain)
            {
                if (this.modelMain.CanCallFLPCIMethod == false)
                {
                    this.modelMain.ErrorMessageToViewModel = "FLPCI is not initialized";
                    return;
                }

                try
                {
                    this.modelMain.FLPCI.SetUplStatus(uplStatus);
                    this.modelMain.ErrorMessageToViewModel = "ULS set.";
                }
                catch (Exception e)
                {
                    this.modelMain.ErrorMessageToViewModel = e.ToString();
                }
            }
        }

        /// <summary>
        /// Modelからのプロパティ変更イベントを受け取ります。
        /// </summary>
        /// <param name="sender">イベント発生源</param>
        /// <param name="e">イベント引数</param>
        private void OnListenerPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var modelMain = (ModelMain)sender;
            var changedProperty = e.PropertyName;
            switch (changedProperty)
            {
                case "CanCallFLPCIMethod":
                    // 自分を参照しているビューモデルにイベントを送る。
                    this.RaisePropertyChanged("CanCallFLPCIMethod");
                    break;
                default:
                    break;
            }
        }
    }
}
