﻿//-----------------------------------------------------------------------
// <copyright file="About.cs" company="Century Systems">
// Copyright (c) 2015 Century Systems Co., Ltd. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System.ComponentModel;
using CenturySystems.WPFSample.Common;
using CenturySystems.WPFSample.Model;

namespace CenturySystems.WPFSample.ViewModel
{
    /// <summary>
    /// バージョン・モデル表示用ViewModel
    /// </summary>
    public class About : ViewModelBase
    {
        /// <summary>
        /// バージョン番号を表すモデル
        /// </summary>
        private FLPCIVersion model;

        /// <summary>
        /// DLLのバージョン
        /// </summary>
        private string dllVersion;

        /// <summary>
        /// APIのバージョン
        /// </summary>
        private string apiVersion;

        /// <summary>
        /// 機種名
        /// </summary>
        private string modelName;

        /// <summary>
        /// ファームのバージョン
        /// </summary>
        private string firmVersion;

        /// <summary>
        /// AboutViewModelを初期化します。
        /// </summary>
        public About()
            : this(null)
        {
        }

        /// <summary>
        /// AboutViewModelを初期化します。
        /// </summary>
        /// <param name="model">対応するモデル</param>
        public About(FLPCIVersion model)
        {
            this.DLLVersion = string.Empty;
            this.APIVersion = string.Empty;
            this.ModelName = string.Empty;
            this.FirmVersion = string.Empty;
            this.Model = model;
        }

        /// <summary>
        /// モデル
        /// </summary>
        public FLPCIVersion Model
        {
            get
            {
                return this.model;
            }

            set
            {
                this.model = value;
                if (value != null)
                {
                    // イベント受け設定
                    this.ModelPropertyChanged += this.OnListenerPropertyChanged;
                    PropertyChangedEventManager.AddListener(this.model, this, string.Empty);

                    // 情報取得
                    this.model.UpdateAPIVersion();
                    this.model.UpdateFirmVersion();
                }
            }
        }

        /// <summary>
        /// DLLのバージョン番号
        /// </summary>
        public string DLLVersion
        {
            get
            {
                return this.dllVersion;
            }

            set
            {
                this.dllVersion = value;
                this.RaisePropertyChanged("DLLVersion");
            }
        }

        /// <summary>
        /// APIのバージョン番号
        /// </summary>
        public string APIVersion
        {
            get
            {
                return this.apiVersion;
            }

            set
            {
                this.apiVersion = value;
                this.RaisePropertyChanged("APIVersion");
            }
        }

        /// <summary>
        /// モデルを表す文字列
        /// </summary>
        public string ModelName
        {
            get
            {
                return this.modelName;
            }

            set
            {
                this.modelName = value;
                this.RaisePropertyChanged("ModelName");
            }
        }

        /// <summary>
        /// ファームウェアのバージョン情報
        /// </summary>
        public string FirmVersion
        {
            get
            {
                return this.firmVersion;
            }

            set
            {
                this.firmVersion = value;
                this.RaisePropertyChanged("FirmVersion");
            }
        }

        /// <summary>
        /// Modelからのプロパティ変更イベントを受け取ります。
        /// </summary>
        /// <param name="sender">イベント発生源</param>
        /// <param name="e">イベント引数</param>
        private void OnListenerPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var version = (FLPCIVersion)sender;
            var changedProperty = e.PropertyName;
            switch (changedProperty)
            {
                case "DLLVersion":
                    this.DLLVersion = version.DLLVersion.ToString();
                    break;
                case "APIVersion":
                    this.APIVersion = version.APIVersion.ToString();
                    break;
                case "ModelName":
                    this.ModelName = version.ModelName.ToString();
                    break;
                case "FirmVersion":
                    this.FirmVersion = version.FirmVersion.ToString();
                    break;
                default:
                    break;
            }
        }
    }
}
