﻿//-----------------------------------------------------------------------
// <copyright file="LogInformation.cs" company="Century Systems">
// Copyright (c) 2015 Century Systems Co., Ltd. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System.ComponentModel;
using CenturySystems.FLNet;
using CenturySystems.WPFSample.Common;

namespace CenturySystems.WPFSample.ViewModel
{
    /// <summary>
    /// ログ情報のビューモデル
    /// </summary>
    public class LogInformation : ViewModelBase
    {
        /// <summary>
        /// モデル
        /// </summary>
        private Model.LogInformation model;

        // ユーザー入力

        /// <summary>
        /// ノード番号
        /// </summary>
        private byte nodeNumber;

        /// <summary>
        /// 表示するログ情報
        /// </summary>
        private LogInfo logInfo;

        /// <summary>
        /// ログ情報を取得するコマンド
        /// </summary>
        private DelegateCommand gettingCommand;

        /// <summary>
        /// ログ情報を取得することが可能かどうか
        /// </summary>
        private bool canUpdateLogInfo;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public LogInformation()
            : this(null)
        {
        }

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="model">対応するモデル</param>
        public LogInformation(Model.LogInformation model)
        {
            this.Model = model;
            this.nodeNumber = 1;
            this.canUpdateLogInfo = false;
        }

        /// <summary>
        /// モデル
        /// </summary>
        public Model.LogInformation Model
        {
            set
            {
                this.model = value;
                if (this.model != null)
                {
                    // イベントで受ける変数の初期値を設定
                    this.canUpdateLogInfo = this.model.CanUpdateLogInfo;

                    // イベントを受ける設定
                    this.ModelPropertyChanged += this.OnListenerPropertyChanged;
                    PropertyChangedEventManager.AddListener(this.model, this, string.Empty);
                }
                else
                {
                    this.canUpdateLogInfo = false;
                }
            }
        }

        /// <summary>
        /// ノード番号
        /// </summary>
        public byte NodeNumber
        {
            get
            {
                return this.nodeNumber;
            }

            set
            {
                this.nodeNumber = value;
            }
        }

        // ここに表示用の属性がずらずらと

        /// <summary>
        /// ソケット送信回数
        /// </summary>
        public string SendCount
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.SendCount.ToString());
            }
        }

        /// <summary>
        /// ソケット送信エラー回数
        /// </summary>
        public string SendError
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.SendError.ToString());
            }
        }

        /// <summary>
        /// イーサネット送信エラー回数
        /// </summary>
        public string EtherSendError
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.EtherSendError.ToString());
            }
        }

        /// <summary>
        /// ソケット受信回数
        /// </summary>
        public string RecvCount
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.RecvCount.ToString());
            }
        }

        /// <summary>
        /// ソケット受信エラー回数
        /// </summary>
        public string RecvError
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.RecvError.ToString());
            }
        }

        /// <summary>
        /// イーサネット受信エラー回数
        /// </summary>
        public string EtherRecvError
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.EtherRecvError.ToString());
            }
        }
        
        /// <summary>
        /// トークン送信回数
        /// </summary>
        public string FrameSendToken
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.FrameSendToken.ToString());
            }
        }

        /// <summary>
        /// サイクリックフレーム送信回数
        /// </summary>
        public string FrameSendCyclic
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.FrameSendCyclic.ToString());
            }
        }

        /// <summary>
        /// 1対1メッセージ送信回数
        /// </summary>
        public string FrameSend1_1Message
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.FrameSend1_1Message.ToString());
            }
        }

        /// <summary>
        /// 1対Nメッセージ送信回数
        /// </summary>
        public string FrameSend1_NMessage
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.FrameSend1_NMessage.ToString());
            }
        }

        /// <summary>
        /// トークン受信回数
        /// </summary>
        public string FrameRecvToken
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.FrameRecvToken.ToString());
            }
        }

        /// <summary>
        /// サイクリックフレーム受信回数
        /// </summary>
        public string FrameRecvCyclic
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.FrameRecvCyclic.ToString());
            }
        }

        /// <summary>
        /// 1対1メッセージ受信回数
        /// </summary>
        public string FrameRecv1_1Message
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.FrameRecv1_1Message.ToString());
            }
        }

        /// <summary>
        /// 1対Nメッセージ受信回数
        /// </summary>
        public string FrameRecv1_NMessage
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.FrameRecv1_NMessage.ToString());
            }
        }

        /// <summary>
        /// サイクリック受信エラー総数
        /// </summary>
        public string ErrorCyclicTotal
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorCyclicTotal.ToString());
            }
        }

        /// <summary>
        /// サイクリックアドレスサイズエラー回数
        /// </summary>
        public string ErrorCyclicAddressSize
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorCyclicAddressSize.ToString());
            }
        }

        /// <summary>
        /// サイクリック CBNエラー
        /// </summary>
        public string ErrorCyclicCBN
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorCyclicCBN.ToString());
            }
        }

        /// <summary>
        /// サイクリック TBNエラー
        /// </summary>
        public string ErrorCyclicTBN
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorCyclicTBN.ToString());
            }
        }

        /// <summary>
        /// サイクリック BSIZEエラー
        /// </summary>
        public string ErrorCyclicBSize
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorCyclicBSize.ToString());
            }
        }

        /// <summary>
        /// メッセージ再送回数
        /// </summary>
        public string ErrorMessageRetry
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorMessageRetry.ToString());
            }
        }

        /// <summary>
        /// メッセージ再送オーバー回数
        /// </summary>
        public string ErrorMessageRetryOver
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorMessageRetryOver.ToString());
            }
        }

        /// <summary>
        /// メッセージ受信エラー回数
        /// </summary>
        public string ErrorMessageRecvTotal
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorMessageRecvTotal.ToString());
            }
        }

        /// <summary>
        /// メッセージ通番エラー回数
        /// </summary>
        public string ErrorMessageTuban
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorMessageTuban.ToString());
            }
        }

        /// <summary>
        /// メッセージ通番再送認識
        /// </summary>
        public string ErrorMessageVersion
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorMessageVersion.ToString());
            }
        }

        /// <summary>
        /// ACKエラー回数
        /// </summary>
        public string ErrorAckTotal
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorAckTotal.ToString());
            }
        }

        /// <summary>
        /// ACKバージョン・エラー回数
        /// </summary>
        public string ErrorAckVersion
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorAckVersion.ToString());
            }
        }

        /// <summary>
        /// ACK通番エラー回数
        /// </summary>
        public string ErrorAckTuban
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorAckTuban.ToString());
            }
        }

        /// <summary>
        /// ACKノード番号エラー回数
        /// </summary>
        public string ErrorAckNodeNum
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorAckNodeNum.ToString());
            }
        }

        /// <summary>
        /// ACK TCDエラー回数
        /// </summary>
        public string ErrorAckTCD
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorAckTCD.ToString());
            }
        }

        /// <summary>
        /// トークン多重化認識回数
        /// </summary>
        public string ErrorTokenTajuka
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorTokenTajuka.ToString());
            }
        }

        /// <summary>
        /// トークン破棄回数
        /// </summary>
        public string ErrorTokenHaki
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorTokenHaki.ToString());
            }
        }

        /// <summary>
        /// トークン再発行回数
        /// </summary>
        public string ErrorTokenSaihakko
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorTokenSaihakko.ToString());
            }
        }

        /// <summary>
        /// トークン保持タイムアウト回数
        /// </summary>
        public string ErrorTokenHojiTimeout
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorTokenHojiTimeout.ToString());
            }
        }

        /// <summary>
        /// トークン監視タイムアウト回数
        /// </summary>
        public string ErrorTokenKanshiTimeout
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.ErrorTokenKanshiTimeout.ToString());
            }
        }

        /// <summary>
        /// トータル稼働時間
        /// </summary>
        public string StatusTotalTime
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.StatusTotalTime.ToString());
            }
        }

        /// <summary>
        /// フレーム待ち状態回数
        /// </summary>
        public string StatusFrameWait
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.StatusFrameWait.ToString());
            }
        }

        /// <summary>
        /// 加入回数
        /// </summary>
        public string StatusKanyu
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.StatusKanyu.ToString());
            }
        }

        /// <summary>
        /// 自己離脱回数
        /// </summary>
        public string StatusJikoRidatsu
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.StatusJikoRidatsu.ToString());
            }
        }

        /// <summary>
        /// スキップ離脱回数
        /// </summary>
        public string StatusSkipRidatsu
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.StatusSkipRidatsu.ToString());
            }
        }

        /// <summary>
        /// 他ノード離脱認識回数
        /// </summary>
        public string StatusOtherRidatsu
        {
            get
            {
                return (this.logInfo == null ? string.Empty : this.logInfo.StatusOtherRidatsu.ToString());
            }
        }

        /// <summary>
        /// ログ情報を取得するコマンド
        /// </summary>
        public DelegateCommand GettingCommand
        {
            get
            {
                if (this.gettingCommand == null)
                {
                    this.gettingCommand = new DelegateCommand(this.UpdateLogInfo, this.CanUpdateLogInfo);
                }

                return this.gettingCommand;
            }
        }

        // メソッド

        /// <summary>
        /// ログ情報を取得する。
        /// </summary>
        public void UpdateLogInfo()
        {
            if (this.model != null)
            {
                this.model.UpdateLogInfo(this.nodeNumber);
            }
        }

        /// <summary>
        /// ログ情報を取得可能かどうか
        /// </summary>
        /// <returns>取得可能ならtrue</returns>
        public bool CanUpdateLogInfo()
        {
            return this.canUpdateLogInfo;
        }

        /// <summary>
        /// Modelからのプロパティ変更イベントを受け取ります。
        /// </summary>
        /// <param name="sender">イベント発生源</param>
        /// <param name="e">イベント引数</param>
        private void OnListenerPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var logInformation = (Model.LogInformation)sender;
            var changedProperty = e.PropertyName;
            switch (changedProperty)
            {
                case "LogInfo":
                    this.logInfo = logInformation.LogInfo;
                    //// 全てのプロパティが更新されたイベントを発生させる
                    this.RaisePropertyChanged(string.Empty);
                    break;
                case "CanUpdateLogInfo":
                    this.canUpdateLogInfo = logInformation.CanUpdateLogInfo;
                    break;
                default:
                    break;
            }
        }
    }
}
