﻿//-----------------------------------------------------------------------
// <copyright file="MyCyclic.cs" company="Century Systems">
// Copyright (c) 2015 Century Systems Co., Ltd. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Globalization;
using CenturySystems.FLPCI;
using CenturySystems.WPFSample.Common;

namespace CenturySystems.WPFSample.ViewModel
{
    /// <summary>
    /// 自ノードサイクリック画面ビューモデル
    /// </summary>
    public class MyCyclic : ViewModelBase
    {
        /// <summary>
        /// モデル
        /// </summary>
        private Model.MyCyclic model;

        /// <summary>
        /// 「変更」のオフセット
        /// </summary>
        private ushort offset1;

        /// <summary>
        /// 「変更」の長さ
        /// </summary>
        private ushort length1;

        /// <summary>
        /// 「変更」のデータ
        /// </summary>
        private ushort data1;

        /// <summary>
        /// 「変更」のオフセット
        /// </summary>
        private ushort offset2;

        /// <summary>
        /// 「変更」の長さ
        /// </summary>
        private ushort length2;

        /// <summary>
        /// 「変更」のデータ
        /// </summary>
        private ushort data2;

        /// <summary>
        /// 「全変更」でセットするデータ
        /// </summary>
        private ushort dataToAll;

        /// <summary>
        /// 「変更」「全変更」の実行が可能かどうか
        /// </summary>
        private bool canSet;

        /// <summary>
        /// ノードを初期化したときの情報
        /// </summary>
        private NodeInitInfo nodeInitInfo;

        /// <summary>
        /// 「変更1」コマンド
        /// </summary>
        private DelegateCommand setCommand1;

        /// <summary>
        /// 「変更2」コマンド
        /// </summary>
        private DelegateCommand setCommand2;

        /// <summary>
        /// 「全変更」コマンド
        /// </summary>
        private DelegateCommand setToAllCommand;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public MyCyclic()
            : this(null)
        {
        }

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="model">モデル</param>
        public MyCyclic(Model.MyCyclic model)
        {
            this.Model = model;

            this.CommonDataList1 = new ObservableCollection<CommonData>();
            this.CommonDataList2 = new ObservableCollection<CommonData>();
        }

        /// <summary>
        /// モデル
        /// </summary>
        public Model.MyCyclic Model
        {
            get
            {
                return this.model;
            }

            private set
            {
                this.model = value;
                if (this.model != null)
                {
                    // イベントで受ける変数の初期値を設定
                    this.canSet = this.model.CanCallFLPCIMethod;

                    // イベントを受ける設定
                    this.ModelPropertyChanged += this.OnListenerPropertyChanged;
                    PropertyChangedEventManager.AddListener(this.model, this, string.Empty);
                }
                else
                {
                    this.canSet = false;
                }
            }
        }

        /// <summary>
        /// 表示するコモンデータのリスト1
        /// </summary>
        public ObservableCollection<CommonData> CommonDataList1 { get; private set; }

        /// <summary>
        /// 表示するコモンデータのリスト2
        /// </summary>
        public ObservableCollection<CommonData> CommonDataList2 { get; private set; }

        /// <summary>
        /// 「変更」のオフセット
        /// </summary>
        public string Offset1
        {
            get
            {
                return this.offset1.ToString();
            }

            set
            {
                ushort number;
                ushort.TryParse(value, out number);
                this.offset1 = number;
            }
        }

        /// <summary>
        /// 「変更」の長さ
        /// </summary>
        public string Length1
        {
            get
            {
                return this.length1.ToString();
            }

            set
            {
                ushort number;
                ushort.TryParse(value, out number);
                this.length1 = number;
            }
        }

        /// <summary>
        /// 「変更」のデータ
        /// </summary>
        public string Data1
        {
            get
            {
                return this.data1.ToString("X4");
            }

            set
            {
                ushort number;
                ushort.TryParse(value, NumberStyles.HexNumber, CultureInfo.InvariantCulture, out number);
                this.data1 = number;
            }
        }

        /// <summary>
        /// 「変更」のオフセット
        /// </summary>
        public string Offset2
        {
            get
            {
                return this.offset2.ToString();
            }

            set
            {
                ushort number;
                ushort.TryParse(value, out number);
                this.offset2 = number;
            }
        }

        /// <summary>
        /// 「変更」の長さ
        /// </summary>
        public string Length2
        {
            get
            {
                return this.length2.ToString();
            }

            set
            {
                ushort number;
                ushort.TryParse(value, out number);
                this.length2 = number;
            }
        }

        /// <summary>
        /// 「変更」のデータ
        /// </summary>
        public string Data2
        {
            get
            {
                return this.data2.ToString("X4");
            }

            set
            {
                ushort number;
                ushort.TryParse(value, NumberStyles.HexNumber, CultureInfo.InvariantCulture, out number);
                this.data2 = number;
            }
        }

        /// <summary>
        /// 「全変更」のデータ
        /// </summary>
        public string DataToAll
        {
            get
            {
                return this.dataToAll.ToString("X4");
            }

            set
            {
                ushort number;
                ushort.TryParse(value, NumberStyles.HexNumber, CultureInfo.InvariantCulture, out number);
                this.dataToAll = number;
            }
        }

        /// <summary>
        /// 「変更1」のコマンド
        /// </summary>
        public DelegateCommand SetCommand1
        {
            get
            {
                if (this.setCommand1 == null)
                {
                    this.setCommand1 = new DelegateCommand(this.Set1, this.CanSet);
                }

                return this.setCommand1;
            }
        }

        /// <summary>
        /// 「変更2」のコマンド
        /// </summary>
        public DelegateCommand SetCommand2
        {
            get
            {
                if (this.setCommand2 == null)
                {
                    this.setCommand2 = new DelegateCommand(this.Set2, this.CanSet);
                }

                return this.setCommand2;
            }
        }

        /// <summary>
        /// 「全変更」のコマンド
        /// </summary>
        public DelegateCommand SetToAllCommand
        {
            get
            {
                if (this.setToAllCommand == null)
                {
                    this.setToAllCommand = new DelegateCommand(this.SetToAll, this.CanSet);
                }

                return this.setToAllCommand;
            }
        }

        /// <summary>
        /// 「変更1」を行う
        /// </summary>
        public void Set1()
        {
            if (this.model == null)
            {
                return;
            }

            this.model.WriteCommonMemory(1, this.offset1, this.length1, this.data1);
            this.ChangeCommonDataList(this.CommonDataList1, this.offset1, this.length1, this.data1);
        }

        /// <summary>
        /// 「変更2」を行う
        /// </summary>
        public void Set2()
        {
            if (this.model == null)
            {
                return;
            }

            this.model.WriteCommonMemory(2, this.offset2, this.length2, this.data2);
            this.ChangeCommonDataList(this.CommonDataList2, this.offset2, this.length2, this.data2);
        }

        /// <summary>
        /// 「全変更」を行う
        /// </summary>
        public void SetToAll()
        {
            if (this.model == null)
            {
                return;
            }

            this.model.WriteCommonMemoryAll(this.dataToAll);
            this.ChangeCommonDataListAll(this.dataToAll);
        }

        /// <summary>
        /// 「変更」「全変更」が可能かどうか
        /// </summary>
        /// <returns>「変更」「全変更」が可能ならtrue</returns>
        public bool CanSet()
        {
            return this.canSet;
        }

        /// <summary>
        /// Modelからのプロパティ変更イベントを受け取ります。
        /// </summary>
        /// <param name="sender">イベント発生源</param>
        /// <param name="e">イベント引数</param>
        private void OnListenerPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var myCyclic = (Model.MyCyclic)sender;
            var changedProperty = e.PropertyName;
            switch (changedProperty)
            {
                case "CanCallFLPCIMethod":
                    this.canSet = myCyclic.CanCallFLPCIMethod;
                    break;
                case "NodeInitInfo":
                    this.nodeInitInfo = myCyclic.NodeInitInfo;
                    this.MakeCommonDataList();
                    break;
                default:
                    break;
            }
        }

        /// <summary>
        /// 自ノード表示用のデータを作成する
        /// </summary>
        private void MakeCommonDataList()
        {
            this.CommonDataList1.Clear();
            this.CommonDataList2.Clear();

            if (this.nodeInitInfo == null)
            {
                return;
            }

            for (ushort i = 0; i < this.nodeInitInfo.Common1Size; i++)
            {
                var commonData = new CommonData(i, (ushort)(this.nodeInitInfo.Common1Address + i), (ushort)0);
                this.CommonDataList1.Add(commonData);
            }

            for (ushort i = 0; i < this.nodeInitInfo.Common2Size; i++)
            {
                var commonData = new CommonData(i, (ushort)(this.nodeInitInfo.Common2Address + i), (ushort)0);
                this.CommonDataList2.Add(commonData);
            }
        }

        /// <summary>
        /// 表示用コモンデータを更新します。
        /// </summary>
        /// <param name="list">表示用データのリスト</param>
        /// <param name="offset">開始オフセット</param>
        /// <param name="size">個数</param>
        /// <param name="data">データ</param>
        private void ChangeCommonDataList(ObservableCollection<CommonData> list, ushort offset, ushort size, ushort data)
        {
            for (ushort i = 0; i < size; i++)
            {
                if (offset + i >= list.Count)
                {
                    break;
                }

                list[offset + i].Data = data;
            }
        }

        /// <summary>
        /// 表示用コモンデータを全て同一のデータで更新します。
        /// </summary>
        /// <param name="data">データ</param>
        private void ChangeCommonDataListAll(ushort data)
        {
            foreach (CommonData commonData in this.CommonDataList1)
            {
                commonData.Data = data;
            }

            foreach (CommonData commonData in this.CommonDataList2)
            {
                commonData.Data = data;
            }
        }

        /// <summary>
        /// 個々のデータを表すクラス
        /// </summary>
        public class CommonData : INotifyPropertyChanged
        {
            /// <summary>
            /// データ(プロパティ変更イベントを起こすためにこの形にしている)
            /// </summary>
            private ushort data;

            /// <summary>
            /// コンストラクタ
            /// </summary>
            /// <param name="offset">オフセット</param>
            /// <param name="address">アドレス(絶対アドレス)</param>
            /// <param name="data">データ</param>
            public CommonData(ushort offset, ushort address, ushort data)
            {
                this.Offset = offset;
                this.Address = address;
                this.Data = data;
            }

            /// <summary>
            /// プロパティが変わった場合のイベントハンドラ
            /// </summary>
            public event PropertyChangedEventHandler PropertyChanged;

            /// <summary>
            /// オフセット
            /// </summary>
            public ushort Offset { get; set; }

            /// <summary>
            /// アドレス(絶対アドレス)
            /// </summary>
            public ushort Address { get; set; }

            /// <summary>
            /// データ(プロパティ変更イベントを起こすためにこの形にしている)
            /// </summary>
            public ushort Data
            {
                get
                {
                    return this.data;
                }

                set
                {
                    this.data = value;
                    var h = this.PropertyChanged;
                    if (h != null)
                    {
                        h(this, new PropertyChangedEventArgs("Data"));
                    }
                }
            }
        }
    }
}
