﻿//-----------------------------------------------------------------------
// <copyright file="NodeInformation.cs" company="Century Systems">
// Copyright (c) 2015 Century Systems Co., Ltd. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System;
using System.ComponentModel;
using CenturySystems.FLPCI;
using CenturySystems.WPFSample.Common;

namespace CenturySystems.WPFSample.ViewModel
{
    /// <summary>
    /// ノード情報のビューモデル
    /// </summary>
    public class NodeInformation : ViewModelBase
    {
        /// <summary>
        /// モデル
        /// </summary>
        private Model.NodeInformation model;

        // ユーザー入力

        /// <summary>
        /// ノード番号
        /// </summary>
        private byte nodeNumber;

        /// <summary>
        /// 表示するノード情報
        /// </summary>
        private NodeInfo nodeInfo;

        /// <summary>
        /// ノード情報を取得するコマンド
        /// </summary>
        private DelegateCommand gettingCommand;

        /// <summary>
        /// ノード情報を取得することが可能かどうか
        /// </summary>
        private bool canUpdateNodeInfo;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public NodeInformation()
            : this(null)
        {
        }

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="model">対応するモデル</param>
        public NodeInformation(Model.NodeInformation model)
        {
            this.Model = model;
            this.nodeNumber = 1;
            this.canUpdateNodeInfo = false;
        }

        /// <summary>
        /// モデル
        /// </summary>
        public Model.NodeInformation Model
        {
            set
            {
                this.model = value;
                if (this.model != null)
                {
                    // イベントで受ける変数の初期値を設定
                    this.canUpdateNodeInfo = this.model.CanUpdateNodeInfo;

                    // イベントを受ける設定
                    this.ModelPropertyChanged += this.OnListenerPropertyChanged;
                    PropertyChangedEventManager.AddListener(this.model, this, string.Empty);
                }
                else
                {
                    this.canUpdateNodeInfo = false;
                }
            }
        }

        /// <summary>
        /// ノード番号
        /// </summary>
        public byte NodeNumber
        {
            get
            {
                return this.nodeNumber;
            }

            set
            {
                this.nodeNumber = value;
            }
        }

        /// <summary>
        /// 参加/離脱
        /// </summary>
        public bool Join
        {
            get
            {
                return this.nodeInfo == null ? false : this.nodeInfo.Join;
            }
        }

        /// <summary>
        /// コモン領域1アドレス
        /// </summary>
        public string Common1Address
        {
            get
            {
                return this.nodeInfo == null ? string.Empty : this.nodeInfo.Common1Address.ToString();
            }
        }

        /// <summary>
        /// コモン領域1サイズ
        /// </summary>
        public string Common1Size
        {
            get
            {
                return this.nodeInfo == null ? string.Empty : this.nodeInfo.Common1Size.ToString();
            }
        }

        /// <summary>
        /// コモン領域2アドレス
        /// </summary>
        public string Common2Address
        {
            get
            {
                return this.nodeInfo == null ? string.Empty : this.nodeInfo.Common2Address.ToString();
            }
        }

        /// <summary>
        /// コモン領域2サイズ
        /// </summary>
        public string Common2Size
        {
            get
            {
                return this.nodeInfo == null ? string.Empty : this.nodeInfo.Common2Size.ToString();
            }
        }

        /// <summary>
        /// トークン監視時間
        /// </summary>
        public string TokenWatchTimeout
        {
            get
            {
                return this.nodeInfo == null ? string.Empty : this.nodeInfo.TokenWatchTimeout.ToString();
            }
        }

        /// <summary>
        /// 最小許容フレーム間隔
        /// </summary>
        public string MinFrameintervalTime
        {
            get
            {
                return this.nodeInfo == null ? string.Empty : this.nodeInfo.MinFrameintervalTime.ToString();
            }
        }

        /// <summary>
        /// ベンダー名
        /// </summary>
        public string VendorName
        {
            get
            {
                return this.nodeInfo == null ? string.Empty : this.nodeInfo.VendorName.ToString();
            }
        }

        /// <summary>
        /// メーカ名
        /// </summary>
        public string MakerName
        {
            get
            {
                return this.nodeInfo == null ? string.Empty : this.nodeInfo.MakerName.ToString();
            }
        }

        /// <summary>
        /// ノード名
        /// </summary>
        public string NodeName
        {
            get
            {
                return this.nodeInfo == null ? string.Empty : this.nodeInfo.NodeName.ToString();
            }
        }

        /// <summary>
        /// 上位層 RUN/STOP
        /// </summary>
        public bool Run
        {
            get
            {
                return this.nodeInfo == null ? false : this.nodeInfo.UPLStatus.Run;
            }
        }

        /// <summary>
        /// 上位層 Alarm
        /// </summary>
        public bool Alarm
        {
            get
            {
                return this.nodeInfo == null ? false : this.nodeInfo.UPLStatus.Alarm;
            }
        }

        /// <summary>
        /// 上位層 Warning
        /// </summary>
        public bool Warning
        {
            get
            {
                return this.nodeInfo == null ? false : this.nodeInfo.UPLStatus.Warning;
            }
        }

        /// <summary>
        /// 上位層 エラーナンバー
        /// </summary>
        public string ErrorNumber
        {
            get
            {
                return this.nodeInfo == null ? string.Empty : this.nodeInfo.UPLStatus.ErrorNumber.ToString("X");
            }
        }

        /// <summary>
        /// リンク 通信無効検知
        /// </summary>
        public bool DisableComm
        {
            get
            {
                return this.nodeInfo == null ? false : this.nodeInfo.LinkStatus.DisableComm;
            }
        }

        /// <summary>
        /// リンク 上位層信号エラー
        /// </summary>
        public bool UPLError
        {
            get
            {
                return this.nodeInfo == null ? false : this.nodeInfo.LinkStatus.UplError;
            }
        }

        /// <summary>
        /// リンク コモンメモリ・データ有効
        /// </summary>
        public bool EnableCommonMemoryData
        {
            get
            {
                return this.nodeInfo == null ? false : this.nodeInfo.LinkStatus.EnableCommonMemoryData;
            }
        }

        /// <summary>
        /// リンク コモンメモリ設定完了
        /// </summary>
        public bool FinishCommonMemoryConfig
        {
            get
            {
                return this.nodeInfo == null ? false : this.nodeInfo.LinkStatus.FinishCommonMemoryConfig;
            }
        }

        /// <summary>
        /// リンク アドレス重複検知
        /// </summary>
        public bool DuplicateAddress
        {
            get
            {
                return this.nodeInfo == null ? false : this.nodeInfo.LinkStatus.DuplicateAddress;
            }
        }

        /// <summary>
        /// ノード情報を取得するコマンド
        /// </summary>
        public DelegateCommand GettingCommand
        {
            get
            {
                if (this.gettingCommand == null)
                {
                    this.gettingCommand = new DelegateCommand(this.UpdateNodeInfo, this.CanUpdateNodeInfo);
                }

                return this.gettingCommand;
            }
        }

        // メソッド

        /// <summary>
        /// ノード情報を取得する。
        /// </summary>
        public void UpdateNodeInfo()
        {
            if (this.model != null)
            {
                this.model.UpdateNodeInfo(this.nodeNumber);
            }
        }

        /// <summary>
        /// ノード情報を取得可能かどうか
        /// </summary>
        /// <returns>取得可能ならtrue</returns>
        public bool CanUpdateNodeInfo()
        {
            return this.canUpdateNodeInfo;
        }

        /// <summary>
        /// Modelからのプロパティ変更イベントを受け取ります。
        /// </summary>
        /// <param name="sender">イベント発生源</param>
        /// <param name="e">イベント引数</param>
        private void OnListenerPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var nodeInfomation = (Model.NodeInformation)sender;
            var changedProperty = e.PropertyName;
            switch (changedProperty)
            {
                case "NodeInfo":
                    this.nodeInfo = nodeInfomation.NodeInfo;
                    //// 全てのプロパティが更新されたイベントを発生させる
                    this.RaisePropertyChanged(string.Empty);
                    break;
                case "CanUpdateNodeInfo":
                    this.canUpdateNodeInfo = nodeInfomation.CanUpdateNodeInfo;
                    break;
                default:
                    break;
            }
        }
    }
}
