﻿//-----------------------------------------------------------------------
// <copyright file="SetULS.cs" company="Century Systems">
// Copyright (c) 2015 Century Systems Co., Ltd. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System;
using System.ComponentModel;
using System.Globalization;
using CenturySystems.FLNet;
using CenturySystems.WPFSample.Common;
using CenturySystems.WPFSample.Model;

namespace CenturySystems.WPFSample.ViewModel
{
    /// <summary>
    /// 自ノードULS設定ビューモデル
    /// </summary>
    public class SetULS : ViewModelBase
    {
        /// <summary>
        /// モデル
        /// </summary>
        private ModelMain model;

        /// <summary>
        /// セットする上位層の状態
        /// </summary>
        private UPLStatus uplStatus;

        /// <summary>
        /// ULSをセットするコマンド
        /// </summary>
        private DelegateCommand setULSCommand;

        /// <summary>
        /// ULSセットが可能かどうかの状態
        /// </summary>
        private bool canSetUPLStatus;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public SetULS()
            : this(null)
        {
        }

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="model">モデル</param>
        public SetULS(ModelMain model)
        {
            this.uplStatus = new UPLStatus(false, false, false, 0);
            this.Model = model;
        }

        /// <summary>
        /// モデル
        /// </summary>
        public ModelMain Model
        {
            get
            {
                return this.model;
            }

            set
            {
                this.model = value;
                if (this.model != null)
                {
                    // イベントで受ける変数の初期値を設定
                    this.canSetUPLStatus = this.model.CanCallFLPCIMethod;

                    // イベントを受ける設定
                    this.ModelPropertyChanged += this.OnListenerPropertyChanged;
                    PropertyChangedEventManager.AddListener(this.model, this, string.Empty);
                }
                else
                {
                    this.canSetUPLStatus = false;
                }
            }
        }

        /// <summary>
        /// Run / Stop
        /// </summary>
        public bool Run
        {
            get
            {
                return this.uplStatus.Run;
            }

            set
            {
                this.uplStatus.Run = value;
            }
        }

        /// <summary>
        /// Alarm
        /// </summary>
        public bool Alarm
        {
            get
            {
                return this.uplStatus.Alarm;
            }

            set
            {
                this.uplStatus.Alarm = value;
            }
        }

        /// <summary>
        /// Warning
        /// </summary>
        public bool Warning
        {
            get
            {
                return this.uplStatus.Warning;
            }

            set
            {
                this.uplStatus.Warning = value;
            }
        }

        /// <summary>
        /// エラーNo.
        /// </summary>
        public string ErrorNumber
        {
            get
            {
                return this.uplStatus.ErrorNumber.ToString("X3");
            }

            set
            {
                try
                {
                    var number = ushort.Parse(value, NumberStyles.HexNumber);
                    this.uplStatus.ErrorNumber = number;
                }
                catch (ArgumentNullException)
                {
                }
                catch (FormatException)
                {
                }
                catch (OverflowException)
                {
                }
                catch (ArgumentOutOfRangeException)
                {
                }
            }
        }

        /// <summary>
        /// ULSを設定するコマンド
        /// </summary>
        public DelegateCommand SetULSCommand
        {
            get
            {
                if (this.setULSCommand == null)
                {
                    this.setULSCommand = new DelegateCommand(this.SetUPLStatus, this.CanSetUPLStatus);
                }

                return this.setULSCommand;
            }
        }

        /// <summary>
        /// ULSを設定する。
        /// </summary>
        public void SetUPLStatus()
        {
            this.Model.Other.SetUPLStatus(this.uplStatus);
        }

        /// <summary>
        /// ULSの設定が可能かどうか
        /// </summary>
        /// <returns>ULSの設定が可能ならtrue</returns>
        public bool CanSetUPLStatus()
        {
            return (this.uplStatus != null) && (this.canSetUPLStatus);
        }

        /// <summary>
        /// Modelからのプロパティ変更イベントを受け取ります。
        /// </summary>
        /// <param name="sender">イベント発生源</param>
        /// <param name="e">イベント引数</param>
        private void OnListenerPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var modelMain = (ModelMain)sender;
            var changedProperty = e.PropertyName;
            switch (changedProperty)
            {
                case "CanCallFLPCIMethod":
                    this.canSetUPLStatus = modelMain.CanCallFLPCIMethod;
                    break;
                default:
                    break;
            }
        }
    }
}
