/*------------------------------------------------------------------------------
TITLE: FL-PCI Card Firmware Downloder for LINUX
DESCRIPTION:
HISTORY:
2010.07.12	FL-PCI/V2-100、FL-PCI/V2-100LのLinux対応
------------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>

#define VERSION		"1.00"

#define O_BINARY	0
#define PCI_CMDSTS	0x0000
#define PCI_DUALRAM	0x20000
#define MAX_BUF 256
#define READ_SIZE (16*1024)
#define BUFF_SIZE 0x1000

#define FLWR_DATA			(PCI_DUALRAM + 0x0B000) // B000-F000 16kbyte
#define FLWR_CS_OPCODE 		(PCI_DUALRAM + 0x0F000)	// char
#define FLWR_CS_STCTRL 		(PCI_DUALRAM + 0x0F001)	// char
#define FLWR_CS_BC			(PCI_DUALRAM + 0x0F002)	// short
#define FLWR_CS_OFFSET 		(PCI_DUALRAM + 0x0F004)	// long

#define DW 0
#define PCI


/*------------------------------------------------------------------------------
グローバル変数定義
------------------------------------------------------------------------------*/
/* グローバル変数 : FL-net PCIカードのオブジェクトハンドル */
int hDevice0 = -1;

#define FLASH_SIZE	(0x10000 * 31)
unsigned char fary[FLASH_SIZE];

FILE *f1, *f2;
unsigned long base_address = 0x00000;
unsigned char buffer[READ_SIZE];
unsigned long highest_address;

/* help message */
#define USAGE	\
	"Usage: download OPTIONS\n"						\
	" OPTIONS := [-h]         : help\n"				\
	"            [-v]         : verbose output\n"	\
	"            [-b {0 | 1}  : board number\n"		\
	"            [-w <string> : writting file\n"	\
	"            [-f <string> : firmware file\n"

/* verbose output flag */
int verbose = 0;


/*------------------------------------------------------------------------------
プロトタイプ定義
------------------------------------------------------------------------------*/
int Download_Exe(unsigned int board_no, char *writting_file, char *firmware_file);
int ConvertFile(char *FileNameToConvert, char *NewFileName);
int jfgets();
long jhtoi(char *s);
int mprocess(char *pnt);
void printv(char *format, ...);


/*------------------------------------------------------------------------------
FUNCTION:	main関数
DESCRIPTION:
ARGUMENTS:
	argc
	argv
RETURN:
	正常：0
	異常：0以外
GLOBAL:
HISTORY:
------------------------------------------------------------------------------*/
int main(int argc, char **argv)
{
	int c;
	int rc = -1;
	char writting_file[256];
	char firmware_file[256];
	unsigned int board_no = -1;

	memset(writting_file, 0x00, sizeof(writting_file));
	memset(firmware_file, 0x00, sizeof(firmware_file));

	// オプション処理
	while ((c = getopt(argc, argv, "b:w:f:hv")) != EOF) {
		switch (c) {
		case 'b':
			if ((strcmp(optarg, "0") == 0) || (strcmp(optarg, "1") == 0)) {
				board_no = atoi(optarg);
			} else {
				printv("Invalid board number. value: %s.\n", optarg);
				goto pexit;
			}
			break;
		case 'w':
			strcpy(writting_file, optarg);
			break;
		case 'f':
			strcpy(firmware_file, optarg);
			break;
		case 'v':
			verbose = 1;
			break;
		case 'h':
			printf("%s", USAGE);
		default:
			goto pexit;
		}
	}

	printv("FL-PCI Card Firmware Downloder V%s\n", VERSION);
	printv("Copyright (C) Century Systems, 2003\n");
	printv("board number : %d\n", board_no);
	printv("writting file: %s\n", writting_file);
	printv("firmware file: %s\n", firmware_file);

	if ((board_no != 0) && (board_no != 1)) {
		printv("Must be set board number.\n");
		goto pexit;
	}
	if ((strlen(writting_file) == 0) || strlen(firmware_file) == 0) {
		printv("Must be set file option.\n");
		goto pexit;
	} 

	rc = Download_Exe(board_no, writting_file, firmware_file);

pexit:
	if (rc == 0){
		printv("Download is successfully completed.\n");
	} else {
		printv("Download is failed. rc = %d\n", rc);
	}
	return rc;
}


/*------------------------------------------------------------------------------
FUNCTION:	ファームウェアの書き込みプログラムのロード、および
			ファームウェアのフラッシュメモリへの書き込み処理
DESCRIPTION:
ARGUMENTS:
	board_no		ボード番号 0 or 1
	writting_file	書き込みプログラムファイル名
	firmware_file	ファームウェアのファイル名
RETURN:
	正常：0
	異常：0以外
GLOBAL:
HISTORY:
------------------------------------------------------------------------------*/
int Download_Exe(unsigned int board_no, char *writting_file, char *firmware_file)
{
	int i, j;
	int rc;
	int fp;
	int r_size;
	unsigned long addr;
	unsigned char sts;
	// pnpci: 共有メモリのアクセスは最適化を抑止する
	volatile unsigned char *pnpci;
	char *NewFileName = "/tmp/tm";

	// open device
	if (board_no == 0) {
		hDevice0 = open("/dev/flnet0", O_RDWR);
		if (hDevice0 < 0) {
			printv("[Writting] open /dev/flnet0 error.\n");
			return -1;
		}
	} else {
		hDevice0 = open("/dev/flnet1", O_RDWR);
		if (hDevice0 < 0) {
			printv("[Writting] open /dev/flnet1 error.\n");
			return -1;
		}
	}
	pnpci = mmap(0, 256*1024, PROT_WRITE | PROT_READ, MAP_SHARED, hDevice0, 0);
	if (pnpci == MAP_FAILED) {
	    close(hDevice0);
		printv("[Writting] mmap error.\n");
		return -1;
	}

	/* (1) Reset ON */
	*pnpci = 0x00;

	//--------------------------------
	// Step1 load SH flash write prg
	//--------------------------------
	printv("[Writting] Step1: Start.\n");

	// convert S->binary
	rc = ConvertFile(writting_file, NewFileName); 
	if (rc < 0){
		printv("[Writting] Step1: File Convert Error.\n");
		return rc;
    }

	// open binary file
	fp = open(NewFileName, O_RDWR | O_BINARY);
	if ( fp < 0 ) {
		printv("[Writting] Step1: File Open Error.\n");
		return -1;
	}

	// do copy
	addr = PCI_DUALRAM; // point dual ram top
	for (;;) {
		// read SH firm file till CRLR
		r_size = read(fp, buffer, READ_SIZE);
		if (r_size == 0) {
			break;
		}
		// write to pci
		for (j = 0; j < r_size; j++) {
			*(addr + pnpci) = buffer[j];
			if (*(addr + pnpci) != buffer[j]) {
				printv("[Writting] Step1: Verify error. (%x:%x)\n",  *(addr + pnpci), buffer[j]);
				return -1;
			}
			addr++;
		}
		printv("[Writting] Step1: Writting block.\n");
		if (r_size < READ_SIZE) {
			break;
		}
	}
	rc = close(fp);	
	printv("[Writting] Step1: Completed.\n");
	// now flash write prg load done

	//--------------------------------
	// Step2 load SH Application prg
	//--------------------------------
	printv("[Writting] Step2: Start.\n");

	// clear dual
	for (i = FLWR_DATA; i < FLWR_DATA + 0x4000; i++) {
		*(i + pnpci) = 0;
	}
	*(FLWR_CS_OPCODE + pnpci) = 0;
	*(FLWR_CS_STCTRL + pnpci) = 0;
	*(FLWR_CS_BC + pnpci) = 0;
	*(FLWR_CS_BC+1 + pnpci) = 0;
	*(FLWR_CS_OFFSET + pnpci) = 0;
	*(FLWR_CS_OFFSET+1 + pnpci) = 0;
	*(FLWR_CS_OFFSET+2 + pnpci) = 0;
	*(FLWR_CS_OFFSET+3 + pnpci) = 0;

	// PCI RESET off, DUAL_START_MODE(not start from flash)
	*(PCI_CMDSTS + pnpci) = 0x0C0 | DW; 

	// convert S->binary
	rc = ConvertFile(firmware_file, NewFileName); 
	if (rc < 0) {
		printv("[Writting] Step2: File Convert Error.\n");
		return rc;
    }

	// open binary file
	fp = open( NewFileName, O_RDWR | O_BINARY);
	if (fp < 0) {
		printv("[Writting] Step2: File Open Error.\n");
		return -1;
	}

	// erase falsh
	*(FLWR_CS_OPCODE + pnpci) = 0x01; // erase ope
	*(FLWR_CS_STCTRL + pnpci) = 0x01; // go
	for (;;) {
		sts = *(FLWR_CS_STCTRL + pnpci);
		if ((sts & 0x01) == 0x00) { // SH side done
			break;
		} else if ((sts & 0x80) == 0x80) {
			printv("[Writting] Step2: Flash Erase error.\n");
			return -1;
		}
	}
	printv("[Writting] Step2: Flash Erase Completed.\n");

	// do copy
	addr = 0;
	for (;;) {
		// バッファを0xffで初期化する。チェックサムの計算は、空き領域がすべて0xff
		// であることが前提なのでここでバッファを0xffにセットすることは重要である。
		// 2006.7.10 Maru
		for (j = 0; j < READ_SIZE; j++) {
			buffer[j] = 0xff;
		}

		// read SH firm file till CRLR
		r_size = read( fp, buffer, READ_SIZE );
		if (r_size == 0) {
			break;
		}

		// write to pci
		for (j = 0; j < READ_SIZE; j++) {
			*(j+FLWR_DATA + pnpci) = buffer[j];
			if (*(j+FLWR_DATA + pnpci) != buffer[j]) {
				printv("[Writting] Step2: Verify error. (%x:%x)\n", *(j+FLWR_DATA + pnpci), buffer[j]);
				return -1;
			}
		}

		// wait untill SH firm finish writing flash
		*(FLWR_CS_OPCODE + pnpci)   = 0x02;               // write ope
		*(FLWR_CS_BC + pnpci)       = READ_SIZE>>8;	      // BC upper 8 bit
		*(FLWR_CS_BC+1 + pnpci)     = READ_SIZE & 0x0ff;  // BC lower 8 bit
		*(FLWR_CS_OFFSET + pnpci)   = (unsigned char)(addr>>24); // offset
		*(FLWR_CS_OFFSET+1 + pnpci) = (unsigned char)(addr>>16); // offset		
		*(FLWR_CS_OFFSET+2 + pnpci) = (unsigned char)(addr>>8);  // offset
		*(FLWR_CS_OFFSET+3 + pnpci) = (unsigned char)(addr);     // offset
		*(FLWR_CS_STCTRL + pnpci)   = 0x01;               // go

		for (;;) {
			sts = *(FLWR_CS_STCTRL + pnpci);
			if((sts & 0x01) == 0x00) {
				break;			// SH side done
			} else if((sts & 0x80) == 0x80) {
				printv("[Writting] Step2: Flash write error.\n");
				return -1;
			}
		}
		addr += READ_SIZE;

		// update windows
		printv("[Writting] Step2: Writting block.\n");

		if(r_size < READ_SIZE) {
			break;
		}
	}
	close(fp);	

	// now SH firm prg load done
	close(hDevice0);

	printv("[Writting] Step2: Completed.\n");
	return 0;
}


/*------------------------------------------------------------------------------
FUNCTION:	
DESCRIPTION:
ARGUMENTS:
RETURN:
	正常：0
	異常：0以外
GLOBAL:
HISTORY:
------------------------------------------------------------------------------*/
int ConvertFile(char *FileNameToConvert, char *NewFileName)
{
	int i, j, quit, r_code, size;
	unsigned short sum, tm;

	if ((f1 = fopen(FileNameToConvert, "rb")) == NULL) {
		printv("[ConvertFile] open error. (%s)\n", FileNameToConvert);
		return -1;
	}
	
	if ((f2 = fopen(NewFileName, "wb")) == NULL) {
		fclose(f1);
		printv("[ConvertFile] open error. (%s)\n", NewFileName);
		return -1;
	}

	// clear the array
	for (i = 0, j = sizeof(fary); j > 0; i++, j--) {
//		fary[i] = 0x00;
		fary[i] = 0xff; // 2006.7.10 Maru
	}
	quit = 0;
	do {
		// Get one line at a time and process
		r_code = jfgets();
		if (r_code != 0) {
			r_code = mprocess((char*)buffer);
			if (r_code == 0) {
				;
			} else {
				fclose(f1);
				fclose(f2);
				printv("[ConvertFile] mprocess error. (%d)\n", r_code );
				return -1;
			}
		} else {
			r_code = fwrite(fary, sizeof(char), highest_address - base_address, f2);
			quit = 1;
		}
	} while(quit == 0);

//-- prep for check sum
	sum = 0;
	size = highest_address - base_address;
	for (i = 0; i < size; i += 2) {
		tm = (((unsigned short)fary[i] << 8) & 0x0ff00) |
			  ((unsigned short)(fary[i + 1]) & 0x0ff);
		sum += tm;
	}
	for (i = size + 2; i < 0x80000; i += 2) { // 0x80000 = 512Kbyte
		sum += 0xffff;
	}
	//TRACE("\n bsum size = %x %x", sum, size);
	sum = ~sum;
	sum += 1;
	tm = (sum << 8) & 0x0ff00;
	tm |= ((sum >> 8) & 0x0ff);
	r_code = fwrite(&tm, sizeof(char), 2, f2);

//--

	fclose(f1);
	fclose(f2);
	return 0;
}


/*------------------------------------------------------------------------------
FUNCTION:	
DESCRIPTION:
ARGUMENTS:
RETURN:
GLOBAL:
HISTORY:
------------------------------------------------------------------------------*/
int jfgets()
{
	long i, j;
	char temp;

	for (i = 0, j = sizeof(buffer) - 1; j > 0; j--) {
		temp = fgetc(f1);
		if (temp != EOF) {
//			if ((temp == '\n') || (temp == '\t') || (temp == '\r')) {
			if ((temp == '\n') || (temp == '\t')) {
				buffer[i++] = '\0';
				break;
			} else if(temp == '\r') {
				;
			} else {
				buffer[i++] = temp;
			}
		}
	}
	if (i > 0) {
		return i;
	} else {
		return (int)NULL;
	}
}


/*------------------------------------------------------------------------------
FUNCTION:	
DESCRIPTION:
ARGUMENTS:
RETURN:
GLOBAL:
HISTORY:
------------------------------------------------------------------------------*/
int mprocess(char *pnt)
{
	long byte_count;
	long address;
	char tbuff[BUFF_SIZE];
	char *fpnt;
	unsigned char cdata;

	// Check to see if first character is 'S'
	if (pnt[0] != 'S') {
		return(-1);
	}

	// Throw away S0, S7, S8, and S9 records
	switch(pnt[1]) {
	case	'0':
	case	'7':
	case	'8':
	case	'9':	return(0); break;
	// Error on S4, S5 and S6 records
	case	'4':
	case	'5':
	case	'6':	return(-2); break;
	// Process S1, S2, and S3 recores
	case	'1':
	case	'2':
	case	'3':	break;
	// Anything else, error
	default:		return(-3); break;
	}

	// Got here, must be a good line
	// Get the byte count from pnt[2] and pnt[3]
	tbuff[0] = pnt[2];
	tbuff[1] = pnt[3];
	tbuff[2] = 0x00;

	// Convert to int
	byte_count = jhtoi(tbuff);

	// Now get the address, 3 forms, S1 (4 bytes), S2 (6 bytes), and S3 (8 bytes)
	switch(pnt[1]) {
	case '1':	tbuff[0] = pnt[4];
				tbuff[1] = pnt[5];
				tbuff[2] = pnt[6];
				tbuff[3] = pnt[7];
				tbuff[4] = 0x00;
				fpnt	 = &pnt[8];
				byte_count -= 3;	// 2 for address, 1 for checksum
				break;

	case '2':	tbuff[0] = pnt[4];
				tbuff[1] = pnt[5];
				tbuff[2] = pnt[6];
				tbuff[3] = pnt[7];
				tbuff[4] = pnt[8];
				tbuff[5] = pnt[9];
				tbuff[6] = 0x00;
				fpnt	 = &pnt[10];
				byte_count -= 4;	// 3 for address, 1 for checksum
				break;
	default:	tbuff[0] = pnt[4];
				tbuff[1] = pnt[5];
				tbuff[2] = pnt[6];
				tbuff[3] = pnt[7];
				tbuff[4] = pnt[8];
				tbuff[5] = pnt[9];
				tbuff[6] = pnt[10];
				tbuff[7] = pnt[11];
				tbuff[8] = 0x00;
				byte_count -= 5;	// 4 for address, 1 for checksum
				fpnt	 = &pnt[12];
				break;
	}

	// Convert to int
	address = jhtoi(tbuff);

	// cal the highest_address
	highest_address = address + byte_count;

	// Now convert the address to an offset
	address -= base_address;	// base address of the FLASH

	if (address < 0) {
		return (-4);
	}
	if (address >= 0x06000000) {
		return 0;
	}

	while (byte_count > 0) {
		// Nov convert data and write in the memory array
		// tpnt address was setup in the above case statement
		tbuff[0] = *fpnt++;
		tbuff[1] = *fpnt++;
		tbuff[2] = 0x00;

		// Now convert the data
		cdata = (unsigned char) jhtoi(tbuff);
		// Now write to memory array
		fary[address++] = cdata;
		byte_count--;
	}

	// Got here, no errors
	return(0);
}


/*------------------------------------------------------------------------------
FUNCTION:	
DESCRIPTION:
ARGUMENTS:
RETURN:
GLOBAL:
HISTORY:
------------------------------------------------------------------------------*/
long jhtoi(char *s)
{
	int n;

	n = 0;
	while (*s != '\0') {
		if(*s >= '0' && *s <= '9') {
			n = (0x10 * n) + (*s++ - '0');
		}
		else if(*s >= 'a' && *s <= 'f') {
			n = (0x10 * n) + (*s++ - 'W');
		}
		else if(*s >= 'A' && *s <= 'F') {
			n = (0x10 * n) + (*s++ - '7');
		}
		else {
			break;
		}
	}
	return(n);
}


/*------------------------------------------------------------------------------
FUNCTION:	printv - verbose output
DESCRIPTION:
ARGUMENTS:	printf互換
RETURN:		None
GLOBAL:
	verbose
HISTORY:
------------------------------------------------------------------------------*/
void printv(char *format, ...)
{
	va_list args;
	char buff[1024];

	va_start(args, format);
	vsnprintf(buff, sizeof(buff), format, args);
	syslog(LOG_USER | LOG_ERR, buff);
	if (verbose) {
		fputs(buff, stdout);
	}
	va_end(args);
}
