/*
 FLNet PCI Card Driver for LINUX 2.6.x 
 Copyright(C) Century Systems Inc.
  (www.centurysys.co.jp)

*/

#define MODVERSIONS
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <asm/uaccess.h>
#include <linux/vmalloc.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/cdev.h>

/* debugs */
#define DEBUG(x) ;

/* typedefs */
typedef volatile unsigned char VBYTE;
typedef volatile u32 V32;
typedef u32 DWORD;

/* defines for PCI Card */
#define VENDOR_ID 	  0x123c 	            /* century systems inc. */
#define QL_DEVICE_ID  0x0005                /* FL-PCI/V2-100  QUICK LOGIC  */
#define PLX_DEVICE_ID 0x001d 	            /* FL-PCI/V2-100L PLX PCI9030  */
#define PCI_RANGE     0x40000 	            /* 256kbyte */
#define DEVICE_NAME "FLNet"  	            /* device name */
#define FL_BOARD(dev) (MINOR(dev)&0x0f) /* borad # = device minor # */
#define FL_BD_MAX 	2                      /* max board # */
#define DEV_COUNT	2
/* proto types */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
#define __devinit
#define __devexit
#define __devinitdata
#endif

#ifndef VM_RESERVED
#define VM_RESERVED (VM_DONTEXPAND | VM_DONTDUMP)
#endif

static int fl_open(struct inode *inode, struct file *file);
static int fl_release(struct inode *inode, struct file *file);
static int fl_mmap(struct file *file, struct vm_area_struct *vma);

static void fl_vma_open(struct vm_area_struct *vma);
static void fl_vma_close(struct vm_area_struct *vma);

struct vm_operations_struct fl_vm_ops = {
	fl_vma_open, fl_vma_close,
};

static dev_t       dev_num;
static struct cdev char_dev;

/* file operation  */
struct file_operations fl_fops = {
    owner:        THIS_MODULE, 
    mmap:         fl_mmap,
    open:         fl_open,
    release:      fl_release,
};

/* driver structures */
struct fl_struct {
	DWORD paddr;	              /* physical address */
	unsigned char *vaddr;	      /* virtical address for driver */
	DWORD memstart;	/* virtical address for driver */
    DWORD memlen;                 /* virtical memory length */
	struct resource *resource;    /* struct resource addr   */
};
struct fl_struct *FL_dev[FL_BD_MAX]={0,0};
int FL_bd_no = 0;	/* deteced board # */
int FL_dev_major = 0;	/* driver Major # */

MODULE_LICENSE("GPL"); 

/* load driver */
static int __devinit flpci_init(struct pci_dev *pci_dev, const struct pci_device_id *device_id)  
{  
	int               i;
	unsigned char pci_bus;
	unsigned int   pci_device_fn;
	int            bar;
	DWORD  addr;
	unsigned short command;
	int				ret;
	int            baseminor = 0;

// memo 2008.11.13 debug
	DEBUG( printk("[flpci_init] called.\n") );
	if( FL_BD_MAX <= FL_bd_no ){
		printk("[flpci_init] FL_bd_no is greater than FL_BD_MAX." );
		goto ERR;
	} 

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	printk("\n[flpci_init] Require kernel 2.6.0 later\n");
	return -ENODEV;
#endif
	DEBUG(printk("\n[flpci_init] Found card\n"));

#if 0
    printk("\n[FLNet] FL_bd_no=%x\n",FL_bd_no);
    printk("\n[FLNet] bus=%x dev_fn=%x\n", pci_bus, pci_device_fn);
#endif
	printk( "\n[flpci_init] device = %x\n", pci_dev->device );
    if(pci_enable_device( pci_dev ))
		goto ERR;

	/* set board info into device structure */
	FL_dev[FL_bd_no] = kmalloc( sizeof(struct fl_struct)+32,
							 GFP_KERNEL);
	if(!FL_dev[FL_bd_no]) {
		printk("[flpci_init] kmalloc err\n");
		goto ERR;
	}

	FL_dev[FL_bd_no]->vaddr = 0;
	if( pci_dev -> device == QL_DEVICE_ID ){
		bar = 0;
    }
	else if( pci_dev -> device == PLX_DEVICE_ID ){
		bar = 2;
	}
	else{
		goto ERR;
	}

	FL_dev[FL_bd_no]->memstart = pci_resource_start( pci_dev, bar );
	DEBUG( printk("[flpci_init] memstart = %lx\n", FL_dev[FL_bd_no]->memstart ));

	FL_dev[FL_bd_no]->memlen   = pci_resource_len( pci_dev, bar );
	DEBUG( printk("[flpci_init] memlen = %lx\n", FL_dev[FL_bd_no]->memlen ));
	if( !(FL_dev[FL_bd_no]->resource = request_mem_region( FL_dev[FL_bd_no]->memstart,
															FL_dev[FL_bd_no]->memlen,
															"fldrv" )) ){
		printk("[flpci_init] request_mem_region error /start=%x len=%x\n",
						FL_dev[FL_bd_no]->memstart, FL_dev[FL_bd_no]->memlen );
		goto ERR;
    }

	pci_bus = pci_dev->bus->number;
	pci_device_fn = pci_dev->devfn;

	/* retrieve BASE Address */
	if( bar == 0 ){
		pci_read_config_dword(pci_dev, PCI_BASE_ADDRESS_0, &addr);
	}
	else{
		/* PLX の場合、コマンドレジスタのI/O空間ビットをO　にする必要がある。*/
		pci_read_config_word(pci_dev, PCI_COMMAND, (unsigned short *)&command);
		command &= 0xfe;
		pci_write_config_word(pci_dev, PCI_COMMAND, command );
		pci_read_config_dword(pci_dev, PCI_BASE_ADDRESS_2, &addr);
	}

	FL_dev[FL_bd_no]->paddr = (DWORD)addr;

	/* Map bus(physical) address into "KERNEL" virtual address. */
	FL_dev[FL_bd_no]->vaddr = ioremap(addr, PCI_RANGE);
	if(!FL_dev[FL_bd_no]->vaddr) {
		printk("[flpci_init] ioremap err\n");
		goto ERR;
	}

	DEBUG( printk("\n[flpci_init] pci Bus(phys)addr=%lx, Virtual addr= %lx\n",
		FL_dev[FL_bd_no]->paddr, FL_dev[FL_bd_no]->vaddr) );

#if 0
	DEBUG(printk("\n virt, bus, phys addr=%lx %lx %lx",
			bus_to_virt(FL_dev[FL_bd_no]->paddr),
			virt_to_bus(FL_dev[FL_bd_no]->vaddr),
			virt_to_phys(FL_dev[FL_bd_no]->vaddr)));
#endif

    FL_bd_no += 1;

	/* turn on LEDs */
	if(FL_bd_no == 1){
	    /* register device */
    	ret = alloc_chrdev_region( &dev_num, baseminor, DEV_COUNT, DEVICE_NAME );	  
   		if( ret < 0 ) {
			printk( "[flpci_init] Touroku Shippai.\n" );
		 	goto ERR;
		}
		FL_dev_major = MAJOR(dev_num);
		cdev_init( &char_dev, &fl_fops );
		char_dev.owner = THIS_MODULE;
		ret = cdev_add( &char_dev, dev_num, DEV_COUNT );
		if( ret < 0 ) {
			goto ERR;
		}
		printk( "[flpci_init] Touroku Seikou. Major Number = %d\n", FL_dev_major ); 
		*(V32 *)(FL_dev[0]->vaddr) |= 0x00001000; /* LED #1 on */
	}
	else if(FL_bd_no == 2) 
		*(V32 *)(FL_dev[1]->vaddr) |= 0x00003000; /* LED #1 & #2 on */

	return 0;

ERR:
	printk("[flpci_init] Error Detect\n");
	for(i=0; i<FL_BD_MAX; i++) {
		if(FL_dev[i]){
		    if(FL_dev[i]->resource)
                release_mem_region(FL_dev[i]->memstart, FL_dev[i]->memlen);
			if(FL_dev[i]->vaddr)
			    iounmap((void *)FL_dev[i]->vaddr);
			kfree(FL_dev[i]);
		    FL_dev[i] = 0;
		}
	}
	if(FL_dev_major > 0){
		cdev_del( &char_dev );
		unregister_chrdev_region( dev_num, DEV_COUNT );
	}
    FL_dev_major = 0;
	FL_bd_no = 0;

	return -ENODEV;
}

/* unload driver */
static void __devexit flpci_exit(struct pci_dev *pci_dev)
{
	int i;

	if(FL_dev_major == 0)
		return;

	if(FL_bd_no == 1){ 
		if(FL_dev[0] && FL_dev[0]->vaddr)
		    *(V32 *)(FL_dev[0]->vaddr) &= ~(0x0000108f);  /* LED #1 off, Reset ON */
	}
	else if(FL_bd_no == 2){  
		if(FL_dev[0] && FL_dev[0]->vaddr)
		    *(V32 *)(FL_dev[0]->vaddr) &= ~(0x0000108f); /* LED #1 off, Reset ON */
		if(FL_dev[1] && FL_dev[1]->vaddr)
		    *(V32 *)(FL_dev[1]->vaddr) &= ~(0x0000308f); /* LED #1 & #2 off, Reset ON */
    }
	else{
		printk("[FL-cleanup_module] bd_no = %d\n",FL_bd_no );
	}

	for(i=0; i<FL_bd_no; i++) {
		if(FL_dev[i]) {
            release_mem_region(FL_dev[i]->memstart, FL_dev[i]->memlen);
			iounmap((void *)FL_dev[i]->vaddr);
			kfree(FL_dev[i]);
			FL_dev[i] = 0;
		}
		else {
			printk("[FLNet-cleanup_module] fldev invalid\n");
		}
	}
	FL_bd_no = 0;

	if(FL_dev_major){
		cdev_del( &char_dev );
		unregister_chrdev_region( dev_num, DEV_COUNT );
		printk( "[FL-cleanup_module] Sakujyo Seikou. Major Number = %d\n", MAJOR(dev_num) ); 
	}
	else
	    printk("[FLNet-cleanup_module] Major# invalid\n");

	FL_dev_major = 0;

	DEBUG( printk("[FLNet]See ya later\n") );
	return;
}

/* open */
static int fl_open(struct inode *inode, struct file *file)
{
	if(FL_BOARD(inode->i_rdev) >= FL_bd_no) {
		printk("[fl_open] error1=%x\n",FL_BOARD(inode->i_rdev));
		return -ENODEV;
	}
	file->private_data = (void *)((uintptr_t)FL_BOARD(inode->i_rdev));
	return 0;
}

/* close */
static int fl_release(struct inode *inode, struct file *file)
{
	file->private_data = (void *)0;
	return 0;
}

/* mmap */
static int fl_mmap(struct file *file, struct vm_area_struct *vma)
{
	int bno;
	uintptr_t bno_temp;
	unsigned long off = vma->vm_pgoff << PAGE_SHIFT;    /* Ver 2.4.x */
	DWORD physical;
	DWORD vsize = vma->vm_end - vma->vm_start;
	DWORD psize = PCI_RANGE - off;

	bno_temp = (uintptr_t) file->private_data; /* board # */
	bno = (int) bno_temp;
	if(bno >= FL_bd_no) {
		printk("[fl_mmap] bd no  error\n");
		return -EINVAL;
	}
	physical = FL_dev[bno]->paddr + off ;

	if(off & (PAGE_SIZE-1)) {
		printk("[fl_mmap] offset alignment error\n");
		return -ENXIO; 
	}
	if(vsize > psize) { /* required size ?? */
		printk("[fl_mmap] request size error\n");
		return -EINVAL;
	}

	if( (off >= __pa(high_memory)) || (file->f_flags & O_SYNC) )   /* Ver 2.4.x */
		vma->vm_flags |= VM_IO;
    vma->vm_flags |= VM_RESERVED; /* Don't swap */

#if LINUX_VERSION_CODE == KERNEL_VERSION(2, 6, 9)
	if( remap_page_range( vma, vma->vm_start, physical, vsize, vma->vm_page_prot) ) {
#else   
	if( remap_pfn_range(vma, vma->vm_start, physical >> PAGE_SHIFT, vsize, vma->vm_page_prot) ) {
#endif
		printk("[fl_mmap] remap error\n");
		return -EAGAIN;
	}

	if(vma->vm_ops) {
		printk("[fl_mmap] fops error\n");
		return -EINVAL;
	}

	vma->vm_ops = &fl_vm_ops;

	DEBUG(printk("[fl_mmap] remap=%lx %lx %lx\n", vma->vm_start, physical, vsize));
	DEBUG(printk("\n off=%lx padrr=%lx\n", vma->vm_pgoff, FL_dev[bno]->paddr));

	return 0;
}
/* mmap open */
static void fl_vma_open(struct vm_area_struct *vma)
{
	DEBUG(printk("[fl_vma_open]%lx\n", vma->vm_start));
}
/* mmap close */
static void fl_vma_close(struct vm_area_struct *vma)
{
	DEBUG(printk("[fl_vma_close]%lx\n", vma->vm_start));
}

/* Ver 2.4.x */
static struct pci_device_id flpci_id_tbl[] __devinitdata = {
	{ VENDOR_ID, QL_DEVICE_ID,  PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ VENDOR_ID, PLX_DEVICE_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0, }
};
static struct pci_driver flpci_driver = {
    name:        DEVICE_NAME,
    id_table:     flpci_id_tbl,
    probe:       flpci_init,
    remove:     flpci_exit,
};

static int __init pci_drv_init(void)
{
	int rc;
	rc = pci_register_driver(&flpci_driver);
	printk("[fldrv] pci_drv_init, rc: %d\n", rc);
	return rc;
}
static void __exit pci_drv_exit(void)
{
    pci_unregister_driver( &flpci_driver );
}

module_init(pci_drv_init);
module_exit(pci_drv_exit);


/* end of file */
