//
// flpci.c : FL-PCI用APIライブラリ
//

#include "flpci.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/file.h>

#pragma pack(1)

#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define FALSE	0
#endif

// DLL Version No.
#define DLL_VERSION		0x0003

// Application <--> FL-net Card Command/Responce
// AP to FL Command
#define REQ_INIT		((unsigned short) 0x0001)
#define REQ_NODE_INF	((unsigned short) 0x0003)
#define REQ_LOG			((unsigned short) 0x0004)
#define REQ_NET_INF		((unsigned short) 0x0005)
#define REQ_MSG_SND		((unsigned short) 0x0006)
#define REQ_SANKA_NODE	((unsigned short) 0x0007)
#define REQ_SET_ULS		((unsigned short) 0x0008)
#define REQ_ULS_LKS		((unsigned short) 0x0009)
#define REQ_REVISION	((unsigned short) 0x000A)
#define REQ_SPEED		((unsigned short) 0x000B)

// FL to AP Responce
#define RSP_INIT		((unsigned short) 0x0081)
#define RSP_NODE_INF	((unsigned short) 0x0083)
#define RSP_LOG			((unsigned short) 0x0084)
#define RSP_NET_INF		((unsigned short) 0x0085)
#define RSP_MSG_SND		((unsigned short) 0x0086)
#define RSP_SANKA_NODE	((unsigned short) 0x0087)
#define RSP_SET_ULS		((unsigned short) 0x0088)
#define RSP_ULS_LKS		((unsigned short) 0x0089)
#define RSP_REVISION	((unsigned short) 0x008A)
#define RSP_SPEED		((unsigned short) 0x008B)
#define RSP_TEST		((unsigned short) 0x0090)

// FL to AP Command
#define COM_ENABLE		((unsigned short) 0x0020)
#define MSG_RCV			((unsigned short) 0x0021)
#define MSG_SND_END		((unsigned short) 0x0022)
#define REP_FATAL_ERR	((unsigned short) 0x0023)

// AP to FL Responce
#define RSP_ENABLE		((unsigned short) 0x00A0)
#define RSP_RCV			((unsigned short) 0x00A1)
#define RSP_SND_END		((unsigned short) 0x00A2)

// グローバル変数 : FL-net PCIカードの先頭アドレス
unsigned char	*pnpci = 0;
unsigned long	PnSize = 256*1024;

// デュアルポートRAMアドレスの定義
#define APtoFL_CMD		(pnpci + 0x20000)
#define APtoFL_RSP		(pnpci + 0x20002)
#define FLtoAP_CMD		(pnpci + 0x20004)
#define FLtoAP_RSP		(pnpci + 0x20006)
#define CYC_CTL_TBL		(pnpci + 0x20008)
#define APtoFL_DATA		(pnpci + 0x21000)
#define FLtoAP_DATA		(pnpci + 0x21800)
#define CMNBUF1_COM1	(pnpci + 0x22000)
#define CMNBUF1_COM2	(pnpci + 0x22400)
#define CMNBUF2_COM1	(pnpci + 0x26400)
#define CMNBUF2_COM2	(pnpci + 0x26800)
#define APtoFL_CBUF		(pnpci + 0x2A800)
#define FLtoAP_CBUF		(pnpci + 0x2B000)


///////////////////////////////////
// Dual Port RAM I/F 構造体
///////////////////////////////////
// PCI Card 初期化コマンドのデータ部の構造
struct PciCardInitData {
	unsigned int	mode;		// 動作モード
	unsigned int	ip_adr;		// IPアドレス
	unsigned short	c1_adr;		// コモンメモリ１のアドレス
	unsigned short	c1_siz;		// コモンメモリ１のサイズ
	unsigned short	c2_adr;		// コモンメモリ２のアドレス
	unsigned short	c2_siz;		// コモンメモリ２のサイズ
	unsigned char	tw;			// トークン監視間隔
	unsigned char	mft;		// 最小フレーム間隔
	unsigned char	vender[10];	// ベンダ名
	unsigned char	maker[10];	// メーカー名
	unsigned char	node[10];	// ノード名
	unsigned int	sum_flag;	// SUMチェック有無 12345678h=有り
	unsigned int	sum;
};

// サイクリックコントロールテーブルの構造
struct CyclicControlTable {
	unsigned char	node;		// ノード番号(1～254)
	unsigned char	last_buf;	// 最新のデータを持つバッファNo. 0:NO DATA 1:BUFF1 2:BUFF2
	unsigned char	ap_use;		// アプリケーション側使用バッファ 0:NOT USE 1:BUFF1 2:BUFF2
	unsigned char	if_use;		// PCI Card側使用バッファ 0:NOT USE 1:BUFF1 2:BUFF2
	unsigned short	sanka_flg;	// 0:離脱 1:ネットワーク参加
	unsigned short	cmn1_addr;	// コモン領域１アドレス
	unsigned short	cmn1_size;	// コモン領域１サイズ
	unsigned short	cmn2_addr;	// コモン領域２アドレス
	unsigned short	cmn2_size;	// コモン領域２サイズ
};

// ノード情報要求コマンド構造体
struct NodeInfoTable {
	char			result;			// 結果コード
	unsigned char	sanka;			// 参加／離脱 フラグ
	unsigned short	uls;			// 上位層ステータス
	unsigned short	cmn1_addr;		// コモン領域１アドレス
	unsigned short	cmn1_size;		// コモン領域１サイズ
	unsigned short	cmn2_addr;		// コモン領域２アドレス
	unsigned short	cmn2_size;		// コモン領域２サイズ
	unsigned short	rct;			// リフレッシュサイクル許容時間
	unsigned char	tw;				// トークン監視時間
	unsigned char	mft;			// 最小フレーム間隔
	unsigned char	link_status;	// リンクの状態
	unsigned char	my_status;		// 自ノードの状態
	unsigned char	vender[10];		// ベンダー名
	unsigned char	maker[10];		// メーカー名
	unsigned char	name[10];		// ノード名
};

// メッセージコマンド構造体
struct MessageCmdTable {
	unsigned char	msg_node;		// ノード番号(1～254) 
	unsigned char	msg_result;		// 結果コード
	unsigned short	dummy;			// 未使用
	unsigned int	msg_tcd;		// トランザクションコード
	unsigned short	msg_bc;			// メッセージデータバイト長
	unsigned short	msg_lng;		// 仮想アドレス空間 データ長
	unsigned int	msg_adr;		// 仮想アドレス空間 アドレス
	unsigned char	msg_data[1024];	// データ
};

// メッセージ送信終了コマンド構造体
struct MessageEndTable {
	char			result;	// 0:正常終了、－1:送信失敗
	unsigned char	node;	// ノード番号
	unsigned short	tcd;	// 送信したＴＣＤ
};

// 通信モード取得 構造体
struct GetModeInfo {
	char			result;  	// 結果コード
	unsigned char	reserve[3];	// 未使用
	unsigned int	speed;		// 0:10Mbps 1:100Mbps
	unsigned int	duplex;		// 0:half 1:full
};

////////////////////////////////////////////
// Libraly 関数 グローバル変数
////////////////////////////////////////////
#define RCV_MAX		8
#define END_MAX		8
#define MU_MAX		3
#define BOARD_MAX	2
#define DEV_MAX		2

#define MUTEX_AP_CMD	0
#define MUTEX_FL_CMD	1
#define MUTEX_CYC		2

char *dev_name[DEV_MAX] = {"/dev/flnet0", "/dev/flnet1"};
pthread_mutex_t mutex[DEV_MAX][MU_MAX];	/*	MUTEX	*/

// グローバル変数 : FL-net PCIカードのノードＮｏ．(任意の値を指定可能）
unsigned short dev_opn[DEV_MAX] = {0,0};

// グローバル変数 : サイクリックデータバッファ（ライト用）
unsigned short com_buf1[BOARD_MAX][512];
unsigned short com_buf2[BOARD_MAX][512 * 16];

// グローバル変数 : メッセージ送信終了 管理領域
struct MessageEndTable msgend_tbl[BOARD_MAX][END_MAX];
struct MessageCmdTable msgrcv_tbl[BOARD_MAX][RCV_MAX];

/* グローバル変数 : FL-net PCIカードのオブジェクトハンドル */
int hDevice[DEV_MAX] = {-1,-1};

unsigned int dev_speed[BOARD_MAX] = {0,0};	/* Etherner speed : 10Mbps default  06/03/26 */

struct dev_info {
unsigned char	*pnpci;
unsigned short	my_node;
unsigned short	c1_adr;
unsigned short	c1_siz;
unsigned short	c2_adr;
unsigned short	c2_siz;
unsigned short	sum;
} dev_tbl[DEV_MAX] = {{0,0,0,0,0,0,0},
                      {0,0,0,0,0,0,0}};

struct MessageCtl {
unsigned short	rcv_cnt;	// 受信数
unsigned short	rcv_rxp;	// 受信セット 位置
unsigned short	rcv_txp;	// 受信通知   位置
unsigned short	end_cnt;	// 終了数
unsigned short	end_rxp;	// 終了セット 位置
unsigned short	end_txp;	// 終了通知   位置
} msg_ctl[BOARD_MAX] = {{0,0,0,0,0,0},
                        {0,0,0,0,0,0}};


// ローカル関数
static unsigned short EndRspSave(unsigned short);
static unsigned short RcvCmdSave(unsigned short);
static unsigned short CalSharedSum(struct dev_info *ptr, unsigned short num);


//
// FL_Initialize : FL-net PCIカードを初期化する。
//
// RETURN CODE
//   0 : 初期化成功
//  -1 : 初期化失敗
//
unsigned short FL_Initialize(struct InitInfo *init_p)
{
	clock_t				s_time, c_time;
	clock_t				duration;	/* 05.09.16 */
	unsigned short		dev_no;
	struct MessageCtl	*mc_p;
	unsigned int		sum, i;
	unsigned int		*l_ptr;
	unsigned short		*d_p;		/* 100Mbps 06/03/26 */
	unsigned short		version;	/* 100Mbps 06/03/26 */
	unsigned int		com_dat;

	if (init_p->dev_no != 0 && init_p->dev_no != 1) {
		return(PRM_ERR);
	}
	
	if(dev_opn[init_p->dev_no] == 1){    /* 2000/7/20 */
		return(DEV_DUPL_OPN);
	}
	
	/* デバイス名の指定					*/
	/* ボード１の場合："/dev/flnet0"	*/
	/* ボード２の場合："/dev/flnet1"	*/
	if (init_p->dev_no == 0) {
		hDevice[init_p->dev_no] = open("/dev/flnet0", O_RDWR);
		if (hDevice[init_p->dev_no] < 0) {
			return(DEV_ERR1);
		}
	} else {
		hDevice[init_p->dev_no] = open("/dev/flnet1", O_RDWR);
		if (hDevice[init_p->dev_no] < 0) {
			return(DEV_ERR1);
		}
	}
	
	if (flock(hDevice[init_p->dev_no], LOCK_EX | LOCK_NB) == -1) {
		if (errno == EWOULDBLOCK) {
			close(hDevice[init_p->dev_no]);
			return(DEV_DUPL_OPN);
		}
	}
	
	// 初期化
	memset(com_buf1, 0, sizeof(com_buf1));
	memset(com_buf2, 0, sizeof(com_buf2));
	memset(msgend_tbl, 0, sizeof(msgend_tbl));
	memset(msgrcv_tbl, 0, sizeof(msgrcv_tbl));
	
	dev_no = init_p->dev_no;
	dev_opn[dev_no] = 1;		// 05.10.06
	
	pnpci = mmap(0, PnSize, PROT_WRITE | PROT_READ, MAP_SHARED, hDevice[dev_no], 0);
	if(pnpci == 0) {
		DeviceClose(dev_no, 0);
		return(DEV_ERR1);
	}
	
	dev_tbl[dev_no].pnpci = pnpci;	// 05.10.06

	//-------------------------------
	// FL-net PCIカードを初期化する。
	//-------------------------------
	// (1) Reset ON
	com_dat = (*pnpci & 0x00003000) | 0x0000000f;
	*pnpci = com_dat;

	// (2) 10msec待つ。
	usleep(10 * 1000);

	// (3) デュアルポートRAM(20000h～2fffAh)をゼロクリアする。
	memset(pnpci + 0x20000, 0, 0xfffa);

	// (4) Reset解除
	*pnpci = com_dat | 0x00000080;

	// (5) FL-net PCIカードがReadyになるまで待つ。
	s_time = clock();
	while (*(unsigned short *)FLtoAP_CMD != COM_ENABLE) {
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;	/* 05.09.16 */
		if (duration > 6) {		/* 6 秒  */
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/* DeviceClose(dev_no, 1);  重大エラー取得のためクローズしない 05.07.27 */
				return(FATAL_ERR);
			} else {
				if (*(unsigned short *)FLtoAP_CMD == COM_ENABLE) {
					break;
				} else {
					DeviceClose(dev_no, 1);
					return(TIM_OUT);
				}
			}
		}
	}
	/* 100Mbps 06/03/26 */
	version = *(unsigned short *)FLtoAP_DATA & 0xff00;

	// (6) FLtoAP_CMD領域をゼロクリアする。
	*(unsigned short *)FLtoAP_CMD = 0x00;

	// (7)「コマンド受け付け可能」コマンドに対するレスポンスを
	//     FL-net PCIカードに返す。
	*(unsigned short *)APtoFL_RSP = RSP_ENABLE;

	//-----------------------------------------------
	// 初期化コマンドを発行する。
	//-----------------------------------------------
	// 初期化パラメータをAPtoFL_DATA領域にセットする。
	struct PciCardInitData *ptr;
	ptr = (struct PciCardInitData *)APtoFL_DATA;
	ptr->mode   = dev_speed[dev_no];  /* 100Mbps support 06/03/26 */
	ptr->ip_adr = init_p->ip_adr;
	ptr->c1_adr = init_p->c1_adr;
	ptr->c1_siz = init_p->c1_size;
	ptr->c2_adr = init_p->c2_adr;
	ptr->c2_siz = init_p->c2_size;
	ptr->tw     = init_p->tw;
	ptr->mft    = init_p->mft;
	memcpy(ptr->vender, init_p->vender, 10);
	memcpy(ptr->maker, init_p->maker, 10);
	memcpy(ptr->node, init_p->name, 10);

	// SUM値をセットする
	l_ptr = (unsigned int *)APtoFL_DATA;
	sum = 0;
	for (i = (sizeof(struct PciCardInitData) - 8) / 4; i > 0; i--) {
		sum += *l_ptr++;
	}
	ptr->sum_flag = 0x12345678;
	ptr->sum      = sum;

	// 初期化コマンドを発行する。
	*(unsigned short *)FLtoAP_RSP = 0;
	*(unsigned short *)APtoFL_CMD = REQ_INIT;

	// FL-net PCIカードからのレスポンスを待つ。
	s_time = clock();
	while (*(unsigned short *)FLtoAP_RSP != RSP_INIT) {
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;	/* 05.09.16 */
		if (duration > 10) {
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/* DeviceClose(dev_no, 1);   重大エラー取得のためクローズしない 05.07.27 */
				return(FATAL_ERR);
			} else {
				if (*(unsigned short *)FLtoAP_RSP == RSP_INIT) {
					break;
				} else {
					DeviceClose(dev_no, 1);
					return(TIM_OUT);
				}
			}
		}
	}
	// レスポンスのデータ部をチェックする。
	if (*(unsigned short *)FLtoAP_DATA != 0) {
		DeviceClose(dev_no, 1);
		return(PRM_ERR);
	}

	switch (dev_no) {
	case 0:
		// グローバル変数 my_node をセットする
		dev_tbl[0].my_node = (unsigned char)(ptr->ip_adr & 0x000000ff);

		dev_tbl[0].c1_adr = init_p->c1_adr;
		dev_tbl[0].c1_siz = init_p->c1_size;
		dev_tbl[0].c2_adr = init_p->c2_adr;
		dev_tbl[0].c2_siz = init_p->c2_size;

		// メッセージ 管理領域をクリアする。
		mc_p = &msg_ctl[0];
		mc_p->rcv_cnt = 0;	// 受信数
		mc_p->rcv_rxp = 0;	// 受信セット 位置
		mc_p->rcv_txp = 0;	// 受信通知   位置
		mc_p->end_cnt = 0;	// 終了数
		mc_p->end_rxp = 0;	// 終了セット 位置
		mc_p->end_txp = 0;	// 終了通知   位置
		
		/*	MUTEX初期化	*/
		pthread_mutex_init(&mutex[0][MUTEX_AP_CMD], NULL);
		pthread_mutex_init(&mutex[0][MUTEX_FL_CMD], NULL);
		pthread_mutex_init(&mutex[0][MUTEX_CYC], NULL);
		
		break;
	case 1:
		// グローバル変数 my_node をセットする
		dev_tbl[1].my_node = (unsigned char)(ptr->ip_adr & 0x000000ff);

		dev_tbl[1].c1_adr = init_p->c1_adr;
		dev_tbl[1].c1_siz = init_p->c1_size;
		dev_tbl[1].c2_adr = init_p->c2_adr;
		dev_tbl[1].c2_siz = init_p->c2_size;

		// メッセージ 管理領域をクリアする。
		mc_p = &msg_ctl[1];
		mc_p->rcv_cnt = 0;	// 受信数
		mc_p->rcv_rxp = 0;	// 受信セット 位置
		mc_p->rcv_txp = 0;	// 受信通知   位置
		mc_p->end_cnt = 0;	// 終了数
		mc_p->end_rxp = 0;	// 終了セット 位置
		mc_p->end_txp = 0;	// 終了通知   位置
		
		/*	MUTEX初期化	*/
		pthread_mutex_init(&mutex[1][MUTEX_AP_CMD], NULL);
		pthread_mutex_init(&mutex[1][MUTEX_FL_CMD], NULL);
		pthread_mutex_init(&mutex[1][MUTEX_CYC], NULL);
		
		break;
	}
	/* 06/03/26 */
	d_p = &com_buf1[dev_no][0];
	for (i = 512; i > 0; i--) {
		*d_p++ = 0;
	}
	d_p = &com_buf2[dev_no][0];
	for (i = 512 * 16; i > 0; i--) {
		*d_p++ = 0;
	}

	// FLtoAP_RSP領域をゼロクリアする。
	*(unsigned short *)FLtoAP_RSP = 0;
	dev_tbl[dev_no].sum = CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1);

	return(0);
}

//
// Close : FL-net PCIカードをクローズする。
//
unsigned short DeviceClose(unsigned short dev_no, unsigned short reset_f)
{
	int				ret;
	unsigned int	com_dat;
	
	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}
	
	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	pnpci = dev_tbl[dev_no].pnpci;
	
	// FL-net PCIカードをリセットする。(FL-net停止)
	if (reset_f == 1) {
		if (pnpci != 0) {
			com_dat = (*pnpci & 0x00003000) | 0x0000000f;
			*pnpci = com_dat;
		}
	}
	
	if (pnpci != 0) {
		munmap(pnpci, PnSize);
	}
	
	flock(hDevice[dev_no], LOCK_UN);
	
	ret = close(hDevice[dev_no]);
	if (ret != 0) {
		return(DEV_ERR5);
	}
	
	dev_opn[dev_no] = 0;
	
	return(0);
}

//
// WriteCyclicData All
//
unsigned short WriteCyclicALL(unsigned short *buf1, unsigned short *buf2, unsigned short dev_no)
{
	struct CyclicControlTable	*ptr;
	char						ap_use;
	unsigned short				*write_addr1, *write_addr2;
	unsigned short				size1, size2, i, *d_p;
	int							ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}
	
	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_CYC]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}

	// サイクリックコントロールテーブルを得る。
	ptr = (struct CyclicControlTable *)CYC_CTL_TBL;
	ptr += (dev_tbl[dev_no].my_node - 1);

	// コモン領域１、２のサイズが０か？
	if (ptr->cmn1_size == 0 && ptr->cmn2_size == 0) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	}

	size1 = ptr->cmn1_size;
	size2 = ptr->cmn2_size;

	// バッファ１、２のどちらに書き込むかを決定する。
	if ((ptr->last_buf & 0x0f) == 1) {
		if((ptr->last_buf & 0x10) == 0 ) {
			ap_use = 2;
		} else {
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(BUF_FUL);
		}
	} else {
		if((ptr->last_buf & 0x20) == 0 ){
			ap_use = 1;
		} else {
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(BUF_FUL);
		}
	}

	// ap_useのバッファへ、サイクリックデータを書く。
	if (ap_use == 1) {
		write_addr1 = (unsigned short *)(CMNBUF1_COM1 + (ptr->cmn1_addr * 2));
		write_addr2 = (unsigned short *)(CMNBUF1_COM2 + (ptr->cmn2_addr * 2));
	} else {
		write_addr1 = (unsigned short *)(CMNBUF2_COM1 + (ptr->cmn1_addr * 2));
		write_addr2 = (unsigned short *)(CMNBUF2_COM2 + (ptr->cmn2_addr * 2));
	}

	d_p = &com_buf1[dev_no][0];
	for (i = 0; i < size1; i++) {
		*write_addr1++ = *buf1;
		*d_p++         = *buf1++;
	}
	d_p = &com_buf2[dev_no][0];
	for (i = 0; i < size2; i++) {
		*write_addr2++ = *buf2;
		*d_p++         = *buf2++;
	}

	// 書き込みが終了したことをPCI Cardに通知する。
	while (1) {
		if (ptr->if_use == 0) {
			break;
		}
		usleep(1 * 1000);
	}
	if (ap_use == 1) {
		ptr->last_buf = (ptr->last_buf & 0xf0) | 0x11;
	} else {
		ptr->last_buf = (ptr->last_buf & 0xf0) | 0x22;
	}
	ptr->ap_use = 0;

	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	return(0);
}

//
// WriteCyclicData Relative
//
unsigned short WriteCyclicRelative(unsigned short common, unsigned short offset,
								   unsigned short size,unsigned short *buff,
								   unsigned short dev_no)
{
	struct CyclicControlTable	*ptr;
	char						ap_use;
	unsigned short				cmn_addr, cmn_size;
	unsigned short				*write_addr1, *write_addr2;
	unsigned short				size1, size2, i, *d_p;
	int							ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_CYC]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	// サイクリックコントロールテーブルを得る。
	ptr = (struct CyclicControlTable *)CYC_CTL_TBL;
	ptr += (dev_tbl[dev_no].my_node - 1);

	// コモン領域１、２のサイズが０か？
	if (ptr->cmn1_size == 0 && ptr->cmn2_size == 0) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	}

	if (common != 1 && common != 2) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	}
	if (common == 1) {
		if ((offset + size) > dev_tbl[dev_no].c1_siz) {		/* 07.12.03 */
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(PRM_ERR);
		}
	} else {
		if ((offset + size) > dev_tbl[dev_no].c2_siz) {		/* 07.12.03 */
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(PRM_ERR);
		}
	}

	if (common == 1) {
		cmn_addr = ptr->cmn1_addr;
		cmn_size = ptr->cmn1_size;
		d_p = &com_buf1[dev_no][offset];
	} else {
		cmn_addr = ptr->cmn2_addr;
		cmn_size = ptr->cmn2_size;
		d_p = &com_buf2[dev_no][offset];
	}
	// ライト開始アドレスが領域外
	if (offset > (cmn_size - 1)) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	}

	for (i = 0; i < size; i++) {
		*d_p++ = *buff++;
		if ((offset + i) > (cmn_addr + cmn_size - 1)) {
			break;
		}
	}

	size1 = ptr->cmn1_size;
	size2 = ptr->cmn2_size;

	// バッファ１、２のどちらに書き込むかを決定する。
	// 古いバッファを選択する。
	if ((ptr->last_buf & 0x0f) == 1) {
		if((ptr->last_buf & 0x10) == 0) {
			ap_use = 2;
		} else {
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(BUF_FUL);
		}
	} else {
		if ((ptr->last_buf & 0x20) == 0) {
			ap_use = 1;
		} else {
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(BUF_FUL);
		}
	}
	
	// ap_useのバッファへ、サイクリックデータを書く。
	if (ap_use == 1) {
		write_addr1 = (unsigned short *)(CMNBUF1_COM1 + (ptr->cmn1_addr * 2));
		write_addr2 = (unsigned short *)(CMNBUF1_COM2 + (ptr->cmn2_addr * 2));
	} else {
		write_addr1 = (unsigned short *)(CMNBUF2_COM1 + (ptr->cmn1_addr * 2));
		write_addr2 = (unsigned short *)(CMNBUF2_COM2 + (ptr->cmn2_addr * 2));
	}

	d_p = &com_buf1[dev_no][0];
	for (i = 0; i < size1; i++) {
		*write_addr1++ = *d_p++;
	}
	d_p = &com_buf2[dev_no][0];
	for (i = 0; i < size2; i++) {
		*write_addr2++ = *d_p++;
	}

	// 書き込みが終了したことをPCI Cardに通知する。
	while (1) {
		if (ptr->if_use == 0) {
			break;
		}
		usleep(1 * 1000);
	}
	if (ap_use == 1) {
		ptr->last_buf = (ptr->last_buf & 0xf0) | 0x11;
	} else {
		ptr->last_buf = (ptr->last_buf & 0xf0) | 0x22;
	}
	ptr->ap_use = 0;

	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	return(0);
}

//
// ReadCyclicDataRelative
//
unsigned short ReadCyclicRelative(unsigned short node, unsigned short common,
								  unsigned short offset, unsigned short size,
								  unsigned short *buff, unsigned short dev_no)
{
	struct CyclicControlTable	*ptr;
	unsigned short				cmn_addr, cmn_size, set_p;
	unsigned short				*read_addr;
	char						ap_use;
	int							ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0){
		return(DEV_NOT_OPEN);
	}
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_CYC]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}

	if (node == 0 || node == 255) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	}

	// サイクリックコントロールテーブルを得る。
	ptr = (struct CyclicControlTable *)CYC_CTL_TBL;
	ptr += node - 1;

	// ネットワーク参加・離脱フラグをチェックする。
	if (ptr->sanka_flg == 0) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(OUT_RING);
	}
	// last_bufをチェックする。
	if (ptr->last_buf == 0) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(NO_DATA);
	}

	if (common != 1 && common != 2) {	/* 07.12.03 */
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	}
	if (common == 1) {
		if ((offset + size) > ptr->cmn1_size) {		/* 07.12.03 */
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(PRM_ERR);
		}
	} else {
		if ((offset + size) > ptr->cmn2_size) {		/* 07.12.03 */
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(PRM_ERR);
		}
	}

	if (common == 1) {
		cmn_addr = ptr->cmn1_addr;
		cmn_size = ptr->cmn1_size;
	} else {
		cmn_addr = ptr->cmn2_addr;
		cmn_size = ptr->cmn2_size;
	}

	if (common == 1 && size > 512) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	} else if (common == 2 && size > 512 * 16) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	} else if (size <= 0) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	} else {
	}
	
	// リード開始アドレスが領域外
	if ((cmn_addr + offset) > (cmn_addr + cmn_size - 1)) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	}

	// バッファ１、２のどちらを読み込むかを決定する。
	while (1) {		/* 04/05/25 */
		ptr->ap_use = (volatile unsigned char)ptr->last_buf & 0x0f;
		if ((volatile unsigned char)ptr->if_use == (volatile unsigned char)ptr->ap_use) {
			ptr->ap_use = 0;
			usleep(1 * 1000);
		} else {
			break;
		}
	}
	ap_use = ptr->ap_use;

	// ap_useのバッファから、サイクリックデータを読み出す。
	if (ap_use == 1) {
		if (common == 1) {
			read_addr = (unsigned short *)(CMNBUF1_COM1 + (cmn_addr + offset) * 2);
		} else {
			read_addr = (unsigned short *)(CMNBUF1_COM2 + (cmn_addr + offset) * 2);
		}
	} else {
		if (common == 1) {
			read_addr = (unsigned short *)(CMNBUF2_COM1 + (cmn_addr + offset) * 2);
		} else {
			read_addr = (unsigned short *)(CMNBUF2_COM2 + (cmn_addr + offset) * 2);
		}
	}

	set_p = cmn_addr + offset;
	for (; size > 0; size--) {
		//dat = *read_addr++;
		//*buff++ = dat;
		*buff++ = *read_addr++;
		set_p++;
		if (set_p > (cmn_addr + cmn_size - 1)) {
			break;
		}
	}

	// バッファ使用中フラグをクリアする。
	ptr->ap_use = 0;
    
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	return(0);
}

//
// ReadCyclicDataAbsolute
//
unsigned short ReadCyclicAbsolute(unsigned short common, unsigned short offset,
								  unsigned short size, unsigned short *buff,
								  unsigned short dev_no)
{
	short						node, hit_flag;
	unsigned short				*buf_p, i, ed_p, cmn_size, cmn_addr;
	struct CyclicControlTable	*ptr;
	unsigned short				*read_addr;
	char						ap_use;
	int							ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	// 引数チェック	07.12.03
	if (common != 1 && common != 2) {
		return(PRM_ERR);
	}
	
	if (common == 1) {
		if ((offset + size) > 512) {
			return(PRM_ERR);
		}
	} else {
		if ((offset + size) > 8192) {
			return(PRM_ERR);
		}
	}

	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_CYC]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	// 出力用バッファの指定領域分を０クリアする。
	buf_p = buff;
	for (i = 0; i < size; i++) {
		*buf_p++ = 0;
	}

	// 指定された領域に該当するノードをサーチする。
	ed_p = offset + size - 1;		/*	04/02/03	*/
	ptr = (struct CyclicControlTable *)CYC_CTL_TBL;
	for (node = 1; node <= 254; node++, ptr++) {
		// このノードは、ネットワークに参加しているか？
		if (ptr->sanka_flg == 0) {	// 離脱している
			continue;
		}
		
		// コモン領域アドレス、コモン領域サイズを得る。
		if (common == 1) {
			cmn_addr = ptr->cmn1_addr;
			cmn_size = ptr->cmn1_size;
		} else {
			cmn_addr = ptr->cmn2_addr;
			cmn_size = ptr->cmn2_size;
		}
		
		// このノードのコモン領域アドレスおよびサイズは、指定した絶対アドレス
		// およびサイズの範囲内にあるか？
		hit_flag = 0;
		if (cmn_size > 0) {
			if (cmn_addr >= offset && cmn_addr <= ed_p) {
				hit_flag = 1;
			}
			if ((cmn_addr + cmn_size) >= offset && (cmn_addr + cmn_size) <= ed_p) {
				hit_flag = 1;
			}
			if (cmn_addr <= offset && (cmn_addr + cmn_size) >= ed_p) {
				hit_flag = 1;
			}
		}
		if (hit_flag == 0) {
			continue;
		}
		
		// last_bufをチェックする。
		if (ptr->last_buf == 0) {
			continue;
		}
		
		// バッファ１、２のどちらを読み込むかを決定する。
		while (1) {		/*	04/05/25	*/
			ptr->ap_use = (volatile unsigned char)ptr->last_buf & 0x0f;
			if ((volatile unsigned char)ptr->if_use == (volatile unsigned char)ptr->ap_use) {
				ptr->ap_use = 0;
				usleep(1 * 1000);
			} else {
				break;
			}
		}
		ap_use = ptr->ap_use;

		// ap_useのバッファから、サイクリックデータを読み出す。
		if (cmn_addr < offset) {
			cmn_size = (cmn_addr + cmn_size) - offset;
			cmn_addr = offset;
		}
		
		if (ap_use == 1) {
			if (common == 1) {
				read_addr = (unsigned short *)(CMNBUF1_COM1 + cmn_addr * 2);
			} else {
				read_addr = (unsigned short *)(CMNBUF1_COM2 + cmn_addr * 2);
			}
		} else {
			if (common == 1) {
				read_addr = (unsigned short *)(CMNBUF2_COM1 + cmn_addr * 2);
			} else {
				read_addr = (unsigned short *)(CMNBUF2_COM2 + cmn_addr * 2);
			}
		}
		
		buf_p = buff + (cmn_addr - offset);
		for (; cmn_size > 0; cmn_size--) {
			*buf_p++ = *read_addr++;
			cmn_addr++;
			if (cmn_addr > ed_p) {
				break;
			}
		}
		// バッファ使用中フラグをクリアする。
		ptr->ap_use = 0;
	}
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	return(0);
}

//
// ReadCyclicData All
//
unsigned short ReadCyclicALL(unsigned short node, unsigned short *buf1,
							 unsigned short *buf2, unsigned short dev_no)
{
	struct CyclicControlTable	*ptr;
	char						ap_use;
	unsigned short				*read_addr1, *read_addr2;
	unsigned short				size1, size2, i;
	int							ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}

    if (node < 1 || node > 254) {		/* 07.12.03 */
		return(PRM_ERR);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_CYC]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	// サイクリックコントロールテーブルを得る。
	ptr = (struct CyclicControlTable *)CYC_CTL_TBL;
	ptr += node - 1;

	// ネットワーク参加・離脱フラグをチェックする。
	if (ptr->sanka_flg == 0) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(OUT_RING);
	}

	// last_bufをチェックする。
	if (ptr->last_buf == 0) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(NO_DATA);
	}

	// コモン領域１、２のサイズが０か？
	if (ptr->cmn1_size == 0 && ptr->cmn2_size == 0) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	}

	size1 = ptr->cmn1_size;
	size2 = ptr->cmn2_size;

	// バッファ１、２のどちらを読み込むかを決定する。
	while (1) {		/*	04/05/25	*/
		ptr->ap_use = (volatile unsigned char)ptr->last_buf & 0x0f;
		if ((volatile unsigned char)ptr->if_use == (volatile unsigned char)ptr->ap_use) {
			ptr->ap_use = 0;
			usleep(1 * 1000);
		} else {
			break;
		}
	}
	ap_use = ptr->ap_use;

	// ap_useのバッファの、サイクリックデータをRead
	if (ap_use == 1) {
		read_addr1 = (unsigned short *)(CMNBUF1_COM1 + (ptr->cmn1_addr * 2));
		read_addr2 = (unsigned short *)(CMNBUF1_COM2 + (ptr->cmn2_addr * 2));
	} else {
		read_addr1 = (unsigned short *)(CMNBUF2_COM1 + (ptr->cmn1_addr * 2));
		read_addr2 = (unsigned short *)(CMNBUF2_COM2 + (ptr->cmn2_addr * 2));
	}

	for (i = 0; i < size1; i++) {
		*buf1++ = *read_addr1++;
	}
	for (i = 0; i < size2; i++) {
		*buf2++ = *read_addr2++;
	}

	// バッファ使用中フラグをクリアする。
	ptr->ap_use = 0;
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_CYC]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	return(0);
}

//
// 参加ノードＮＯ．表示
//
unsigned short GetRingInfo(struct SankaReqTable *ring_infp, unsigned short dev_no)
{
	struct SankaReqTable	*ptr1;
	unsigned long			i;
	clock_t					s_time, c_time;
	clock_t					duration;		/*	05.09.16	*/
	int						ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	// 参加ノード情報要求コマンドをセット。
	*(unsigned short *)FLtoAP_RSP = 0;
	*(unsigned short *)APtoFL_CMD = REQ_SANKA_NODE;

	// FL-net PCIカードからのレスポンスを待つ。
	s_time = clock();
	while (*(volatile unsigned short *)FLtoAP_RSP != RSP_SANKA_NODE) {	/* add volatile  05.09.16 */
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;	/*	05.09.16	*/
		// メッセージ送信終了レスポンスをチェックする。
		if (duration > 1) {		/* 1 秒  */
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(FATAL_ERR);
			} else {
				if (*(volatile unsigned short *)FLtoAP_RSP == RSP_SANKA_NODE) {	/* add volatile  05.09.16 */
					break;
				} else {
					*(unsigned short *)APtoFL_CMD = 0;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
		}
	}

	ptr1 = (struct SankaReqTable *)FLtoAP_DATA;

	// 参加ノードＮｏ．情報をコピーする
	ring_infp->node_cnt = ptr1->node_cnt;

	for (i = 0; i < ptr1->node_cnt; i++) {
		ring_infp->node_no[i] = ptr1->node_no[i];
	}

	// FLtoAP_RSP領域をゼロクリアする。
	*(unsigned short *)FLtoAP_RSP = 0;
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	return(0);
}

//
// ノード情報を得る
//
unsigned short GetNodeInfo(unsigned short node, struct NodeInfo *node_infp, unsigned short dev_no)
{
	struct NodeInfoTable	*ptr1;
	unsigned long			i;
	clock_t					s_time, c_time;
	clock_t					duration;	/*	05.09.16	*/
	int						ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}

	// 自ノードの参加ノード情報要求コマンドをセット。
	*(unsigned short *)FLtoAP_RSP = 0;
	*(unsigned char *)APtoFL_DATA = (unsigned char)node;
	*(unsigned short *)APtoFL_CMD = REQ_NODE_INF;

	// FL-net PCIカードからのレスポンスを待つ。
	s_time = clock();
	while (*(volatile unsigned short *)FLtoAP_RSP != RSP_NODE_INF) {	/* add volatile  05.09.16 */
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;		/*	05.09.16	*/
		// メッセージ送信終了レスポンスをチェックする。
		if (duration > 1) {		/* 1 秒  */
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(FATAL_ERR);
			} else {
				if (*(volatile unsigned short *)FLtoAP_RSP == RSP_NODE_INF) {	/* add volatile  05.09.16 */
					break;
				} else {
					*(unsigned short *)APtoFL_CMD = 0;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
		}
	}

	ptr1 = (struct NodeInfoTable *)FLtoAP_DATA;
	if (ptr1->result == -1) {	// 結果コード
		*(unsigned short *)FLtoAP_RSP = 0;
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	}
	
	if (ptr1->sanka == 1 || dev_tbl[dev_no].my_node == node) {
		node_infp->sanka		= ptr1->sanka;		// 参加／離脱 フラグ
		node_infp->uls			= ptr1->uls;		// 上位層ステータス
		node_infp->cmn1_addr	= ptr1->cmn1_addr;	// コモン領域１アドレス
		node_infp->cmn1_size	= ptr1->cmn1_size;	// コモン領域１サイズ
		node_infp->cmn2_addr	= ptr1->cmn2_addr;	// コモン領域２アドレス
		node_infp->cmn2_size	= ptr1->cmn2_size;	// コモン領域２サイズ
		node_infp->rct			= ptr1->rct;		// リフレッシュサイクル許容時間
		node_infp->tw			= ptr1->tw;			// トークン監視時間
		node_infp->mft			= ptr1->mft;		// 最小フレーム間隔
		node_infp->link_status	= ptr1->link_status;// リンクの状態
		node_infp->my_status	= ptr1->my_status;	// 自ノードの状態
		for (i = 0; i < 10; i++) {
			node_infp->vender[i]	= ptr1->vender[i];	// ベンダー名
			node_infp->maker[i]		= ptr1->maker[i];	// メーカー名
			node_infp->name[i]		= ptr1->name[i];	// ノード名
		}
		node_infp->vender[10]	= 0;
		node_infp->maker[10]	= 0;
		node_infp->name[10]		= 0;
	} else {	/*	07.12.03	*/
		node_infp->sanka		= 0;
		node_infp->uls			= 0;
		node_infp->cmn1_addr	= 0;
		node_infp->cmn1_size	= 0;
		node_infp->cmn2_addr	= 0;
		node_infp->cmn2_size	= 0;
		node_infp->rct			= 0;
		node_infp->tw			= 0;
		node_infp->mft			= 0;
		node_infp->link_status	= 0;
		node_infp->my_status	= 0;
		node_infp->vender[0]	= 0;
		node_infp->maker[0]		= 0;
		node_infp->name[0]		= 0;
	}

	// FLtoAP_RSP領域をゼロクリアする。
	*(unsigned short *)FLtoAP_RSP = 0;
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	return(0);
}

//
// ネットワーク情報を得る
//
unsigned short GetNetInfo(struct NetInfo *net_infp, unsigned short dev_no)
{
	struct NetInfo	*ptr1;
	clock_t			s_time, c_time;
	clock_t			duration;		/*	05.09.16	*/
	int				ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}

	// ネットワーク情報要求コマンドをセット。
	*(unsigned short *)FLtoAP_RSP = 0;
	*(unsigned short *)APtoFL_CMD = REQ_NET_INF;  /*!*/

	// FL-net PCIカードからのレスポンスを待つ。
	s_time = clock();
	while (*(volatile unsigned short *)FLtoAP_RSP != RSP_NET_INF) {	/* add volatile  05.09.16 */
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;		/*	05.09.16	*/
		// メッセージ送信終了レスポンスをチェックする。
		if (duration > 1) {		/* 1 秒  */
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(FATAL_ERR);
			} else {
				if (*(volatile unsigned short *)FLtoAP_RSP == RSP_NET_INF) {	/* add volatile  05.09.16 */
					break;
				} else {
					*(unsigned short *)APtoFL_CMD = 0;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
		}
	}

	ptr1 = (struct NetInfo *)FLtoAP_DATA;
	net_infp->rct = ptr1->rct;
	net_infp->rcm = ptr1->rcm;
	net_infp->rcm_max = ptr1->rcm_max;
	if (ptr1->rcm_min == 0xffff) {		/* 07.12.03 */
		net_infp->rcm_min = 0;
	} else {
		net_infp->rcm_min = ptr1->rcm_min;
	}
	net_infp->link_status = ptr1->link_status;
	*(unsigned short *)FLtoAP_RSP = 0;
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	return(0);
}

//
//
//
unsigned short GetLogInfo(unsigned int *bufp, unsigned short dev_no)
{
	unsigned long	i;
	unsigned int	*ptr;
	clock_t			s_time, c_time;
	clock_t			duration;		/*	05.09.16	*/
	int				ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}

	// ログ情報要求コマンドをセット。
	*(unsigned short *)FLtoAP_RSP = 0;
	*(unsigned short *)APtoFL_CMD = REQ_LOG;

	// FL-net PCIカードからのレスポンスを待つ。
	s_time = clock();
	while (*(volatile unsigned short *)FLtoAP_RSP != RSP_LOG) {	/* add volatile  05.09.16 */
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;		/*	05.09.16	*/
		// メッセージ送信終了レスポンスをチェックする。
		if (duration > 1) {		/* 1 秒  */
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(FATAL_ERR);
			} else {
				if (*(volatile unsigned short *)FLtoAP_RSP == RSP_LOG) {	/* add volatile  05.09.16 */
					break;
				} else {
					*(unsigned short *)APtoFL_CMD = 0;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
		}
	}

	ptr = (unsigned int *)FLtoAP_DATA;
	for (i = 0; i < 128; i++) {
		*bufp++ = *ptr++;
	}
	*(unsigned short *)FLtoAP_RSP = 0;
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	return(0);
}

//
//
//
unsigned short SetUplStatus(unsigned short status, unsigned short dev_no)
{
	clock_t	s_time, c_time;
	clock_t	duration;		/*	05.09.16	*/
	int		ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}

	// 上位層ステータス指定 要求コマンドをセット。
	*(unsigned short *)FLtoAP_RSP = 0;
	*(unsigned short *)APtoFL_DATA = status;
	*(unsigned short *)APtoFL_CMD = REQ_SET_ULS;

	// FL-net PCIカードからのレスポンスを待つ。
	s_time = clock();
	while (*(volatile unsigned short *)FLtoAP_RSP != RSP_SET_ULS) {	/* add volatile  05.09.16 */
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;		/*	05.09.16	*/
		// メッセージ送信終了レスポンスをチェックする。
		if (duration > 1) {		/* 1 秒  */
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(FATAL_ERR);
			} else {
				if (*(volatile unsigned short *)FLtoAP_RSP == RSP_SET_ULS) {	/* add volatile  05.09.16 */
					break;
				} else {
					*(unsigned short *)APtoFL_CMD = 0;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
		}
	}
	
	*(unsigned short *)FLtoAP_RSP = 0;
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	return(0);
}

//
// 重大エラー情報を得る
//
unsigned short GetErrInfo(struct ErrInfo *err_infp, unsigned short dev_no)
{
	struct ErrInfo	*ptr;
	unsigned short	i;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}
	
	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	ptr = (struct ErrInfo *)FLtoAP_CBUF;
	if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
		err_infp->err_code = ptr->err_code;
		err_infp->length = ptr->length;
		for (i = 0; i < ptr->length; i++) {
			err_infp->err_msg[i] = ptr->err_msg[i];
		}
		err_infp->err_msg[i] = 0;
	} else {
		err_infp->err_code = 0;
		err_infp->length = 0;
		err_infp->err_msg[0] = 0;
	}
	
	return(0);
}

//
//
//
unsigned short WriteMessage(struct MessageInfo *msg_prm, unsigned char *msg_buf,
							unsigned short wait_flg, unsigned short dev_no)
{
	struct CyclicControlTable	*ptr;
	struct MessageCmdTable		*ptr1;
	struct MessageEndTable		*ptr2;
	unsigned short				i, rc;
	short						result;
	clock_t						s_time, c_time;
	clock_t						duration;		/*	05.09.16	*/
	int							ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}
	
	if (wait_flg != 0 && wait_flg != 1) {
		return(PRM_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}

	// もし、デュアルポート上に通知があった場合は取り込む  07.04.27
	if (*(unsigned short *)FLtoAP_CMD > 0) {
		switch (*(unsigned short *)FLtoAP_CMD) {
		// メッセージ受信通知
		case MSG_RCV:
			RcvCmdSave(dev_no);
			break;
		// メッセージ送信終了
		case MSG_SND_END:
			EndRspSave(dev_no);
			break;
			// その他のコマンド(通常は、重大エラー通知）
		default:
			break;
		}
	}

	// サイクリックコントロールテーブルを得る。
	ptr = (struct CyclicControlTable *)CYC_CTL_TBL;
	ptr += (dev_tbl[dev_no].my_node - 1);
	if (ptr->sanka_flg == 0) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(OUT_RING);
	}
	if (*(unsigned short *)APtoFL_CMD != 0) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(BUF_FUL);
	}
	if (msg_prm->msg_node == 0 || msg_prm->msg_bc > 1024 || msg_prm->msg_result > 2) {
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	}

	ptr1 = (struct MessageCmdTable *)APtoFL_DATA;
	ptr1->msg_node		= msg_prm->msg_node;
	ptr1->msg_tcd		= msg_prm->msg_tcd;
	ptr1->msg_result	= msg_prm->msg_result;
	ptr1->msg_adr		= msg_prm->msg_adr;
	ptr1->msg_bc		= msg_prm->msg_bc;
	ptr1->msg_lng		= msg_prm->msg_lng;
	for (i = 0; i < msg_prm->msg_bc; i++) {
		ptr1->msg_data[i] = *msg_buf++;
	}
	*(unsigned short *)FLtoAP_RSP = 0;
	*(unsigned short *)APtoFL_CMD = REQ_MSG_SND;

	// FL-net PCIカードからのレスポンスを待つ。
	s_time = clock();
	while (*(volatile unsigned short *)FLtoAP_RSP != RSP_MSG_SND) {	/* add volatile  05.09.16 */
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;		/*	05.09.16	*/
		// メッセージ送信終了レスポンスをチェックする。
		if (duration > 2) {		/* 2 秒  */
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(FATAL_ERR);
			} else {
				if (*(volatile unsigned short *)FLtoAP_RSP == RSP_MSG_SND) {	/* add volatile  05.09.16 */
					break;
				} else {
					*(unsigned short *)APtoFL_CMD = 0;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
		}
	}
	
	if (*(char *)FLtoAP_DATA == -1) {
		*(unsigned short *)FLtoAP_RSP = 0;
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(PRM_ERR);
	} else if (*(char *)FLtoAP_DATA == -2) {
		*(unsigned short *)FLtoAP_RSP = 0;
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(BUF_FUL);
	} else {
	}
	
	*(unsigned short *)FLtoAP_RSP = 0;  /* 05.09.16 */
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	// メッセージ送信終了待ち処理
	if (wait_flg == 1) {
		/*	MUTEX取得	*/
		ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_FL_CMD]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		
		s_time = clock();
		duration = 0;
		while (1) {
			if (duration > 2) {		/* 2 秒  */
				if (*(volatile unsigned short *)FLtoAP_CMD == 0) {	/* 04/04/20 , add volatile  05.09.16 */
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
			if (*(volatile unsigned short *)FLtoAP_CMD > 0) {		/* add volatile  05.09.16 */
				switch (*(volatile unsigned short *)FLtoAP_CMD) {	/* add volatile  05.09.16 */
				// メッセージ受信通知
				case MSG_RCV:
					rc = RcvCmdSave(dev_no);
					if (rc != 0) {
						/*	MUTEX解放	*/
						ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
						if (ret != 0) {
							return(MUTEX_ERR);
						}
						return(FATAL_ERR);
					}
					s_time = clock();
					break;
				// メッセージ送信終了
				case MSG_SND_END:
					ptr2 = (struct MessageEndTable *)FLtoAP_CBUF;
					// 今送ったメッセージの応答か
					if (ptr2->node == (unsigned char)msg_prm->msg_node &&
						ptr2->tcd == (unsigned short)msg_prm->msg_tcd) {
						result = ptr2->result;
						*(unsigned short *)FLtoAP_CMD = 0; 
						*(unsigned short *)APtoFL_RSP = RSP_SND_END;
						if (result == 0) {
							/*	MUTEX解放	*/
							ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
							if (ret != 0) {
								return(MUTEX_ERR);
							}
							return(0);
						} else if(result == -2) {
							/*	MUTEX解放	*/
							ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
							if (ret != 0) {
								return(MUTEX_ERR);
							}
							return(BUF_FUL);
						} else {
							/*	MUTEX解放	*/
							ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
							if (ret != 0) {
								return(MUTEX_ERR);
							}
							return(TX_FAIL);
						}
					} else {
						rc = EndRspSave(dev_no);
						if (rc != 0) {
							/*	MUTEX解放	*/
							ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
							if (ret != 0) {
								return(MUTEX_ERR);
							}
							return(FATAL_ERR);
						}
						s_time = clock();
					}
					break;
				// その他のコマンド(通常は、重大エラー通知）
				default:
					*(unsigned short *)FLtoAP_CMD = 0;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(FATAL_ERR);
				}
			}
			c_time = clock();
			duration = (c_time - s_time) / CLOCKS_PER_SEC;
		}
	} else {
		return(0);
	}
}

//
//
//
static unsigned short EndRspSave(unsigned short dev_no)
{
	struct MessageEndTable	*ptr1, *ptr2;
	struct MessageCtl		*mc_p;

	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	mc_p = &msg_ctl[dev_no];
	if (mc_p->end_cnt < END_MAX) {
		ptr1 = (struct MessageEndTable *)FLtoAP_CBUF;
		ptr2 = &msgend_tbl[dev_no][mc_p->end_rxp];

		ptr2->node   = ptr1->node;
		ptr2->result = ptr1->result;
		ptr2->tcd    = ptr1->tcd;
		mc_p->end_cnt +=1;
		mc_p->end_rxp +=1;
		if (mc_p->end_rxp == END_MAX) {
			mc_p->end_rxp = 0;
		}
		*(unsigned short *)FLtoAP_CMD = 0;
		*(unsigned short *)APtoFL_RSP = RSP_SND_END;
	} else {	// 07.04.27
		*(unsigned short *)FLtoAP_CMD = 0;
		*(unsigned short *)APtoFL_RSP = RSP_SND_END;
	}
	
	return(0);
}

//
//
//
static unsigned short RcvCmdSave(unsigned short dev_no)
{
	struct MessageCmdTable	*ptr1, *ptr2;
	struct MessageCtl		*mc_p;
	unsigned long			i;
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}
	
	mc_p = &msg_ctl[dev_no];
	if (mc_p->rcv_cnt < RCV_MAX) {
		ptr1 = (struct MessageCmdTable *)FLtoAP_CBUF;
		ptr2 = &msgrcv_tbl[dev_no][mc_p->rcv_rxp];
		ptr2->msg_node   = ptr1->msg_node;
		ptr2->msg_result = ptr1->msg_result;
		ptr2->msg_tcd    = ptr1->msg_tcd;
		ptr2->msg_bc     = ptr1->msg_bc;
		ptr2->msg_lng    = ptr1->msg_lng;
		ptr2->msg_adr    = ptr1->msg_adr;
		
		for (i = 0; i < ptr2->msg_bc; i++){
			ptr2->msg_data[i] = ptr1->msg_data[i];
		}
		
		mc_p->rcv_cnt +=1;
		mc_p->rcv_rxp +=1;
		if (mc_p->rcv_rxp == RCV_MAX) {
			mc_p->rcv_rxp = 0;
		}
		
		*(unsigned short *)FLtoAP_CMD = 0;
		*(unsigned short *)APtoFL_RSP = RSP_RCV;
	} else {	// 07.04.27
		*(unsigned short *)FLtoAP_CMD = 0;
		*(unsigned short *)APtoFL_RSP = RSP_RCV;
	}
	
	return(0);
}

//
//
//
unsigned short CheckMessageEnd(unsigned short *node, unsigned short *tcd, unsigned short dev_no)
{
	struct MessageEndTable	*ptr;
	struct MessageCtl		*mc_p;
	unsigned short			cmd, result;
	int						ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}

	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_FL_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	mc_p = &msg_ctl[dev_no];
	// すでに取り込まれた結果がある。
	if (mc_p->end_cnt > 0) {
		ptr = &msgend_tbl[dev_no][mc_p->end_txp];
		*node = ptr->node;
		*tcd  = (unsigned short)ptr->tcd;
		mc_p->end_txp += 1;
		mc_p->end_cnt -= 1;
		if (mc_p->end_txp == END_MAX) {
			mc_p->end_txp = 0;
		}
		
		// もし、デュアルポート上に通知があった場合は取り込む   07.04.27
		if (*(unsigned short *)FLtoAP_CMD > 0) {
			switch (*(unsigned short *)FLtoAP_CMD) {
			// メッセージ受信通知
			case MSG_RCV:
				RcvCmdSave(dev_no);
				break;
			// メッセージ送信終了
			case MSG_SND_END:
				EndRspSave(dev_no);
				break;
			// その他のコマンド(通常は、重大エラー通知）
			default:
				break;
			}
		}
		
		if (ptr->result == 0) {
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(0);
		} else {
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(TX_FAIL);
		}
	} else {	// デュアルポートでメッセージ送信終了を確認する。
		cmd = *(unsigned short *)FLtoAP_CMD;
		switch (cmd) {
		case MSG_SND_END:
			ptr = (struct MessageEndTable *)FLtoAP_CBUF;
			*node = ptr->node;
			*tcd  = (unsigned short)ptr->tcd;
			result = ptr->result;
			*(unsigned short *)FLtoAP_CMD = 0;
			*(unsigned short *)APtoFL_RSP = RSP_SND_END;
			if (result == 0) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(0);
			} else {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(TX_FAIL);
			}
		case MSG_RCV:
			RcvCmdSave(dev_no);
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(TX_NOT_END);
		case REP_FATAL_ERR:
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(FATAL_ERR);
		default:
			/*	MUTEX解放	*/
			ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
			if (ret != 0) {
				return(MUTEX_ERR);
			}
			return(TX_NOT_END);
		}
	}
}

unsigned short ReadMessage(struct MessageInfo *rcv_msg, unsigned char *buff,
						   unsigned short wait_flg, unsigned short dev_no)
{
	struct MessageCmdTable	*ptr;
	struct MessageCtl		*mc_p;
	unsigned long			i;
	unsigned short			rc;
	clock_t					s_time, c_time;
	clock_t					duration;		/*	05.09.16	*/
	int						ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}

	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}
	
	if (wait_flg != 0 && wait_flg != 1) {
		return(PRM_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_FL_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	mc_p = &msg_ctl[dev_no];
	// すでに取り込まれた受信データがある。
	if (mc_p->rcv_cnt > 0) {
		ptr = &msgrcv_tbl[dev_no][mc_p->rcv_txp];
		rcv_msg->msg_node   = ptr->msg_node;
		rcv_msg->msg_result = ptr->msg_result;
		rcv_msg->msg_tcd    = ptr->msg_tcd;
		rcv_msg->msg_bc     = ptr->msg_bc;
		rcv_msg->msg_lng    = ptr->msg_lng;
		rcv_msg->msg_adr    = ptr->msg_adr;
		for (i = 0; i < ptr->msg_bc; i++) {
			*buff++ = ptr->msg_data[i];
		}
		mc_p->rcv_cnt -=1;
		mc_p->rcv_txp +=1;
		if (mc_p->rcv_txp == RCV_MAX) {
			mc_p->rcv_txp = 0;
		}
		
		// もし、デュアルポート上に通知があった場合は取り込む   07.04.27
		if (*(unsigned short *)FLtoAP_CMD > 0) {
			switch (*(unsigned short *)FLtoAP_CMD) {
			// メッセージ受信通知
			case MSG_RCV:
				RcvCmdSave(dev_no);
				break;
			// メッセージ送信終了
			case MSG_SND_END:
				EndRspSave(dev_no);
				break;
			// その他のコマンド(通常は、重大エラー通知）
			default:
				break;
			}
		}
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(0);
	} else {	// デュアルポートに受信メッセージがあるかを確認する。
		s_time = clock();
		duration = 0;
		while (1) {
			if (duration > 2) {		/* 2 秒  */
				if (*(volatile unsigned short *)FLtoAP_CMD == 0) {	/* 04/04/20 , add volatile  05.09.16 */
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
			if (*(volatile unsigned short *)FLtoAP_CMD > 0) {		/* add volatile  05.09.16 */
				switch (*(volatile unsigned short *)FLtoAP_CMD) {	/* add volatile  05.09.16 */
				// メッセージ受信通知
				case MSG_RCV:
					ptr = (struct MessageCmdTable *)FLtoAP_CBUF;
					rcv_msg->msg_node   = ptr->msg_node;
					rcv_msg->msg_result = ptr->msg_result;
					rcv_msg->msg_tcd    = ptr->msg_tcd;
					rcv_msg->msg_bc     = ptr->msg_bc;
					rcv_msg->msg_lng    = ptr->msg_lng;
					rcv_msg->msg_adr    = ptr->msg_adr;
					for (i = 0; i < ptr->msg_bc; i++) {
						*buff++ = ptr->msg_data[i];
					}
					*(unsigned short *)FLtoAP_CMD = 0;
					*(unsigned short *)APtoFL_RSP = RSP_RCV;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(0);
				// メッセージ送信終了
				case MSG_SND_END:
					rc = EndRspSave(dev_no);
					if (rc != 0) {
						/*	MUTEX解放	*/
						ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
						if (ret != 0) {
							return(MUTEX_ERR);
						}
						return(FATAL_ERR);
					}
					s_time = clock();
					break;
				// その他のコマンド(通常は、重大エラー通知）
				default:
					*(unsigned short *)FLtoAP_CMD = 0;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(FATAL_ERR);
				}
			}
			if (wait_flg == 0) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_FL_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(NO_DATA);
			}
			c_time = clock();
			duration = (c_time - s_time) / CLOCKS_PER_SEC;		/*	05.09.16	*/
		}
	}
}

unsigned short ReadUls_All(struct UlsInfo *uinf_p, unsigned short dev_no)
{
	unsigned char	*s_ptr, *d_ptr;
	unsigned long	i;
	clock_t			s_time, c_time;
	clock_t			duration;		/*	05.09.16	*/
	int				ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}

	// ログ情報要求コマンドをセット。
	*(unsigned short *)FLtoAP_RSP = 0;
	*(unsigned short *)APtoFL_CMD = REQ_ULS_LKS;

	// FL-net PCIカードからのレスポンスを待つ。
	s_time = clock();   
	while (*(volatile unsigned short *)FLtoAP_RSP != RSP_ULS_LKS) {
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;		/*	05.09.16	*/
		// メッセージ送信終了レスポンスをチェックする。
		if (duration > 2) {		/* 2 秒  */
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(FATAL_ERR);
			} else {
				if (*(volatile unsigned short *)FLtoAP_RSP == RSP_ULS_LKS) {
					break;
				} else {
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
		}
	}

	s_ptr = (unsigned char *)FLtoAP_DATA;
	d_ptr = (unsigned char *)uinf_p;
	for (i = 0; i < 762; i++) {
		*d_ptr++ = *s_ptr++;
	}
	*(unsigned short *)FLtoAP_RSP = 0;
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	return(0);
}

//
// Dll VerisonNoを得る
//
unsigned short Dll_Version(void)
{
	return(DLL_VERSION);
}

//
//
// 04/04/06
unsigned short Firmware_Version(unsigned short *version, unsigned short dev_no)
{
	clock_t	s_time, c_time;
	clock_t	duration;		/*	05.09.16	*/
	int		ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}

	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}

	// 上位層ステータス指定 要求コマンドをセット。
	*(unsigned short *)FLtoAP_RSP = 0;
	*(unsigned short *)APtoFL_CMD = REQ_REVISION;

	// FL-net PCIカードからのレスポンスを待つ。
	s_time = clock();   
	while (*(volatile unsigned short *)FLtoAP_RSP != RSP_REVISION) {
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;
		// メッセージ送信終了レスポンスをチェックする。
		if (duration > 1) {		/* 1 秒  */
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(FATAL_ERR);
			} else {
				if (*(volatile unsigned short *)FLtoAP_RSP == RSP_REVISION) {
					break;
				} else {
					*(unsigned short *)APtoFL_CMD = 0;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
		}
	}
	
	*version = *(unsigned short *)FLtoAP_DATA;
	*(unsigned short *)FLtoAP_RSP = 0;
	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	return(0);
}

//
// Set Ethernet Speed (0:10Mbps, 1:100Mbps, 2:AutoNegotiation)
// 06/03/26
unsigned short Set_Speed(unsigned int speed, unsigned short dev_no)
{
	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}
	
	if (speed > 2) {
		return(PRM_ERR);
	}
	
	dev_speed[dev_no] = speed;
	
	return(0);
}

//
// Get Ethernet Speed (0:10Mbps/1:100Mbps, 0:half_duplex/1:full_duplex)
// 06/03/26
unsigned short Get_Speed(unsigned int *speed, unsigned int *duplex, unsigned short dev_no)
{
	clock_t				s_time, c_time;
	clock_t				duration;		/*	05.09.16	*/
	struct GetModeInfo	*ptr;
	int					ret;

	if (dev_no != 0 && dev_no!= 1) {
		return(PRM_ERR);
	}
	
	if (dev_opn[dev_no] == 0) {
		return(DEV_NOT_OPEN);
	}
	
	if (dev_tbl[dev_no].sum !=
		CalSharedSum(&dev_tbl[dev_no], sizeof(struct dev_info) / 2 - 1)) {
		return(FATAL_ERR);
	}

	/*	MUTEX取得	*/
	ret = pthread_mutex_lock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	// 上位層ステータス指定 要求コマンドをセット。
	*(unsigned short *)FLtoAP_RSP = 0;
	*(unsigned short *)APtoFL_CMD = REQ_SPEED;
	
	// FL-net PCIカードからのレスポンスを待つ。
	s_time = clock();
	while (*(volatile unsigned short *)FLtoAP_RSP != RSP_SPEED) {
		c_time = clock();
		duration = (c_time - s_time) / CLOCKS_PER_SEC;
		// メッセージ送信終了レスポンスをチェックする。
		if (duration > 1) {		/* 1 秒  */
			if (*(unsigned short *)FLtoAP_CMD == REP_FATAL_ERR) {
				/*	MUTEX解放	*/
				ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
				if (ret != 0) {
					return(MUTEX_ERR);
				}
				return(FATAL_ERR);
			} else {
				if (*(volatile unsigned short *)FLtoAP_RSP == RSP_SPEED) {
					break;
				} else {
					*(unsigned short *)APtoFL_CMD = 0;
					/*	MUTEX解放	*/
					ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
					if (ret != 0) {
						return(MUTEX_ERR);
					}
					return(TIM_OUT);
				}
			}
		}
	}
	
	ptr = (struct GetModeInfo *)FLtoAP_DATA;
	if (ptr->result != 0) {		// 結果コード
		*(unsigned short *)FLtoAP_RSP = 0;
		/*	MUTEX解放	*/
		ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
		if (ret != 0) {
			return(MUTEX_ERR);
		}
		return(LNK_DOWN);
	}
	*speed  = ptr->speed;
	*duplex = ptr->duplex;

	*(unsigned short *)FLtoAP_RSP = 0;

	/*	MUTEX解放	*/
	ret = pthread_mutex_unlock(&mutex[dev_no][MUTEX_AP_CMD]);
	if (ret != 0) {
		return(MUTEX_ERR);
	}
	
	return(0);
}

static unsigned short CalSharedSum(struct dev_info *ptr, unsigned short num)
{
	unsigned short	sum;
	unsigned short	*s_ptr;

	s_ptr = (unsigned short *)ptr; 

	for (sum = 0; num > 0; num--) {
		sum += *s_ptr++;
	}
	
	return(sum);
}

