//
// flpci.h : FL-PCI用APIライブラリのヘッダ
//

///////////////////////////////////
// エラーＣｏｄｅ
///////////////////////////////////
#define PRM_ERR			2	// パラメタ エラー
#define BUF_FUL			3	// バッファ フル
#define NO_DATA			4	// データ なし
#define OUT_RING		5	// 離脱中
#define TX_FAIL			6	// 送信 失敗
#define TX_NOT_END		7	// 送信 未完
#define TIM_OUT			8	// タイムアウト
#define FATAL_ERR		9	// 重大エラー
#define DEV_ERR1		10	// デバイスオープンエラー1
//#define DEV_ERR2		11	// デバイスオープンエラー2
//#define DEV_ERR3		12	// デバイスオープンエラー3
//#define DEV_ERR4		13	// デバイスオープンエラー4
#define DEV_NOT_OPEN	14	// デバイスオープンなし
#define DEV_ERR5		15	// デバイスクローズエラー
#define DEV_DUPL_OPN	16	// デバイス２重オープン
//#define NON_OTER_DEV	17	// Other Device Not Open
//#define DEV_BUSY		18	// デバイスBUSY
#define MUTEX_ERR		19	// 排他処理エラー
#define LNK_DOWN		20	// ケーブル外れによる処理未完 06/03/26

#pragma pack(1)

///////////////////////////////////
// Dual Port RAM I/F 構造体
///////////////////////////////////
// 参加ノード要求コマンド構造体
struct SankaReqTable {
	unsigned char	node_cnt;		// 参加ノード数
	unsigned char	node_no[254];	// 参加ノードＮｏ．
};

// ネットワーク情報コマンド構造体
struct NetInfo {
	unsigned short	rct;		// リフレッシュサイクル 設定値
	unsigned short	rcm;		// リフレッシュサイクル 測定値（現在）
	unsigned short	rcm_max;	// リフレッシュサイクル 測定値（最大）
	unsigned short	rcm_min;	// リフレッシュサイクル 測定値（最小）
	unsigned char	link_status;// リンクの状態
};

// 重大エラー情報コマンド 構造体 
struct ErrInfo {
	unsigned short	err_code;		// エラーコード
	unsigned short	length;			// メッセージ長
	unsigned char	err_msg[128];	// エラーメッセージ
};

/////////////////////////////////////////////
// Library 関数 アプリケーションI/F 構造体
/////////////////////////////////////////////
// 初期化 パラメタ構造体
struct InitInfo {
	unsigned int	ip_adr;		// 自ノードＩＰアドレス
	unsigned short  dev_no;     // デバイスＮｏ．
	unsigned short	c1_adr;		// コモンメモリ 領域１ アドレス
	unsigned short	c1_size;	// コモンメモリ 領域１ サイズ
	unsigned short	c2_adr;		// コモンメモリ 領域２ アドレス
	unsigned short	c2_size;	// コモンメモリ 領域２ サイズ
	unsigned char	tw;			// トークン監視タイムアウト値
	unsigned char	mft;		// 最小フレーム間隔
	unsigned char	vender[11];	// ベンダ名
	unsigned char	maker[11];	// メーカー名
	unsigned char	name[11];	// ノード名
};

// ノード情報要求 構造体
struct NodeInfo {
	unsigned char	sanka;		// 参加／離脱 フラグ
	unsigned short	uls;		// 上位層ステータス
	unsigned short	cmn1_addr;	// コモン領域１アドレス
	unsigned short	cmn1_size;	// コモン領域１サイズ
	unsigned short	cmn2_addr;	// コモン領域２アドレス
	unsigned short	cmn2_size;	// コモン領域２サイズ
	unsigned short	rct;		// リフレッシュサイクル許容時間
	unsigned char	tw;			// トークン監視時間
	unsigned char	mft;		// 最小フレーム間隔
	unsigned char	link_status;// リンクの状態
	unsigned char	my_status;	// 自ノードの状態
	unsigned char	vender[11];	// ベンダー名
	unsigned char	maker[11];	// メーカー名
	unsigned char	name[11];	// ノード名
};

// メッセージ送信 構造体
struct MessageInfo {
	unsigned char	msg_node;	// ノード番号(1～254) 
	unsigned char	msg_result;	// 結果コード
	unsigned int	msg_tcd;	// トランザクションコード
	unsigned short	msg_bc;		// メッセージデータバイト長
	unsigned int	msg_adr;	// 仮想アドレス空間 アドレス
	unsigned short	msg_lng;	// 仮想アドレス空間 データ長
};

// 前ノード ULS & LKS 構造体
struct UlsInfo {
	unsigned short	uls[254];
	unsigned char	lks[254];
};


///////////////////////////////////
// Library 関数
///////////////////////////////////
unsigned short FL_Initialize(struct InitInfo *);
unsigned short DeviceClose(unsigned short, unsigned short);
unsigned short WriteCyclicALL(unsigned short *, unsigned short *, unsigned short);
unsigned short WriteCyclicRelative(unsigned short, unsigned short, unsigned short,
								   unsigned short *, unsigned short);
unsigned short ReadCyclicRelative(unsigned short, unsigned short, unsigned short,
								  unsigned short, unsigned short *, unsigned short);
unsigned short ReadCyclicAbsolute(unsigned short, unsigned short, unsigned short,
								  unsigned short *, unsigned short);
unsigned short ReadCyclicALL(unsigned short, unsigned short *, unsigned short *,
							 unsigned short);
unsigned short WriteMessage(struct MessageInfo *, unsigned char *, unsigned short,
							unsigned short);
unsigned short CheckMessageEnd(unsigned short *, unsigned short *, unsigned short);
unsigned short ReadMessage(struct MessageInfo *, unsigned char *, unsigned short,
						   unsigned short);
unsigned short GetRingInfo(struct SankaReqTable *, unsigned short);
unsigned short GetNodeInfo(unsigned short, struct NodeInfo *, unsigned short);
unsigned short GetNetInfo(struct NetInfo *, unsigned short);
unsigned short GetLogInfo(unsigned int *, unsigned short);
unsigned short SetUplStatus(unsigned short, unsigned short);
unsigned short GetErrInfo(struct ErrInfo *, unsigned short);
unsigned short ReadUls_All(struct UlsInfo *, unsigned short);
unsigned short Dll_Version(void);
unsigned short Firmware_Version(unsigned short *, unsigned short);	/* 04/04/06 */
unsigned short Set_Speed(unsigned int, unsigned short);			/* 100Mbps 06/03/26 */
unsigned short Get_Speed(unsigned int *, unsigned int *, unsigned short);	/* 100Mbps 06/03/26 */

