//
// sample.c : FL-net PCIカード サンプルアプリケーションプログラム
//

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <cs/flpci.h>
#include <sys/types.h>

// メッセージ TCD Code
#define BBR_REQ	65003
#define BBW_REQ	65004
#define BWR_REQ	65005
#define BWW_REQ	65006
#define NTR_REQ	65007
#define NTW_REQ	65008
#define STP_REQ	65009
#define RUN_REQ	65010
#define PFR_REQ	65011
#define LGR_REQ	65013
#define LGC_REQ	65014
#define LPB_REQ	65015

#define BBR_RSP	65203
#define BBW_RSP	65204
#define BWR_RSP	65205
#define BWW_RSP	65206
#define NTR_RSP	65207
#define NTW_RSP	65208
#define STP_RSP	65209
#define RUN_RSP	65210
#define PFR_RSP	65211
#define LPB_RSP	65215

// LOOP BACK TEST : メッセージ送受信 テーブル数
#define MSG_SND_MAX			2
#define MSG_RCV_MAX			6

#define IP_ADDRESS_LEN		15
#define VENDER_LEN			10
#define MAKER_LEN			10
#define NODE_LEN			10
#define WORD_DATA_LEN		4
#define LOG_BUF_SIZE		128
#define MSG_BUF_SIZE		1024
#define ENTER_BUF_SIZE		256
#define COMMON1_SIZE		512
#define COMMON2_SIZE		COMMON1_SIZE * 16	// 8192
#define STRING_NUMERIC		"1234567890"
#define STRING_HEX_NUMERIC	"1234567890ABCDEFabcdef"

//#define MUTEX_TEST

#pragma pack(1)

////////////////////////////////////////
// アプリケーション グローバル変数
////////////////////////////////////////
//自ノード情報
unsigned short	my_node;
int				dev_no;
struct NodeInfo	node_prm;

////////////////////////////////////////
// ローカル関数のプロトタイプ宣言
////////////////////////////////////////
static void Initialize(void);
static void WriteCyc(void);
static void WriteCyc_R(void);
static void ReadCyc_R(void);
static void ReadCyc_A(void);
static void ReadCyc_AL(void);
static void RingInfo(void);
static void DspNodeInfo(void);
static void DspNetInfo(void);
static void DspLogInfo(void);
static void UplStatus(void);
static void DspFatalErr(void);
static void MsgSendReceive(void);
static void WriteMsg(void);
static void CheckMsg(void);
static void ReadMsg(void);
static void MessageLoopBack(void);
static void DevCls(void);
static void UlsLks(void);
static void Firmware_ver_dsp(void);
static void DspEtherMode(void);		/* 100Mbps support 06/03/26 */
static int Address_check(const char *value);
static char* Strskipsearch(const char *ptr, const char *skip);
static char* Strskip(const char *ptr, const char *skip);
static int Strnumeric(char *ptr);
static void Cyc_Dump(FILE *fp, unsigned short *data, unsigned short addr, int len);
static void Dump(FILE *fp, unsigned char *data, int len);
static void Int_Dump(FILE *fp, unsigned int *data, int len);
static int Wrapper_Scanf(int *number);
static int Wrapper_Hex_Scanf(int *number, unsigned int len);
static int axtoi(const char *buffer);

#ifdef MUTEX_TEST
void *func1(void* pParam)
{
	int	i;
	
	fprintf(stderr, "start thread1\n");
	for (i = 0; i < 10; i++) {
		DspEtherMode();
#if 0
		printf("output");
		usleep(5);
		printf(" by");
		usleep(5);
		printf(" func1 i = %d\n",i);
#endif
	}
	
	return NULL;
}

void *func2(void* pParam)
{
	int	i;
	
	fprintf(stderr, "start thread2\n");
	for(i = 0; i < 10; i++) {
		DspEtherMode();
#if 0
		printf("OUTPUT");
		usleep(5);
		printf(" BY");
		usleep(5);
		printf(" FUNC2 I = %d\n",i);
#endif
	}
	
	return NULL;
}
#endif

//
// main
//
int main()
{
	int				number = -1, ret;
	unsigned short	rc;

#ifdef MUTEX_TEST
	pthread_t		thread1, thread2;
#endif

	dev_no = -1;
	printf("FLPCIsample - FL-net PCIカード(100Mbps版) サンプルアプリケーションプログラム\n");
	printf("Copyright (C) Century Systems, 2010\n");

	// デバイスNo.を入力する。
	while (1) {
		printf("\n制御したいデバイスを選択して下さい。(デバイス0 or デバイス1)\n");
		printf("デバイス選択(0 or 1) : ");
		ret = Wrapper_Scanf(&dev_no);
		if (ret != 0) {
			printf("Invalid data!\n");
			continue;
		}
		if (dev_no != 0 && dev_no != 1) {
			printf("Invalid data!\n");
			continue;
		}
		break;
	}
	
	// FL-net PCIカードを初期化する。
	while (1) {
		printf("\n<<<<<<<<< Test Menu(デバイスNo.%d) <<<<<<<<<<\n", dev_no);
		printf("0)  Initialize\n");
		printf("1)  Write cyclic data(all)\n");
		printf("2)  Write cyclic data(relative offset)\n");
		printf("3)  Read cyclic data (relative offset)\n");
		printf("4)  Read cyclic data (absolute offset)\n");
		printf("5)  Read cyclic data (all)\n");
		printf("6)  メッセージ送受信\n");
		printf("7)  参加ノード 表示\n");
		printf("8)  ノード情報 表示\n");
		printf("9)  ネットワーク情報 表示\n");
		printf("10) ログ情報 表示\n");
		printf("11) 上位層ステータス 指定\n");
		printf("12) 重大エラー詳細 表示\n");
		printf("14) デバイス クローズ\n");
		printf("15) ULS/LKS 表示\n");
		printf("16) DLL Version 表示\n");
		printf("17) Firmware Version 表示\n");
		printf("18) FL-PCI Card通信速度 表示\n");  /* 100Mbps support 06/03/26 */
		printf("19) 終了\n");
#ifdef MUTEX_TEST
		printf("20) MUTEXテスト\n");
#endif
		printf("Enter number : ");
		ret = Wrapper_Scanf(&number);
		if (ret != 0) {
			printf("Invalid data!\n");
			continue;
		}
		printf("\n");
		switch(number){
		case 0:
			Initialize();
			break;
		case 1:
			WriteCyc();
			break;
		case 2:
			WriteCyc_R();
			break;
		case 3:
			ReadCyc_R();
			break;
		case 4:
			ReadCyc_A();
			break;
		case 5:
			ReadCyc_AL();
			break;
		case 6:
			MsgSendReceive();
			break;
		case 7:
			RingInfo();
			break;
		case 8:
			DspNodeInfo();
			break;
		case 9:
			DspNetInfo();
			break;
		case 10:
			DspLogInfo();
			break;
		case 11:
			UplStatus();
			break;
		case 12:
			DspFatalErr();
			break;
		case 14:
			DevCls();
			break;
		case 15:
			UlsLks();
			break;
		case 16:
			printf("DLL Version = %04d\n", Dll_Version());
			break;
		case 17:
			Firmware_ver_dsp();
			break;
		case 18:
			DspEtherMode();  /* 100Mbps support 06/03/26 */
			break;
		case 19:
			rc = DeviceClose((unsigned short)dev_no, 1);
			if (rc <= 0) {
				printf("[Device Close] Device No.%d Close completed.\n\n", dev_no);
			}
			return(0);
#ifdef MUTEX_TEST
		case 20:
			if (pthread_create(&thread1, NULL, func1, NULL) != 0) {
				perror("pthread_create()");
			}
			if (pthread_create(&thread2, NULL, func2, NULL) != 0) {
				perror("pthread_create()");
			}
			pthread_join(thread1, NULL);
			pthread_join(thread2, NULL);
			break;
#endif
		default:
			printf("Invalid data!\n");
			break;
		}
	}
	return(0);
}

//
// Initialize : FL-net PCIカードを初期化する。
//
static void Initialize(void)
{
	struct InitInfo	init_prm;
	unsigned short	rc;
	int				temp, speed, ret;
	char			string[32];

	init_prm.dev_no = (unsigned short)dev_no;

	while (1) {
		memset(string, 0, sizeof(string));
		// ＩＰアドレスを入力する。
		printf("IPアドレスを入力してください。（XXX.XXX.XXX.XXX）\n");
		scanf("%s", string);
		
		// 文字列長チェック。
		if (strlen(string) > IP_ADDRESS_LEN) {
			printf("Length Error!\n\n");
			continue;
		}
		
		// IPアドレスの文字列確認。
		ret = Address_check(string);
		if (ret < 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		
		init_prm.ip_adr = (unsigned long)inet_network(string);
		
		// classCチェック
		if ((init_prm.ip_adr & 0xffff0000) != 0xc0a80000) {
			printf("This is not a class C.\n\n");
			continue;
		}
		break;
	}
	
	// 領域１のアドレスを入力する。
	while (1) {
		printf("コモン領域1 アドレスを指定してください。（0-511）\n");
		ret = Wrapper_Scanf(&temp);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (temp < 0 || temp >= COMMON1_SIZE) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	init_prm.c1_adr = (unsigned short)temp;

	// 領域１のサイズを入力する。
	while (1) {
		printf("コモン領域1 サイズを指定してください。（0-%d）\n",
			   COMMON1_SIZE - init_prm.c1_adr);
		ret = Wrapper_Scanf(&temp);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (temp < 0 || (init_prm.c1_adr + temp) > COMMON1_SIZE) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	init_prm.c1_size = (unsigned short)temp;

	// 領域２のアドレスを入力する。
	while (1) {
		printf("コモン領域2 アドレスを指定してください。（0-8191）\n");
		ret = Wrapper_Scanf(&temp);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (temp < 0 || temp >= COMMON2_SIZE) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	init_prm.c2_adr = (unsigned short)temp;

	// 領域２のサイズを入力する。
	while (1) {
		printf("コモン領域2 サイズを指定してください。（0-%d）\n",
			   COMMON2_SIZE - init_prm.c2_adr);
		ret = Wrapper_Scanf(&temp);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (temp < 0 || (init_prm.c2_adr + temp) > COMMON2_SIZE) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	init_prm.c2_size = (unsigned short)temp;

	// トークン監視時間を入力する。
	while (1) {
		printf("トークン監視時間を指定してください。（1-255）\n");
		ret = Wrapper_Scanf(&temp);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (temp < 1 || temp > 255) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	init_prm.tw = (unsigned char)temp;

	// 最小フレーム間隔を入力する。
	while (1) {
		printf("最小フレーム間隔を指定してください。（0-50）\n");
		ret = Wrapper_Scanf(&temp);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (temp < 0 || temp > 50) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	init_prm.mft = (unsigned char)temp;

	while (1) {
		// ベンダー名をを入力する。
		memset(init_prm.vender, 0,  sizeof(init_prm.vender));
		printf("ベンダー名を入力してください。（10文字以内）\n");
		scanf("%s", (char *)&init_prm.vender);
		// 文字列長チェック。
		if (strlen((char *)&init_prm.vender) > VENDER_LEN) {
			printf("Length Error!\n\n");
			continue;
		}
		break;
	}

	while (1) {
		// メーカー名を入力する。
		memset(init_prm.maker, 0,  sizeof(init_prm.maker));
		printf("メーカー名を入力してください。（10文字以内）\n");
		scanf("%s", (char *)&init_prm.maker);
		// 文字列長チェック。
		if (strlen((char *)&init_prm.maker) > MAKER_LEN) {
			printf("Length Error!\n\n");
			continue;
		}
		break;
	}

	while (1) {
		// ノード名を入力する。
		memset(init_prm.name, 0, sizeof(init_prm.name));
		printf("ノード名を入力してください。（10文字以内）\n");
		scanf("%s", (char *)&init_prm.name);
		// 文字列長チェック。
		if (strlen((char *)&init_prm.name) > NODE_LEN) {
			printf("Length Error!\n\n");
			continue;
		}
		break;
	}

	/* Ethernet通信速度を設定する 100Mbps support 06/03/26 */
	while (1) {
		printf("通信速度を選択してください。 (0:10Mbps 1:100Mbps 2:自動選択)\n");
		ret = Wrapper_Scanf(&speed);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (speed < 0 || speed > 2) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}

	rc = Set_Speed((unsigned long)speed, (unsigned short)dev_no);
	switch (rc) {
	case 0:
		break;
	case PRM_ERR:
		printf("[Set_Speed] パラメタエラー\n");
		return;
	default:
		printf("[Set_Speed] Other Error = %d", rc);
		return;
	}

	rc =  FL_Initialize(&init_prm);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("[FL_Initialize] パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("[FL_Initialize] PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("[FL_Initialize] 重大エラー\n");
			break;
		case DEV_ERR1:
			printf("[FL_Initialize] device open error.\n");
			break;
		case DEV_DUPL_OPN:
			printf("[FL_Initialize] デバイス2重オープンエラー\n");
			break;
		default:
			printf("[FL_Initialize] Other Error = %d\n", rc);
			break;
		}
		return;
	} else {
		printf("[Initialize] PCI card initialize completed.\n");
		my_node = (unsigned short)(init_prm.ip_adr & 0x000000ff);
		//以降の処理で自ノード設定値を使用するため、ノード情報を取り込む
		rc = GetNodeInfo(my_node, &node_prm, (unsigned short)dev_no);
		if (rc > 0) {
			switch (rc) {
			case PRM_ERR:
				printf("[GetNodeInfo] パラメタエラー\n");
				break;
			case TIM_OUT:
				printf("[GetNodeInfo] PCIからのレスポンスタイムアウト\n");
				break;
			case FATAL_ERR:
				printf("[GetNodeInfo] 重大エラー\n");
				break;
			case DEV_NOT_OPEN:
				printf("[GetNodeInfo] 指定デバイスはOpenしていません。\n");
				break;
			case MUTEX_ERR:
				printf("[GetNodeInfo] MUTEX制御エラー\n");
				break;
			default:
				printf("[GetNodeInfo] Other Error = %d\n", rc);
				break;
			}
			return;
		}
		return;
	}
}

//
//
//
static void DevCls(void)
{
	unsigned short	rc;
	int				reset_f, ret;

	// RESET指定を入力する。
	while (1) {
		printf("RESETを指定してください。 (0:NOT RESET 1:RESET)\n");
		ret = Wrapper_Scanf(&reset_f);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (reset_f != 0 && reset_f != 1) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	
	rc = DeviceClose((unsigned short)dev_no, (unsigned short)reset_f);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("パラメタエラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("指定デバイスはOpenしていません。\n");
			break;
		case DEV_ERR5:
			printf("デバイスクローズエラー\n");
			break;
		default:
			printf("[Device Close] Other Error = %d\n", rc);
			break;
		}
	} else {
		printf("[Device Close] Device No.%d Close completed.\n", dev_no);
	}
}

//
//
//
static void WriteCyc(void)
{
	unsigned int	i;
	unsigned short	*ptr, rc;
	unsigned short	dsp_buf1[COMMON1_SIZE];
	unsigned short	dsp_buf2[COMMON2_SIZE];
	int				ret, data1, data2;
	
	memset(dsp_buf1, 0, sizeof(dsp_buf1));
	memset(dsp_buf2, 0, sizeof(dsp_buf2));
	
	rc = GetNodeInfo(my_node, &node_prm, (unsigned short)dev_no);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("[GetNodeInfo] パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("[GetNodeInfo] PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("[GetNodeInfo] 重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("[GetNodeInfo] 指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("[GetNodeInfo] MUTEX制御エラー\n");
			break;
		default:
			printf("[GetNodeInfo] Other Error = %d", rc);
			return;
		}
		return;
	}
	
	while (1) {
		// 領域１の Writeデータを入力する
		printf("Enter Common1 Word-Data (Hex) ");
		ret = Wrapper_Hex_Scanf(&data1, WORD_DATA_LEN);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	
	while (1) {
		// 領域２の Writeデータを入力する
		printf("Enter Common2 Word-Data (Hex) ");
		ret = Wrapper_Hex_Scanf(&data2, WORD_DATA_LEN);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	
	ptr = (unsigned short *)&dsp_buf1[0];
	for (i = 0; i < node_prm.cmn1_size; i++) {
		*ptr++ = (unsigned short)data1;
	}
	ptr = (unsigned short *)&dsp_buf2[0];
	for (i = 0; i < node_prm.cmn2_size; i++){
		*ptr++ = (unsigned short)data2;
	}

	rc = WriteCyclicALL(&dsp_buf1[0], &dsp_buf2[0], (unsigned short)dev_no);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("[WriteCyclicALL] パラメタエラー\n");
			break;
		case BUF_FUL:
			printf("[WriteCyclicALL] バッファフル\n");
			break;
		case FATAL_ERR:
			printf("[WriteCyclicALL] 重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("[WriteCyclicALL] 指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("[WriteCyclicALL] MUTEX制御エラー\n");
			break;
		default:
			printf("[WriteCyclicALL] Other Error = %d\n", rc);
			break;
		}
	}
}

//
//
//
static void WriteCyc_R(void)
{
	unsigned short	*ptr, rc;
	unsigned short	dsp_buf[COMMON2_SIZE];
	int				common, ret, i;
	int				offset, cmn_size, size, data1;
	
	memset(dsp_buf, 0, sizeof(dsp_buf));
	
	rc = GetNodeInfo(my_node, &node_prm, (unsigned short)dev_no);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("[GetNodeInfo] パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("[GetNodeInfo] PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("[GetNodeInfo] 重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("[GetNodeInfo] 指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("[GetNodeInfo] MUTEX制御エラー\n");
			break;
		default:
			printf("[GetNodeInfo] Other Error = %d", rc);
			return;
		}
		return;
	}
	
	// 領域１、領域２のどちらを書き込むか？
	while (1) {
		printf("どちらのコモン領域を書き込みますか？\n");
		printf("1) 領域1\n");
		printf("2) 領域2\n");
		printf("Enter number : ");
		ret = Wrapper_Scanf(&common);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (common != 1 && common != 2) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	
	if (common == 1) {
		cmn_size = COMMON1_SIZE;
	} else {
		cmn_size = COMMON2_SIZE;
	}

	// 相対アドレスを入力する。
	while (1) {
		if (common == 1) {
			if (node_prm.cmn1_size == 0) {
				printf("コモン領域1のサイズが0で書込めません!\n");
				return;
			}
			printf("書き込みを開始する相対アドレスを入力してください。（0-%d）: "
				   , node_prm.cmn1_size - 1);
			ret = Wrapper_Scanf(&offset);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (offset > node_prm.cmn1_size - 1) {
				printf("Invalid data!\n\n");
				continue;
			}
		} else {
			if (node_prm.cmn2_size == 0) {
				printf("コモン領域2のサイズが0で書込めません!\n");
				return;
			}
			printf("書き込みを開始する相対アドレスを入力してください。（0-%d）: "
				   , node_prm.cmn2_size - 1);
			ret = Wrapper_Scanf(&offset);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (offset > node_prm.cmn2_size - 1) {
				printf("Invalid data!\n\n");
				continue;
			}
		}
		break;
	}

	// 書き込むワード数を入力する。
	while (1) {
		if (common == 1) {
			printf("書き込むワード数を入力してください。（1-%d） : "
				   , node_prm.cmn1_size - offset);
			ret = Wrapper_Scanf(&size);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (size < 1 || size > node_prm.cmn1_size - offset) {
				printf("Invalid data!\n\n");
				continue;
			}
		} else {
			printf("書き込むワード数を入力してください。（1-%d） : "
				   , node_prm.cmn2_size - offset);
			ret = Wrapper_Scanf(&size);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (size < 1 || size > node_prm.cmn2_size - offset) {
				printf("Invalid data!\n\n");
				continue;
			}
		}
		break;
	}

	while (1) {
		//  Writeデータを入力する
		printf("Enter Common Word-Data (Hex) ");
		ret = Wrapper_Hex_Scanf(&data1, WORD_DATA_LEN);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}

	ptr = (unsigned short *)&dsp_buf[0];
	for (i = 0; i < size; i++) {
		*ptr++ = (unsigned short)data1;
	}

	rc = WriteCyclicRelative((unsigned short)common, (unsigned short)offset,
							 (unsigned short)size, &dsp_buf[0], (unsigned short)dev_no);
	switch (rc) {
	case 0:
		break;
	case PRM_ERR:
		printf("パラメタエラー\n");
		break;
	case BUF_FUL:
		printf("バッファフル\n");
		break;
	case FATAL_ERR:
		printf("重大エラー\n");
		break;
	case DEV_NOT_OPEN:
		printf("指定デバイスはOpenしていません。\n");
		break;
	case MUTEX_ERR:
		printf("MUTEX制御エラー\n");
		break;
	default:
		printf("[WriteCyclicRelative] Other Error = %d\n", rc);
		break;
	}
}

//
//
//
static void ReadCyc_R(void)
{
	unsigned short	rc;
	struct NodeInfo	node_info;
	unsigned short	dsp_buf[COMMON2_SIZE];
	int				ret, node, common, size, offset;
	
	memset(dsp_buf, 0, sizeof(dsp_buf));

	// ノード番号を入力する。
	while (1) {
		printf("Enter node number (1-254) : ");
		ret = Wrapper_Scanf(&node);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (node < 1 || node > 254) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}

	rc = GetNodeInfo((unsigned short)node, &node_info, (unsigned short)dev_no);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("[GetNodeInfo] パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("[GetNodeInfo] PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("[GetNodeInfo] 重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("[GetNodeInfo] 指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("[GetNodeInfo] MUTEX制御エラー\n");
			break;
		default:
			printf("[GetNodeInfo] Other Error = %d", rc);
			return;
		}
		return;
	}
	
	if (node_info.sanka == 0) {
		printf("ノード離脱\n");
		return;
	}
	
	// 領域１、領域２のどちらを読み出すか？
	while (1) {
		printf("どちらのコモン領域を読み出しますか？\n");
		printf("1) 領域1\n");
		printf("2) 領域2\n");
		printf("Enter number : ");
		ret = Wrapper_Scanf(&common);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (common != 1 && common != 2) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	
	// 相対アドレスを入力する。
	while (1) {
		if (common == 1) {
			if (node_info.cmn1_size == 0) {
				printf("コモン領域1のサイズが0で読み込めません!\n");
				return;
			}
			printf("読み出しを開始する相対アドレスを入力してください。（0-%d）: "
				   , node_info.cmn1_size - 1);
			ret = Wrapper_Scanf(&offset);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (offset > node_info.cmn1_size - 1) {
				printf("Invalid data!\n\n");
				continue;
			}
		} else {
			if (node_info.cmn2_size == 0) {
				printf("コモン領域2のサイズが0で読み込めません!\n");
				return;
			}
			printf("読み出しを開始する相対アドレスを入力してください。（0-%d）: "
				   , node_info.cmn2_size - 1);
			ret = Wrapper_Scanf(&offset);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (offset > node_info.cmn2_size - 1) {
				printf("Invalid data!\n\n");
				continue;
			}
		}
		break;
	}

	// 読み出すワード数を入力する。
	while (1) {
		if (common == 1) {
			printf("読み出すワード数を入力してください。（1-%d） : "
				   , node_info.cmn1_size - offset);
			ret = Wrapper_Scanf(&size);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (size < 1 || size > node_info.cmn1_size - offset) {
				printf("Invalid data!\n\n");
				continue;
			}
		} else {
			printf("読み出すワード数を入力してください。（1-%d） : "
				   , node_info.cmn2_size - offset);
			ret = Wrapper_Scanf(&size);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (size < 1 || size > node_info.cmn2_size - offset) {
				printf("Invalid data!\n\n");
				continue;
			}
		}
		break;
	}
	
	rc = ReadCyclicRelative((unsigned short)node, (unsigned short)common,
							(unsigned short)offset, (unsigned short)size,
							&dsp_buf[0], (unsigned short)dev_no);
	switch (rc) {
	case 0:
		if (common == 1) {
			Cyc_Dump(stdout, dsp_buf, node_info.cmn1_addr + offset, size);
		} else {
			Cyc_Dump(stdout, dsp_buf, node_info.cmn2_addr + offset, size);
		}
		break;
	case PRM_ERR:
		printf("[ReadCyclicRelative] パラメタエラー\n");
		break;
	case NO_DATA:
		printf("[ReadCyclicRelative] 指定ノードのデータなし\n");
		break;
	case OUT_RING:
		printf("[ReadCyclicRelative] ノード離脱\n");
		break;
	case FATAL_ERR:
		printf("[ReadCyclicRelative] 重大エラー\n");
		break;
	case DEV_NOT_OPEN:
		printf("[ReadCyclicRelative] 指定デバイスはOpenしていません。\n");
		break;
	case MUTEX_ERR:
		printf("[ReadCyclicRelative] MUTEX制御エラー\n");
		break;
	default:
		printf("[ReadCyclicRelative] Other Error = %d\n", rc);
		break;
	}
}

//
//
//
static void ReadCyc_A(void)
{
	unsigned short	rc;
	unsigned short	dsp_buf[COMMON2_SIZE];
	int				ret, common, offset, size;
	
	memset(dsp_buf, 0, sizeof(dsp_buf));
	
	// 領域１、領域２のどちらを読み出すか？
	while (1) {
		printf("どちらのコモン領域を読み出しますか？\n");
		printf("1) 領域1\n");
		printf("2) 領域2\n");
		printf("Enter number : ");
		ret = Wrapper_Scanf(&common);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (common != 1 && common != 2) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	
	// 絶対アドレスを入力する。
	while (1) {
		if (common == 1) {
			printf("読み出しを開始する絶対アドレスを入力してください。（0-511）: ");
			ret = Wrapper_Scanf(&offset);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (offset > COMMON1_SIZE - 1) {
				printf("Invalid data!\n\n");
				continue;
			}
		} else {
			printf("読み出しを開始する絶対アドレスを入力してください。（0-8191）: ");
			ret = Wrapper_Scanf(&offset);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (offset > COMMON2_SIZE - 1) {
				printf("Invalid data!\n\n");
				continue;
			}
		}
		break;
	}
	
	// 読み出すワード数を入力する。
	while (1) {
		if (common == 1) {
			printf("読み出すワード数を入力してください。（0-512） : ");
			ret = Wrapper_Scanf(&size);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (size > COMMON1_SIZE - offset) {
				printf("Size over!\n\n");
				continue;
			}
		} else {
			printf("読み出すワード数を入力してください。（0-8192） : ");
			ret = Wrapper_Scanf(&size);
			if (ret != 0) {
				printf("Invalid data!\n\n");
				continue;
			}
			if (size > COMMON2_SIZE - offset) {
				printf("Size over!\n\n");
				continue;
			}
		}
		break;
	}
	
	rc = ReadCyclicAbsolute((unsigned short)common, (unsigned short)offset,
							(unsigned short)size, &dsp_buf[0], (unsigned short)dev_no);
	switch (rc) {
	case 0:
		Cyc_Dump(stdout, dsp_buf, offset, size);
		break;
	case PRM_ERR:
		printf("パラメタエラー\n");
		break;
	case FATAL_ERR:
		printf("重大エラー\n");
		break;
	case DEV_NOT_OPEN:
		printf("指定デバイスはOpenしていません。\n");
		break;
	case MUTEX_ERR:
		printf("MUTEX制御エラー\n");
		break;
	default:
		printf("[ReadCyclicAbsolute] Other Error = %d\n", rc);
		break;
	}
}

//
//
//
static void ReadCyc_AL(void)
{
	unsigned short	rc;
	struct NodeInfo	node_info;
	unsigned short	dsp_buf1[COMMON1_SIZE];
	unsigned short	dsp_buf2[COMMON2_SIZE];
	int				ret, node;
	
	memset(dsp_buf1, 0, sizeof(dsp_buf1));
	memset(dsp_buf2, 0, sizeof(dsp_buf2));
	
	// ノード番号を入力する。
	while (1) {
		printf("Enter node number (1-254) : ");
		ret = Wrapper_Scanf(&node);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (node < 1 || node > 254) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}

	rc = GetNodeInfo((unsigned short)node, &node_info, (unsigned short)dev_no);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("[GetNodeInfo] パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("[GetNodeInfo] PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("[GetNodeInfo] 重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("[GetNodeInfo] 指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("[GetNodeInfo] MUTEX制御エラー\n");
			break;
		default:
			printf("[GetNodeInfo] Other Error = %d", rc);
			return;
		}
		return;
	}
	
	rc = ReadCyclicALL((unsigned short)node, dsp_buf1, dsp_buf2, (unsigned short)dev_no);
	
	switch (rc) {
	case 0:
		printf("[Common1]\n");
		Cyc_Dump(stdout, dsp_buf1, node_info.cmn1_addr, node_info.cmn1_size);
		printf("\n[Common2]\n");
		Cyc_Dump(stdout, dsp_buf2, node_info.cmn2_addr, node_info.cmn2_size);
		break;
	case PRM_ERR:
		printf("[ReadCyclicALL] パラメタエラー\n");
		break;
	case NO_DATA:
		printf("[ReadCyclicALL] 指定ノードのデータなし\n");
		break;
	case OUT_RING:
		printf("[ReadCyclicALL] ノード離脱\n");
		break;
	case FATAL_ERR:
		printf("[ReadCyclicALL] 重大エラー\n");
		break;
	case DEV_NOT_OPEN:
		printf("[ReadCyclicALL] 指定デバイスはOpenしていません。\n");
		break;
	case MUTEX_ERR:
		printf("[ReadCyclicALL] MUTEX制御エラー\n");
		break;
	default:
		printf("[ReadCyclicALL] Other Error = %d\n", rc);
		break;
	}
}

//
//
//
static void RingInfo(void)
{
	struct SankaReqTable	sanka_prm, *ptr1;
	unsigned short			rc, i;
	
	rc = GetRingInfo(&sanka_prm, (unsigned short)dev_no);
	
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("MUTEX制御エラー\n");
			break;
		default:
			printf("[GetRingInfo] Other Error = %d\n", rc);
			break;
		}
		return;
	}
	
	// 参加ノードNo.を表示する。
	ptr1 = &sanka_prm;
	if (ptr1->node_cnt == 0) {
		printf("自ノードは離脱しています。\n");
		return;
	}

	printf("自ノード        ：%d\n", my_node);
	for (i = 0; i < ptr1->node_cnt; i++) {
		printf("参加中の他ノード：");
		printf("%d\n", ptr1->node_no[i]);
	}
	printf("\n");
}

//
//
//
static void DspNodeInfo(void)
{
	struct NodeInfo	node_info;
	unsigned short	rc;
	int				ret, node;

	// ノード番号を入力する。
	while (1) {
		printf("Enter node number (1-254) : ");
		ret = Wrapper_Scanf(&node);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (node < 1 || node > 254) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	
	rc = GetNodeInfo((unsigned short)node, &node_info, (unsigned short)dev_no);
	
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("MUTEX制御エラー\n");
			break;
		default:
			printf("[GetNodeInfo] Other Error = %d", rc);
			return;
		}
		return;
	}
	
	printf("\n[ノードNo.%d 情報]\n", node);
	if (node_info.sanka == 0) {
		printf("参加／離脱：離脱\n");
	} else {
		printf("参加／離脱：参加\n");
		printf("Common1 Address  = %d\n", node_info.cmn1_addr);
		printf("Common1 Size     = %d\n", node_info.cmn1_size);
		printf("Common2 Address  = %d\n", node_info.cmn2_addr);
		printf("Common2 Size     = %d\n", node_info.cmn2_size);
		printf("RCT              = %d\n", node_info.rct);
		printf("TW               = %d\n", node_info.tw);
		printf("Link Status      = %xH\n", node_info.link_status);
		if (node == my_node) {
			printf("My Node Status   = %xH\n", node_info.my_status);
		}
		printf("Vender           = %s\n", &node_info.vender[0]);
		printf("Maker            = %s\n", &node_info.maker[0]);
		printf("Node Name        = %s\n", &node_info.name[0]);
	}
}

//
//
//
static void DspNetInfo(void)
{
	unsigned short	rc;
	struct NetInfo	net_infp;

	rc = GetNetInfo(&net_infp, (unsigned short)dev_no);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("MUTEX制御エラー\n");
			break;
		default:
			printf("[GetNetInfo] Other Error = %d\n", rc);
			break;
		}
		return;
	}
	
	printf("RCT         = %d\n", net_infp.rct);
	printf("RC Current  = %d\n", net_infp.rcm);
	printf("RC Max      = %d\n", net_infp.rcm_max);
	printf("RC Min      = %d\n", net_infp.rcm_min);
	printf("Link Status = %xH\n", net_infp.link_status);
}

//
//
//
static void DspLogInfo(void)
{
	unsigned short	rc;
	unsigned int	log_buf[LOG_BUF_SIZE];
	
	memset(log_buf, 0, sizeof(log_buf));
	
	rc = GetLogInfo(&log_buf[0], (unsigned short)dev_no);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("MUTEX制御エラー\n");
			break;
		default:
			printf("[GetLogInfo] Other Error = %d\n", rc);
			break;
		}
		return;
	}
	
	Int_Dump(stdout, log_buf, LOG_BUF_SIZE);
}

//
//
//
static void UplStatus(void)
{
	unsigned short	rc;
	int				ret, status;

	while (1) {
		// 上位層ステータスを入力する。」
		printf("上位層ステータスを入力してください。（Hex/4byte）: ");
		ret = Wrapper_Hex_Scanf(&status, WORD_DATA_LEN);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}

	rc = SetUplStatus((unsigned short)status, (unsigned short)dev_no);
	switch (rc) {
	case 0:
		printf("正常に設定しました。\n");
		break;
	case PRM_ERR:
		printf("パラメタエラー\n");
		break;
	case TIM_OUT:
		printf("PCIからのレスポンスタイムアウト\n");
		break;
	case FATAL_ERR:
		printf("重大エラー\n");
		break;
	case DEV_NOT_OPEN:
		printf("指定デバイスはOpenしていません。\n");
		break;
	case MUTEX_ERR:
		printf("MUTEX制御エラー\n");
		break;
	default:
		printf("[SetUplStatus] Other Error = %d\n", rc);
		break;
	}
}

//
//
//
static void DspFatalErr(void)
{
	struct ErrInfo	err_infp, *ptr;
	unsigned short	rc;

	rc = GetErrInfo(&err_infp, (unsigned short)dev_no);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("パラメタエラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("指定デバイスはOpenしていません。\n");
			break;
		default:
			printf("[GetErrInfo] Other Error = %d\n", rc);
			break;
		}
		return;
	}
	
	ptr = &err_infp;
	printf("Error Code    = %xH\n", ptr->err_code);
	printf("Error Length  = %d\n", ptr->length);
	printf("Error Message = %s\n", &ptr->err_msg[0]);
}

//
//
//
static void MsgSendReceive(void)
{
	int	ret, number;

	while (1) {
		printf("\n********** Message Menu *************\n");
		printf("1) Message Send\n");
		printf("2) Check Message End\n");
		printf("3) Read Message\n");
		printf("4) Loop Back Test\n");
		printf("5) 終了\n");
		printf("Enter number : ");
		ret = Wrapper_Scanf(&number);
		if (ret != 0) {
			printf("Invalid data!\n");
			continue;
		}
		if (number < 1 || number > 5) {
			printf("Invalid data!\n");
			continue;
		}
		
		switch (number) {
		case 1:
			WriteMsg();
			break;
		case 2:
			CheckMsg();
			break;
		case 3:
			ReadMsg();
			break;
		case 4:
			MessageLoopBack();
			break;
		case 5:
			printf("\n");
			return;
		default:
			break;
		}
	}
}

//
//
//
static void WriteMsg(void)
{
	unsigned short		rc;
	struct MessageInfo	msg_prm;
	unsigned char		msg_buf[MSG_BUF_SIZE];
	int					ret, node, msg_no, wait_flg;
	
	memset(msg_buf, 0, sizeof(msg_buf));
	
	// ノード番号を入力する。
	while (1) {
		printf("送信先ノードNo.を指定してください。(1-255) : ");
		ret = Wrapper_Scanf(&node);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (node < 1 || node > 255) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}

	msg_prm.msg_node	= (unsigned char)node;
	msg_prm.msg_bc		= 0;
	msg_prm.msg_result	= 0;
	msg_prm.msg_adr		= 0;
	msg_prm.msg_lng		= 0;

	// TCDを入力する。
	while (1) {
		printf("どのメッセージを送りますか？\n");
		printf("1) ログクリア 要求\n");
		printf("2) ログリード 要求\n");
		printf("3) ネットワークパラメタリード 要求\n");
		printf("Enter number : ");
		ret = Wrapper_Scanf(&msg_no);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (msg_no < 1 || msg_no > 3) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	switch (msg_no) {
	case 1:
		msg_prm.msg_tcd = LGC_REQ;
		break;
	case 2:
		msg_prm.msg_tcd = LGR_REQ;
		break;
	case 3:
		msg_prm.msg_tcd = NTR_REQ;
		break;
	default:
		break;
	}

	// 終了待ち モードを選択する
	while (1) {
		printf("メッセージ送信 終了待ちしますか？\n");
		printf("1) 終了待ちしない\n");
		printf("2) 終了待ちする\n");
		printf("Enter number : ");
		ret = Wrapper_Scanf(&wait_flg);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (wait_flg < 1 || wait_flg > 2) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}

	wait_flg -= 1;

	rc = WriteMessage(&msg_prm, &msg_buf[0], (unsigned short)wait_flg,
					  (unsigned short)dev_no);

	switch (rc) {
	case 0:
		printf("メッセージ送信要求は正常終了しました。\n");
		break;
	case PRM_ERR:
		printf("パラメタエラー\n");
		break;
	case BUF_FUL:
		printf("バッファフル\n");
		break;
	case OUT_RING:
		printf("ノード離脱\n");
		break;
	case TX_FAIL:
		printf("送信失敗\n");
		break;
	case TIM_OUT:
		printf("PCIからのレスポンスタイムアウト\n");
		break;
	case FATAL_ERR:
		printf("重大エラー\n");
		break;
	case DEV_NOT_OPEN:
		printf("指定デバイスはOpenしていません。\n");
		break;
	case MUTEX_ERR:
		printf("MUTEX制御エラー\n");
		break;
	default:
		printf("[WriteMessage] Other Error = %d\n", rc);
		break;
	}
}

//
//
//
static void CheckMsg(void)
{
	unsigned short	rc;
	unsigned short	node, tcd;

	rc = CheckMessageEnd(&node, &tcd, (unsigned short)dev_no);
	
	switch (rc) {
	case 0:
		printf("メッセージが正常終了しました。 (node = %d) (TCD = %d)\n", node, tcd);
		break;
	case PRM_ERR:
		printf("パラメタエラー\n");
		break;
	case TX_FAIL:
		printf("送信失敗\n");
		break;
	case TX_NOT_END:
		printf("送信未完\n");
		break;
	case FATAL_ERR:
		printf("重大エラー\n");
		break;
	case DEV_NOT_OPEN:
		printf("指定デバイスはOpenしていません。\n");
		break;
	case MUTEX_ERR:
		printf("MUTEX制御エラー\n");
		break;
	default:
		printf("[CheckMessageEnd] Other Error = %d\n", rc);
		break;
	}
}

//
//
//
static void ReadMsg(void)
{
	struct MessageInfo	rcv_msg;
	unsigned short		rc;
	unsigned char		msg_buf[MSG_BUF_SIZE];
	int					ret, wait_flg;
	
	memset(msg_buf, 0, sizeof(msg_buf));
	
	// 終了待ち モードを選択する
	while (1) {
		printf("メッセージ受信 終了待ちしますか？\n");
		printf("1) 終了待ちしない\n");
		printf("2) 終了待ちする\n");
		printf("Enter number : ");
		ret = Wrapper_Scanf(&wait_flg);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (wait_flg < 1 || wait_flg > 2) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	wait_flg -= 1;

	rc = ReadMessage(&rcv_msg, &msg_buf[0], (unsigned short)wait_flg,
					 (unsigned short)dev_no);
	
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("パラメタエラー\n");
			break;
		case NO_DATA:
			printf("受信データなし\n");
			break;
		case TIM_OUT:
			printf("PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("MUTEX制御エラー\n");
			break;
		default:
			printf("[ReadMessage] Other Error = %d\n", rc);
			break;
		}
		return;
	}
	
	printf("受信データ表示\n");
	printf("node No. = %d\n", rcv_msg.msg_node);
	printf("result   = %d\n", rcv_msg.msg_result);
	printf("tcd      = %d\n", rcv_msg.msg_tcd);
	printf("bc       = %d\n", rcv_msg.msg_bc);
	Dump(stdout, msg_buf, rcv_msg.msg_bc);
}

//
// メッセージ折り返し試験 
// 
static void MessageLoopBack(void)
{
	struct MessageInfo		*snd_ptr;
	struct MessageInfo		*rcv_ptr;
	struct SankaReqTable	sanka_ptr;
	unsigned int			loop_cnt;
	unsigned int			node_cnt, j;
	unsigned char			dat, err_f, *d_p;
	unsigned short			rc, chk_node, chk_tcd;
	struct MessageInfo		msnd_tbl[MSG_SND_MAX];
	struct MessageInfo		mrcv_tbl[MSG_RCV_MAX];
	unsigned char			msnd_buf[MSG_SND_MAX][MSG_BUF_SIZE];
	unsigned char			mrcv_buf[MSG_RCV_MAX][MSG_BUF_SIZE];
	unsigned short			msnd_cnt = 0;
	unsigned short			mrcv_cnt = 0;
	unsigned short			msnd_st_pnt = 0;
	unsigned short			msnd_tx_pnt = 0;
	unsigned short			mrcv_st_pnt = 0;
	int						ret, i, wait_flg, tmp_i;
	
	memset(msnd_tbl, 0, sizeof(msnd_tbl));
	memset(mrcv_tbl, 0, sizeof(mrcv_tbl));
	memset(msnd_buf, 0, sizeof(msnd_buf));
	memset(mrcv_buf, 0, sizeof(mrcv_buf));
	
	// メッセージ折り返し試験要求 メッセージを生成する。
	snd_ptr = &msnd_tbl[msnd_st_pnt];
	while (1) {
		printf("相手先ノードNo.を指定してください。（2-254）\n");
		ret = Wrapper_Scanf(&tmp_i);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (tmp_i < 2 || tmp_i > 254) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}

	snd_ptr->msg_node = (unsigned char)tmp_i;

	while (1) {
		printf("データバイト数を指定してください。（0-1024)\n");
		ret = Wrapper_Scanf(&tmp_i);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (tmp_i > MSG_BUF_SIZE) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}

	// 終了待ち モードを選択する
	while (1) {
		printf("メッセージ受信 終了待ちしますか？\n");
		printf("1) 終了待ちしない\n");
		printf("2) 終了待ちする\n");
		printf("Enter number : ");
		ret = Wrapper_Scanf(&wait_flg);
		if (ret != 0) {
			printf("Invalid data!\n\n");
			continue;
		}
		if (wait_flg < 1 || wait_flg > 2) {
			printf("Invalid data!\n\n");
			continue;
		}
		break;
	}
	wait_flg -= 1;

	snd_ptr->msg_bc		= (unsigned short)tmp_i;
	snd_ptr->msg_result	= 0;
	snd_ptr->msg_lng	= 0;
	snd_ptr->msg_adr	= 0;
	snd_ptr->msg_tcd	= LPB_REQ;

	// 折り返しデータをインクリメントデータでクリアする。
	dat = 0;
	d_p = &msnd_buf[msnd_st_pnt][0];
	for (i = 0; i < tmp_i; i++) {
		*d_p++ = dat;
		dat++;
	}
	msnd_st_pnt += 1;
	msnd_cnt += 1;

	// Broad Castの場合、参加中の他ノード数を調べる
	if (snd_ptr->msg_node == 255) {
		// 参加ノード情報要求コマンドをセット。
		rc = GetRingInfo(&sanka_ptr, (unsigned short)dev_no);
		if (rc > 0) {
			switch (rc) {
			case PRM_ERR:
				printf("[GetRingInfo] パラメタエラー\n");
				break;
			case TIM_OUT:
				printf("[GetRingInfo] PCIからのレスポンスタイムアウト\n");
				break;
			case FATAL_ERR:
				printf("[GetRingInfo] 重大エラー\n");
				break;
			case DEV_NOT_OPEN:
				printf("[GetRingInfo] 指定デバイスはOpenしていません。\n");
				break;
			case MUTEX_ERR:
				printf("[GetRingInfo] MUTEX制御エラー\n");
				break;
			default:
				printf("[GetRingInfo] Other Error = %d\n", rc);
				break;
			}
			return;
		}
		
		// 他ノード数を得る
		node_cnt = sanka_ptr.node_cnt;
	} else {	// 1:1メッセージのときは、他ノード数＝１固定
		node_cnt = 1;
	}
	
	loop_cnt = 0;
	err_f = 0;
	for ( ; ; ) {
		// メッセージを送信する
		if (msnd_cnt > 0) {
			rc = WriteMessage(&msnd_tbl[msnd_tx_pnt], &msnd_buf[msnd_tx_pnt][0],
							 (unsigned short)wait_flg, (unsigned short)dev_no);
			if (rc == 0) {
				msnd_cnt -= 1;
				msnd_tx_pnt += 1;
				if (msnd_tx_pnt == MSG_SND_MAX) {
					msnd_tx_pnt = 0;
				}
			} else {
				// エラーの時は終了、バッファフルの時は、リトライする。
				if (rc != BUF_FUL) {
					printf("メッセージ送信 要求時 エラーが発生しました。\n");
					err_f = 1;
					break;
				}
			}
		}
		rc = CheckMessageEnd(&chk_node, &chk_tcd, (unsigned short)dev_no);
		if (rc == 0) {
			printf("メッセージ送信が正常終了しました。\n");
		} else {
			if (rc != TX_NOT_END) {
				printf("メッセージ送信時エラーが発生しました。\n");
				err_f = 1;
				break;
			}
		}
		rc = ReadMessage(&mrcv_tbl[mrcv_st_pnt], &mrcv_buf[mrcv_st_pnt][0],
						 (unsigned short)wait_flg, (unsigned short)dev_no);
		if (rc == 0) {
			printf("メッセージ受信が正常終了しました。\n");
			mrcv_cnt += 1;
			mrcv_st_pnt += 1;
			if (mrcv_st_pnt == MSG_RCV_MAX) {
				mrcv_st_pnt = 0;
			}
		} else {
			if (rc != NO_DATA) {
				printf("メッセージ受信時 エラーが発生しました。\n");
				err_f = 1;
				break;
			}
		}
		
		// 終了条件をチェックする
		if (mrcv_cnt == node_cnt) {		// 正常終了
			printf("全てのメッセージ送受信を正常終了しました。\n");
			break;
		} else {
			loop_cnt +=1;
			if (loop_cnt == 0xfffffff) {
				printf("タイムアウトです。処理を終了します。\n");
				err_f = 1;
			}
		}
		if (err_f == 1) {		// 異常終了
			break;
		}
	}
	if (err_f == 1) {
		switch (rc) {
		case PRM_ERR:
			printf("パラメタエラー\n");
			break;
		case BUF_FUL:
			printf("バッファフル\n");
			break;
		case NO_DATA:
			printf("受信データなし\n");
			break;
		case OUT_RING:
			printf("ノード離脱\n");
			break;
		case TX_FAIL:
			printf("送信失敗\n");
			break;
		case TX_NOT_END:
			printf("送信未完\n");
			break;
		case TIM_OUT:
			printf("PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("MUTEX制御エラー\n");
			break;
		default:
			printf("Other Error = %d\n", rc);
			break;
		}
		return;
	}
	
	// 受信メッセージ データのチェックを行う
	if (err_f == 0) {
		for (j = 0; j < mrcv_cnt; j++) {
			rcv_ptr = &mrcv_tbl[j];
			d_p = &mrcv_buf[j][0];
			dat = 0;
			err_f = 0;
			for (i = 0; i < rcv_ptr->msg_bc; i++) {
				if (*d_p != dat) {
					err_f = 1;
					break;
				}
				d_p++;
				dat++;
			}
			if (err_f == 1) {
				printf("折り返しデータ比較 不一致 ノードNo. = %d\n", rcv_ptr->msg_node);
				printf("rcv = %d snd = %d\n", *d_p, dat);
			} else {
				printf("折り返しデータ比較 正常 ノードNo. = %d\n", rcv_ptr->msg_node);
			}
		}
	}
	printf("\n");
}

//
// ULS ＆ LKSの表示
//
static void UlsLks(void)
{
	struct UlsInfo	uls_info, *uls_p;
	unsigned short	rc, i;

	rc = ReadUls_All(&uls_info, (unsigned short)dev_no);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("MUTEX制御エラー\n");
			break;
		default:
			printf("[ReadUls_All] Other Error = %d\n", rc);
			break;
		}
		return;
	}
	
	uls_p = &uls_info;
	for (i = 0; i < 254; i++) {
		if ((uls_p->lks[i] & 0x01) == 1 || (i + 1) == my_node) {	/* 参加 or 自ノード */
			printf("[Node %3d] ULS = %04xH, LKS = %02xH \n",
				   i+1, uls_p->uls[i], uls_p->lks[i]);
		}
	}
}

//
//
//
static void Firmware_ver_dsp(void)
{
	unsigned short	fversion, rc;
	unsigned short	ver_hi, ver_low;

	rc = Firmware_Version(&fversion, (unsigned short)dev_no);
	switch (rc) {
	case 0:
		break;
	case PRM_ERR:
		printf("パラメタエラー\n");
		return;
	case TIM_OUT:
		//FLPCI Ver47(FL-net Ver1)以前のファームウェアはタイムアウトで終了する
		printf("Unknown Version\n");
		return;
	case FATAL_ERR:
		printf("重大エラー\n");
		return;
	case DEV_NOT_OPEN:
		printf("指定デバイスはOpenしていません。\n");
		return;
	case MUTEX_ERR:
		printf("MUTEX制御エラー\n");
		return;
	default:
		printf("[Firmware_Version] Other Error = %d\n", rc);
		return;
	}
	
	ver_low = fversion & 0xff;
	ver_hi  = (fversion & 0xff00) >> 8;
	if (ver_hi == 20) {		/* FL-PCI/v2-100 */
		printf("Firmware Version = %02d%02d\n", ver_hi, ver_low );
	} else {				/* FL-PCI/v2 */
		if (ver_low > 0) {
			printf("Firmware Version = %02d%02d\n", ver_hi, ver_low );
		} else {
			//FLPCI Ver1004(FL-net Ver2)以前のファームウェアはver_low=0固定値が通知される
			printf("Unknown Version");
		}
	}
}

//
// 100Mbps support 06/03/26
//
static void DspEtherMode(void)
{
	unsigned short	rc;
	unsigned int	speed, duplex;

	rc = Get_Speed(&speed, &duplex, (unsigned short)dev_no);
	if (rc > 0) {
		switch (rc) {
		case PRM_ERR:
			printf("パラメタエラー\n");
			break;
		case TIM_OUT:
			printf("PCIからのレスポンスタイムアウト\n");
			break;
		case FATAL_ERR:
			printf("重大エラー\n");
			break;
		case DEV_NOT_OPEN:
			printf("指定デバイスはOpenしていません。\n");
			break;
		case MUTEX_ERR:
			printf("MUTEX制御エラー\n");
			break;
		case LNK_DOWN:
			printf("ケーブルが外れているため速度設定は未完です。\n");
			break;
		default:
			printf("[Get_Speed] Other Error = %d\n", rc);
			break;
		}
		return;
	}
	
	if (speed == 0) {
		if (duplex == 0) {
			printf("通信速度の設定は 10Mbps 半2重です。\n");
		} else {
			printf("通信速度の設定は 10Mbps 全2重です。\n");
		}
	} else {
		if (duplex == 0) {
			printf("通信速度の設定は 100Mbps 半2重です。\n");
		} else {
			printf("通信速度の設定は 100Mbps 全2重です。\n");
		}
	}
	return;
}

/*
 * 機  能：IPアドレスの文字列確認関数
 * 引  数：value    ：IPアドレス文字列
 * 戻り値：int      ：正常終了時は正の整数0を返します。
 *       ：         ：それ以外の場合は負の整数を返します。
 * 説  明：IPアドレスの文字列が正常か確認します。
 */
static int Address_check(const char *value)
{
	char	*mem, *ptr1, *ptr2;
	int		data, i;

	/* 引数の確認 */
	if (!value) {
		return -1;
	}
	
	/* 作業メモリ確保 */
	if ((mem = strdup(value)) == NULL) {
		return -1;
	}

	/* 入力可能文字の確認 */
	if (Strskip(mem, "0123456789.") != NULL) {
		return -1;
	}

	/* 最初の3つまで確認 */
	ptr1 = mem;
	for (i = 0; i < 3; i++) {
		if ((ptr2 = Strskipsearch(ptr1, ".")) == NULL) {
			free(mem);
			return -1;
		}
		*ptr2 = '\0';
		++ptr2;
		if (isdigit((int)*ptr1) == 0) {
			free(mem);
			return -1;
		}
		data = atoi(ptr1);
		if ((data < 0) || (255 < data)) {
			free(mem);
			return -1;
		}
		ptr1 = ptr2;
	}

	/* 最後の4つ目を確認 */
	if (Strskipsearch(ptr1, ".") != NULL) {
		free(mem);
		return -1;
	}
	if (isdigit((int)*ptr1) == 0) {
		free(mem);
		return -1;
	}
	data = atoi(ptr1);
	if ((data < 0) || (255 < data)) {
		free(mem);
		return -1;
	}

	free(mem);

	return 0;
}

/*
 * 機  能：文字スキップ関数
 * 引  数：ptr  ：元文字列
 *         skip ：SKIP文字列
 * 戻り値：SKIP終了位置
 * 説  明：文字列の先頭にSKIP文字が現れるまで文字位置をSKIPさせる。
 *         データが無かったらNULLを返す。
 */
static char* Strskipsearch(const char *ptr, const char *skip)
{
	while (*ptr != '\0') {
		if (strchr(skip, (int)*ptr) != NULL) {
			break;
		}
		++ptr;
	}
	if (*ptr != '\0') {
		return (char*)ptr;
	}
	return NULL;
}

/*
 * 機  能：文字スキップ関数
 * 引  数：ptr  ：元文字列
 *         skip ：SKIP文字列
 * 戻り値：SKIP終了位置
 * 説  明：文字列の先頭にSKIP文字がある場合は文字位置をSKIPさせる。
 */
static char* Strskip(const char *ptr, const char *skip)
{
	while (*ptr != '\0') {
		if (strchr(skip, (int)*ptr) == NULL) {
			break;
		}
		++ptr;
	}
	if (*ptr != '\0') {
		return (char*)ptr;
	}
	return NULL;
}

/*
 * 機  能：数字文字列確認関数
 * 引  数：ptr   ：出力文字列
 * 戻り値：int   ：正常終了時は0を返します。
 *       ：      ：それ以外の場合は負の整数を返します。
 * 説  明：文字列が数字もじか確認します。
 */
static int Strnumeric(char *ptr)
{
	if (Strskip(ptr, STRING_NUMERIC) == NULL)
		return 0;
	else
		return -1;
}

static void Cyc_Dump(FILE *fp, unsigned short *data, unsigned short addr, int len)
{
	int				i, base;
	unsigned short	*buf = data;
	
	for (base = 0; base < len; base += 10, addr += 10) {
		fprintf(fp, "%04d - ", addr);
		for (i = base; i < base + 10; i++) {
			if (i < len) {
				fprintf(fp, "%04x ", buf[i]);
			} else {
				fprintf(fp, "   ");
			}
		}
		fprintf(fp, "\n");
	}
}

static void Dump(FILE *fp, unsigned char *data, int len)
{
	int				i;
	int				base;
	unsigned char	*buf = data;
	
	for (base = 0; base < len; base += 16) {
		fprintf(fp, "%04x ", base);
		for (i = base; i < base + 16; i++) {
			if (i % 16 == 8) {
				fprintf(fp, "- ");
			}
			if (i < len) {
				fprintf(fp, "%02x ", buf[i]);
			} else {
				fprintf(fp, "   ");
			}
		}
		fprintf(fp, "  ");
		for (i = base; i < base + 16; i++) {
			if (i < len) {
				if (isprint(buf[i])) {
					fprintf(fp, "%c", buf[i]);
				} else {
					fprintf(fp, ".");
				}
			} else {
				break;
			}
		}
		fprintf(fp, "\n");
	}
}

static void Int_Dump(FILE *fp, unsigned int *data, int len)
{
	int				i;
	int				base;
	unsigned int	*buf = data;
	
	for (base = 0; base < len; base += 10) {
		fprintf(fp, "%04d ", base);
		for (i = base; i < base + 10; i++) {
			if (i < len) {
				fprintf(fp, "%08x ", buf[i]);
			} else {
				fprintf(fp, "   ");
			}
		}
		fprintf(fp, "\n");
	}
}

static int Wrapper_Scanf(int *number)
{
	char	buf[ENTER_BUF_SIZE];
	int 	ret;
	
	memset(buf ,0, sizeof(buf));
	scanf(" %s", buf);
	ret = Strnumeric(buf);
	if (ret != 0) {
		return -1;
	}
	*number = atoi(buf);
	
	return 0;
}

static int Wrapper_Hex_Scanf(int *number, unsigned int len)
{
	char	buf[ENTER_BUF_SIZE];
	
	memset(buf ,0, sizeof(buf));
	scanf(" %s", buf);
	// 文字列長チェック。
	if (strlen(buf) > len) {
		return -1;
	}
	if (Strskip(buf, STRING_HEX_NUMERIC) != NULL) {
		return -1;
	}
	*number = axtoi(buf);
	
	return 0;
}

/*
 * 機  能：16進文字列から数値への変換関数
 * 引  数：buffer  ：変換対照の文字列
 * 戻り値：正常終了した場合は変換結果の数値、エラーが発生した場合は0
 * 説  明：16進文字列から数値への変換を行います
 */
static int axtoi(const char *buffer)
{
	static unsigned char da[] = {
		0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F
	};
	int i;
	int ret = 0;
	for (i = 0; ; i++) {
		if (*buffer == 0x00)
			break;
		if ((0x30 <= *buffer) && (*buffer <= 0x39)) {
			ret <<= 4;
			ret += *buffer & 0x0F;
		}
		else if ((0x41 <= *buffer) && (*buffer <= 0x46)) {
			ret <<= 4;
			ret += da[*buffer - 0x41];
		}
		else if ((0x61 <= *buffer) && (*buffer <= 0x66)) {
			ret <<= 4;
			ret += da[*buffer - 0x61];
		}
		else
			break;
		++buffer;
	}
	return ret;
}
